########################################################################
# metaf2xml/XML.pm -- write data of METAR, TAF, and/or SYNOP message to XML file
#
# copyright (c) metaf2xml 2006-2009
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

package metaf2xml::XML;

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;

########################################################################
# export the functions provided by this module
########################################################################
BEGIN {
    use Exporter();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

    $VERSION     = sprintf "%d.%03d", '$Revision: 1.36 $0.0' =~ /(\d+)/g;
    @ISA         = qw(Exporter);
    @EXPORT      = ();
    %EXPORT_TAGS = ();
    @EXPORT_OK   = qw(startXML printReport finishXML);
}
our @EXPORT_OK;
my $VERSION_ID = 'XML.pm: $Id: XML.pm,v 1.36 2009/02/28 09:40:03 metaf2xml Exp $';

END { }

=head1 NAME

metaf2xml::XML -- write parsed data of METAR/TAF/SYNOP messages as XML

=head1 SYNOPSIS

  use metaf2xml::XML;
  metaf2xml::XML::startXML(\%opts,\@versions);
  metaf2xml::XML::printReport(\%report);
  ...
  metaf2xml::XML::finishXML();

=head1 DESCRIPTION

This module writes the data obtained from parsing METAR, TAF, and/or SYNOP
messages as XML.

=head1 FUNCTIONS

=cut

########################################################################
# global variables
########################################################################
# number of spaces to prepend. -1: file not open
my $indent_depth = -1;

########################################################################
# helper functions to write XML
########################################################################

sub _XMLtag {
    my $type = shift;
    my $tag = shift;

    print XML ' ' x $indent_depth . '<' . $tag;
    for (my $i = 0; $i < $#_; $i += 2) {
        my ($name, $value) = @_[$i, $i + 1];
        $value =~ s/&/&amp;/g; # must be the first substitution
        $value =~ s/</&lt;/g;
        $value =~ s/>/&gt;/g;
        $value =~ s/"/&quot;/g;
        print XML ' ' . $name . '="' . $value . '"';
    }
    if ($type eq 'start') {
        $indent_depth++;
    } else {
        print XML '/';
    }
    print XML ">\n";
}

sub _XMLstartTag {
    return _XMLtag 'start', @_;
}

sub _XMLendTag {
    print XML ' ' x --$indent_depth . '</' . $_[0] . '>' . "\n";
}

sub _XMLemptyTag {
    return _XMLtag 'empty', @_;
}

sub _XMLdataElement {
    print XML ' ' x $indent_depth . '<' . $_[0] . '>' . $_[1] .
              '</' . $_[0] . ">\n";
}

########################################################################
# helper functions
########################################################################

sub _printXML;
sub _printXML {
    my ($r, $tag, $xml_tag) = @_;

    return unless exists $r->{$tag};
    $r = $r->{$tag};

    $tag = $xml_tag if $xml_tag;
    if (ref $r eq 'HASH') {
        my (@subnodes, @attrs);

        # sort, so XML looks nicer
        for (qw/errorType warningType s v rp rn u q/) {
            push @attrs, $_, $r->{$_} if exists $r->{$_};
        }

        # special tag sequence to comply with the sequence of tags in DTD
        if ($tag eq 'cloudOpacityLvl') {
            @subnodes = qw/locationAnd oktas weather cloudType cloudBase/;
        } elsif ($tag =~ /^(?:tempMinNighttime|tempMaxDaytime|tempMinPeriod|tempMaxPeriod)$/) {
            @subnodes = qw/period periodNotAvailable notAvailable temp/;
        } elsif ($tag eq 'pressureChange') {
            @subnodes = qw/period notAvailable pressureChangeVal pressureTendency/;
        } elsif ($tag eq 'reportModifier') {
            @subnodes = qw/modifierType isLastSegment segment bulletinSeq sequenceLost over24hLate/;
        } elsif ($tag =~ /^(?:phenomenonAtLoc|phenomenonOnly)$/) {
            @subnodes = qw/phenomDescrPre phenomDescrPost weather cloudType cloudCover lightningType otherPhenom locationAnd MOV MOVD cloudTypeAsoctd cloudTypeEmbd/;
        } elsif ($tag eq 'cloudInfo') {
            @subnodes = qw/cloudBaseNotAvailable cloudBase cloudBaseFrom cloudBaseTo visVert visVertFrom visVertTo cloudTypeNotAvailable cloudType cloudOktas/;
        } elsif ($tag eq 'cloudLocation') {
            @subnodes = qw/cloudTypeNotAvailable cloudType cloudNone cloudInvisible cloudDir topsInvisible elevationAngle/;
        } elsif ($tag =~ /^(?:radShortWave|radDirectSolar)$/) {
            @subnodes = qw/radiationPeriod notAvailable radiationValue/;
        } elsif ($tag eq 'location') {
            @subnodes = qw/inDistance inVicinity distance isGrid isQuadrant obscgMtns locationSpec compassDir quadrant/;
        } elsif ($tag eq 'climate') {
            @subnodes = qw/traces precipAmount1Inch precipAmount1MM precipAmount2Inch precipAmount2MM temp1 temp2/;
        } elsif ($tag eq 'precipHourly') {
            @subnodes = qw/period period3o6h periodNotAvailable timeBeforeObsSince notAvailable precipAmount traces/;
        } elsif ($tag eq 'snowFall') {
            @subnodes = qw/period timeBeforeObsSince noMeasurement precipAmount traces /;
        } elsif ($tag eq 'highestMeanSpeed') {
            @subnodes = qw/hour minute measurePeriod period timeBeforeObs timeBeforeObsBegin timeBeforeObsSince timeBeforeObsEnd wind/;
        } else {
            for (sort keys %$r) {
                if (!/^(?:[vuqs]|rp|rn|errorType|warningType)$/) {
                    push @subnodes, $_;
                }
            }
        }

        # *Arr are arrays with different subnodes. Suppress tag in XML.
        if ($#subnodes > -1) {
            _XMLstartTag $tag, @attrs unless $tag =~ /Arr$/;
            for (@subnodes) {
                _printXML $r, $_;
            }
            _XMLendTag $tag unless $tag =~ /Arr$/;
        } else {
            _XMLemptyTag $tag, @attrs unless $tag =~ /Arr$/;
        }
    } elsif (ref $r eq 'ARRAY') {
        if ($#$r > -1) {
            for (@$r) {
                _printXML { $tag => $_ }, $tag;
            }
        } else {
            _XMLemptyTag $tag;
        }
    } elsif ($tag eq 'version') {
        _XMLdataElement 'version', $r;
    } else {
        _XMLemptyTag $tag, defined $r ? ('v',  $r) : ();
    }
}

########################################################################
# variable shared by startXML, finishXML
########################################################################
my $opt_X;

########################################################################
# startXML
########################################################################

=head2 startXML(\%opts,\@versions)

This function opens the file and writes to it the optional data as specified.

The following arguments are expected:

=over

=item C<opts>

Reference to hash of options. The following keys of the hash are recognised:

=for html <ul>

=over

=item C<x>, with value B<E<lt>filenameE<gt>>

enables writing the data to <filename>

=item C<X>

with option B<x>, wrap data in <data>...</data>

=item C<D>

with option B<x>, include DOCTYPE and reference to DTD (implies option B<X>)

=item C<S>, with value B<E<lt>filenameE<gt>>

with option B<x>, include reference to stylesheet <filename> (implies option
B<X>)

=item C<O>, with value B<E<lt>optionsE<gt>>

with option B<x>, include <options> (a space separated list of options)
(implies option B<X>)

=for html </ul>

=back

=item C<versions>

Reference to array of version strings. They are written to the XML file if the
option C<X> is given.

=back

=cut

sub startXML {
    my ($opts, $versions) = @_;

    return 0 unless $indent_depth == -1 && exists $opts->{x};

    # open file, set state
    open(XML, ">$opts->{x}") || return 0;
    $indent_depth = 0;

    # XML header
    print XML '<?xml version="1.0" encoding="UTF-8"?>' . "\n";

    # write additional info to XML, if required
    $opt_X =    exists $opts->{X}
             || exists $opts->{D} || exists $opts->{S} || exists $opts->{O};
    if ($opt_X) {

        # XML meta data
        print XML '<!DOCTYPE data SYSTEM "metaf.dtd">' . "\n"
            if defined $opts->{D};
        print XML '<?xml-stylesheet href="' . $opts->{S} . '" type="text/xsl"?>'
                  . "\n"
            if defined $opts->{S};

        print XML "\n";

        # opening data tag
        _XMLstartTag 'data';

        # version stuff
        _printXML { version => [ @$versions, $VERSION_ID ] }, 'version';

        if (exists $opts->{O}) {
            my (%options, $msg);
            (@options{'type_metaf', 'type_synop', 'lang', 'format', 'src_metaf', 'src_synop', 'mode'}, $msg)
                                                     = split / /, $opts->{O}, 8;
            @options{'msg_metaf', 'msg_synop'} = split '  ', $msg, 2;
            _printXML { options => \%options }, 'options';
        }
    }
    _XMLstartTag 'reports', query_start => gmtime().'';
    return 1;
}

########################################################################
# printReport
########################################################################

=head2 printReport(\%report)

This function writes the content of C<report> to the XML file.
The data for each message will be enclosed in a tag C<<metar>>, C<<taf>>, or
C<<synop>>. The function can be invoked repeatedly.
L<startXML()|/startxml___opts___versions_> must have been invoked already.
L<finishXML()|/finishxml__> must be invoked to complete the XML file.

The following argument is expected:

=over

=item C<report>

reference to hash of data to be written to the file

=back

=cut

sub printReport {
    my $report = shift;

    return 0 unless $indent_depth >= 0;

    if (exists $report->{isSynop}) {
        _XMLstartTag 'synop', 's', $report->{msg};

        for (qw/ERROR warning
                obsStationType callSign obsTime reportModifier
                windIndicator stationPosition obsStationId precipInd wxInd
                baseLowestCloud visPrev visibilityAtLoc totalCloudCover
                sfcWind temperature stationPressure SLPhPa gpSurface
                pressureChange precipHourly weatherSynop cloudTypes
                exactObsTime/)
        {
            _printXML $report, $_;
        }

        if (exists $report->{section3}) {
            if ($#{$report->{section3}} > -1) {
                _XMLstartTag 'section3', s => '333';
                _printXML { section3Arr => $report->{section3}}, 'section3Arr';
                _XMLendTag 'section3';
            } else {
                _XMLemptyTag 'section3', s => '333';
            }
        }

        if (exists $report->{section4}) {
            if ($#{$report->{section4}} > -1) {
                _XMLstartTag 'section4', s => '444';
                _printXML { section4Arr => $report->{section4}}, 'section4Arr';
                _XMLendTag 'section4';
            } else {
                _XMLemptyTag 'section4', s => '444';
            }
        }

        if (exists $report->{section5}) {
            if ($#{$report->{section5}} > -1) {
                _XMLstartTag 'section5', s => '555';
                _printXML { section5Arr => $report->{section5}}, 'section5Arr';
                _XMLendTag 'section5';
            } else {
                _XMLemptyTag 'section5', s => '555';
            }
        }

        _XMLendTag 'synop';
    } else {
        my $is_taf;

        $is_taf = exists $report->{isTaf};
        _XMLstartTag $is_taf ? 'taf' : 'metar', 's', $report->{msg};
        for (qw/ERROR warning isSpeci
                obsStationId obsTime issueTime fcstPeriod reportModifier
                fcstCancelled fcstNotAvbl
                sfcWind windShearLvlArr
                CAVOK visPrev visMin visRwy RVRNO weather cloud visVert
                temperature QNH
                somePressure cloudMaxCover recentWeather windShear rwyState
                colourCode NEFO_PLAYA RH windAtLoc rwyWind TAFsupplArr/)
        {
            _printXML $report, $_;
        }

        if (exists $report->{trend}) {
            for my $td (@{$report->{trend}}) {
                _XMLstartTag 'trend', 's', $td->{s};
                for (qw/trendType trendTime1 trendTime2 probability
                        sfcWind CAVOK visPrev visVert weather cloud
                        rwyState colourCode TAFsupplArr/)
                {
                    _printXML $td, $_;
                }
                _XMLendTag 'trend';
            }
        }

        _printXML $report, 'TAFinfoArr';
        _printXML $report, 'remark', $is_taf ? 'tafRemark' :'remark';

        _XMLendTag $is_taf ? 'taf' : 'metar';
    }
    return 1;
}

########################################################################
# finishXML
########################################################################

=head2 finishXML()

This function must be invoked to complete the writing to the file.
L<startXML()|/startxml___opts___versions_> must have been invoked already.

No arguments are expected.

=cut

sub finishXML {
    return 0 unless $indent_depth >= 0;

    _XMLendTag 'reports';

    # closing data tag, if required
    _XMLendTag 'data' if $opt_X;

    # set state, close file
    $indent_depth = -1;
    return close XML;
}

=head1 SEE ALSO

=begin html

<a href=parser.pm.html>metaf2xml::parser(3pm)</a>,
<a href=metaf2xml.pl.html>metaf2xml(1)</a>,
<a href=metaf.pl.html>metaf(1)</a>, <!--

=end html

B<metaf2xml::parser>(3pm),
B<metaf2xml>(1),
B<metaf>(1),

=for html -->

L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT, LICENSE, DISCLAIMER

copyright (c) 2006-2009 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

=cut

1;
