########################################################################
# metaf2xml/parser.pm -- parse a METAR, TAF, or SYNOP message
#
# copyright (c) metaf2xml 2006-2009
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

package metaf2xml::parser;

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;

use POSIX qw(floor);

########################################################################
# export the functions provided by this module
########################################################################
BEGIN {
    use Exporter();
    our ($VERSION, $VERSION_ID, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

    $VERSION     = sprintf "%d.%03d", '$Revision: 1.49 $0.0' =~ /(\d+)/g;
    $VERSION_ID  = 'parser.pm: $Id: parser.pm,v 1.49 2009/02/28 09:41:05 metaf2xml Exp $';
    @ISA         = qw(Exporter);
    @EXPORT      = ();
    %EXPORT_TAGS = ();
    @EXPORT_OK   = qw(parseReport);
}
our @EXPORT_OK;

END { }

=head1 NAME

metaf2xml::parser -- parse a METAR, TAF, or SYNOP message

=head1 SYNOPSIS

  use metaf2xml::parser;
  my %report = metaf2xml::parser::parseReport($msg,$is_taf,$is_synop);

=head1 DESCRIPTION

This module contains functions to analyze a string (per default as a METAR
message). Its function L<parseReport()|/parsereport__msg__is_taf__is_synop_>
returns a hash with all its components.

=cut

########################################################################
# define lots of regular expressions
########################################################################
my $re_day = '(?:[0-2]\d|3[01])';
my $re_hour = '(?:[01]\d|2[0-3])';
my $re_min = '[0-5]\d';
my $re_bbb = '(?:(?:RR|CC|AA)[A-Z]|P[A-Z]{2})';
my $re_rwy = '(?:0[1-9]|[12]\d|3[0-6])';
my $re_rwy_des = "$re_rwy(?:[LCR]|LL|RR)?";
my $re_rwy_des2 = '(?:[05][1-9]|[1267]\d|[38][0-6])'; # add 50 for rrR
my $re_wind_speed = 'P?[1-9]?\d\d';
my $re_wind_speed_unit = '(?:KTS?|TK|K T| KT|MPS|KMH)';
my $re_wind_dir = '(?:[0-2]\d|3[0-6])';
# EXTENSION: allow wind direction not rounded to 10 degrees
my $re_wind = "(VRB|///|$re_wind_dir\\d)(//|$re_wind_speed)(?:G($re_wind_speed))?($re_wind_speed_unit)";
my $re_vis = '(?:VIS|VSBY)';
my $re_vis_m = '\d{4}';
my $re_vis_km = '[1-9]\d?KM';
my $re_vis_m_km_remark = "(?:$re_vis_m|[1-9]\\d{2,3}M|$re_vis_km)";
my $re_frac16 = '[135]/16';
my $re_frac8 =  '[1357]/8';
my $re_frac4 =  '[13]/4';
my $re_frac2 =  '1/2';
my $re_vis_frac_sm0 = "(?:$re_frac16|$re_frac8|$re_frac4|$re_frac2)";
# MANAIR 2.6.8: no space between whole miles and fraction
my $re_vis_frac_sm1 = "1(?: ?(?:$re_frac8|$re_frac4|$re_frac2))";
# EXTENSION: allow /8 for 2 miles
my $re_vis_frac_sm2 = "2(?: ?(?:$re_frac8|$re_frac4|$re_frac2))";
# EXTENSION: allow /2 for 3 miles
my $re_vis_frac_sm3 = "3(?: ?(?:$re_frac2))";
my $re_vis_whole_sm = '(?:[1-9]\d0|1[0-5]|[2-9][05]|\d)'; # longest match first
my $re_vis_sm =   "(?:M1/4"
                 . "|$re_vis_frac_sm0"
                 . "|$re_vis_frac_sm1"
                 . "|$re_vis_frac_sm2"
                 . "|$re_vis_frac_sm3"
                 . "|$re_vis_whole_sm)"; # last, to find fractions first
my $re_gs_size = "(?:M1/4|[1-9]\\d*(?: $re_frac4| $re_frac2)?|$re_frac4|$re_frac2)";
my $re_rwy_vis = '[PM]?\d{4}';
my $re_rwy_wind = "(?:WIND )?(?:RWY($re_rwy_des) ?|($re_rwy_des)/)((?:///|VRB|${re_wind_dir}0)(?://|$re_wind_speed(?:G$re_wind_speed)?)$re_wind_speed_unit)(?: ($re_wind_dir\\d)V($re_wind_dir\\d))?";
my $re_rwy_wind2 = "((?:///|VRB|${re_wind_dir}0)(?://|$re_wind_speed(?:G$re_wind_speed)?)$re_wind_speed_unit)/RWY($re_rwy_des)";
my $re_rs_rwy_wind = "(RS|$re_rwy_des)($re_wind_dir)($re_wind_speed(?:G$re_wind_speed)?$re_wind_speed_unit)";

my $re_compass_dir = '(?:[NS]?[EW]|[NS])';
my $re_compass_dir16 = "(?:[NE]NE|[ES]SE|[SW]SW|[NW]NW|$re_compass_dir)";

my $re_phen_locdir_comp = "(?:(?:GRID )?(?:[1-9]\\d*(?:KM)? ?)?$re_compass_dir16(?: QUAD)?)";
my $re_phen_locdir_comp_thru = "(?:O?MTNS|OMTS|(?:OVR|ALG) (?:LK|RIVER|MT(?:N?S)?|VLYS?)|(?:(?:OV?HD|(?:OVE?R |AT |NR )?A(?:P|R?PT|D)|IN VLY|STNRY|(?:DISTANT |DSNT |VC )?(?:OBSCG MTNS )?(?:$re_phen_locdir_comp|ALQDS|V/D))(?:(?: ?- ?| THRU )(?:$re_phen_locdir_comp|OV?HD))*))";
my $re_phen_locdir_comp_and = "(?:$re_phen_locdir_comp_thru(?:(?:[/ ]| AND )$re_phen_locdir_comp_thru)*)";
my $re_phen_locdirs = "(?: $re_phen_locdir_comp_and)";
my $re_wx_mov_d2 = "(?: (MOVD?) ($re_compass_dir16(?:-$re_compass_dir16)*|OV?HD|STNRY|UNKN))";
my $re_num_quadrant1 = '(?:1(?:ST)?|2(?:ND)?|3(?:RD)?|4(?:TH)?)';
my $re_num_quadrant = "(?: $re_num_quadrant1(?:[- ]$re_num_quadrant1)* QUAD)";
my $re_weather_desc = '(?:MI|BC|PR|DR|BL|SH|TS|FZ)';
my $re_weather_ra = 'RA(?:DZ|SN|SG|PL|IC|GR|GS)*';
my $re_weather_dz = 'DZ(?:RA|SN|SG|PL|IC|GR|GS)*';
my $re_weather_sn = 'SN(?:RA|DZ|SG|PL|IC|GR|GS)*';
my $re_weather_ic = 'IC(?:RA|DZ|SN|SG|PL|GR|GS)*';
my $re_weather_prec = '(?:DZ|RA|SN|SG|PL|IC|GR|GS)+';
# EXTENSION: allow SHRADZ (ENHK)
my $re_weather_ts_sh = '(?:RA|SN|PL|GR|GS|DZ)+|UP';
my $re_weather_sg_pl_gr_gs = '(?:SG|PL|GR|GS)+';
# EXTENSION: allow SN, BC, PR, BR, PL with FZ
my $re_weather_fz = '(?:RA|DZ|FG|SN|BC|PR|BR|PL)+';
# WMO-No. 306 Vol I.1, Part A, Section C:
#   Intensity shall be indicated only with precipitation, precipitation
#   associated with showers and/or thunderstorms, duststorm or sandstorm.
# EXTENSION: allow +/- with BLSN, BR, TS, FC
# EXTENSION: allow JP (adjacent precipitation in old METAR)
my $re_weather_w_i = "(?:$re_weather_prec|DS|SS"
                      . "|(?:TS|SH)(?:$re_weather_ts_sh)"
                      . "|FZ(?:SH)?(?:$re_weather_fz)"
                      . "|UP|BLSN|BR|FC|TS|JP)";
my $re_weather_wo_i = "(?:$re_weather_ic|FG|BR|SA|DU|HZ|FU|VA|SQ|PO|FC|TS"
                       . "|(?:FZ|MI|BC|PR)*FG"
                       . "|BL(?:$re_weather_sn|SA|DU)"
                       . "|DR(?:$re_weather_sn|SA|DU))";
my $re_weather_vc = "(?:FG|PO|FC|DS|SS|TS(?:$re_weather_ts_sh)?|SH|BL(?:$re_weather_sn|SA|DU)|VA)";
# annex 3, form:
# FZDZ FZRA DZ   RA   SHRA SN SHSN SG SHGR SHGS BLSN SS DS
# TSRA TSSN TSPL TSGR TSGS FC VA   PL UP
# annex 3, text:
# FZDZ FZRA DZ   RA   SHRA SN SHSN SG SHGR SHGS BLSN SS DS
#                          FC VA   PL    TS
# EXTENSION: allow FZFG
my $re_weather_re = "(?:(?:FZ|SH)?$re_weather_prec"
                     . "|TS(?:$re_weather_ts_sh)"
                     . "|FZFG|BLSN|DS|SS|TS|FC|VA|UP)";
# EXTENSION: allow [+-]VC...
my $re_weather = "(?:[+-]?$re_weather_w_i|$re_weather_wo_i|[+-]?VC$re_weather_vc)";

my $re_cloud_cov = '(?:FEW|SCT|BKN|OVC)';
my $re_cloud_base = '(?:///|\d{3})';
my $re_cloud_type = '(?:AC(?:C|SL)?|AS|CB(?:MAM)?|CC(?:SL)?|CI|CS|CU(?:FRA)?|CF'
                     . '|NS|SAC|SC(?:SL)?|ST(?:FRA)?|SF|TCU|///)';

# EXTENSION: allow // for last 2 digits
# EXTENSION: allow tenths of hPa as .\d (FLLS, HUEN, OPRN)
# EXTENSION: allow QNH\d{4}INS
my $re_qnh = '(?:(?:Q[01]|A[23])\d(?:\d\d|//)|QNH[23]\d{3}INS|[AQ]////|Q[01]\d{3}\.\d)';

my $re_colour1 = 'BLU\+?|WHT|GRN|YLO[12]?|AMB|RED';
my $re_colour = "(?:BLACK|(?:BLACK)?(?:$re_colour1)(?: ?(?:$re_colour1|FCST CANCEL))?)";

my $re_be_weather_be   = "(?:[BE]$re_hour?$re_min)";
my $re_be_weather = "(?:(?:FZ|SH)?(?:TS|RA|SN|UP|PL|DZ|GS|FC)+$re_be_weather_be+)";

my $re_rsc_cond = '(?:DRY|WET|SANDED)';
my $re_rsc_deposit = '(?:LSR|SLR|PSR|IR|WR)+(?://|\d\d)?';
my $re_rsc = "(?:$re_rsc_deposit(?:P(?:[ /]$re_rsc_cond)?)?|(?:(?:P[ /])?$re_rsc_cond))";

my $re_snw_cvr_title = '(?:SNO?W? ?(?:CVR|COV(?:ER)?)?[/ ])';
my $re_snw_cvr_state =
       '(?:(?:ONE |MU?CH |TR(?:ACE)? ?)LOOSE|(?:MED(?:IUM)?|HARD) PACK(?:ED)?)';
my $re_snw_cvr = "(?:$re_snw_cvr_title?($re_snw_cvr_state)|${re_snw_cvr_title}NIL)";

my $re_temp = '(?:-?\d+\.\d)';
my $re_precip = '(?:\d{1,2}(?:\.\d)?(?: ?[MC]M)?)';

my $re_opacity_phenom =
     '(?:(?:BL)?SN|FG|DZ|IC|AS|ACC?|CC|CF|CI|CS|CU(?:FRA)?|TCU|NS|ST|SF|SC|CB)';
my $re_trace_cloud = '(?:AC|AS|CF|CI|CS|CU(?:FRA)?|SC|SF|TCU)';

my $re_phen_desc_when = '(?:OCNL|FRE?Q|INTMT|CON(?:TU)?S|PAST HR)';
my $re_phen_desc_how  = '(?:LOW?|LWR|(?:VRY |V|PR )?(?:THN|THIN|THK|THICK)|ISOL|CVCTV|DSIPTD|FREEZING|PTCHY|VERT|PTL)';
my $re_phen_desc_strength  = '(?:(?:VRY |V|PR )?(?:LGT|FBL)|MDT|MOD)';
my $re_phen_desc_where = '(?:ALOFT|ARND)';
my $re_phen_desc = "(?:$re_phen_desc_when|$re_phen_desc_how|$re_phen_desc_strength|$re_phen_desc_where)";
my $re_phen_desc_other = "(?:$re_phen_desc_when|$re_phen_desc_where)";

my $re_ltg_types = '(?:CA|CC|CG|CW|IC)';

my $re_wind_shear_lvl = "WS(\\d{3})/(${re_wind_dir}0${re_wind_speed}KT)";

my $re_phenomenon_other = "(?:LTG$re_ltg_types*|VIRGA|AURBO|AURORA|FG BNK|FULYR|HZY|BINOVC|ICG|SHS?|SHWR|DEW|(?:VIS|CIG|SC) (?:HYR|HIER|LWR|RDCD|REDUCED)|CLDS?(?: EMBD(?: 1ST LYR)?)?|(?:GRASS )?FIRES?|MTN TOPS OBSC|HALO(?: VI?SBL)?)";
my $re_phenomenon4 = "(?:($re_phenomenon_other|PCPN|VLY FG|HIR CLDS)|($re_cloud_type(?:[/-]$re_cloud_type)*)|($re_weather|SMOKE|HAZE|TSTMS)|($re_cloud_cov))";

my $re_loc_quadr3 = "(?:(?: TO)?($re_phen_locdirs)|(?: (DSNT))?($re_num_quadrant))";

my $re_estmd = '(?:EST(?:MD)?|ESTM?D|ESMTD|ESTIMATED)';
my $re_data_estmd = '(?:WI?NDS?(?: DATA)?|(?:CIG )?BLN|ALTM|CIG|CEILING|SLP|ALSTG|CLD HGTS)';

my $re_synop_tt = '(?:[0-6]\d|7[0-5])';
my $re_synop_period = '(?:[0-6]\d)';
my $re_synop_w1w1 = '(?:0[4-9]|1[01379]|2[0-8]|3[09]|4[1-9]|5[0-79]|6[0-7]|[778]\d|9[0-3])';

my $re_record_temp = '(?:(?:HI|LO)(?:[EX])(?:AT|FM|SE|SL|DA))';

########################################################################
# helper functions
########################################################################

sub _rnd {
    my ($val, $prec) = @_;
    return $prec * floor(($val / $prec) + 0.5);
}

sub _makeErrorMsg {
    my ($errorType, $msg, $curr_msg) = @_;
    my $s;

    $s = substr($msg, 0, (length $msg) - length $curr_msg) . '<@> ' . $curr_msg;
    $s =~ s/(<@> +[^ ]+) .+/$1 .../;
    return { errorType => $errorType, s => $s };
}

# $cy: [ 'ABCD', 'AB', 'A' ]
sub _cyInString {
    my ($cy, $str) = @_;

    return    index($str, ' ' . $cy->[0] . ' ') > -1
           || index($str, ' ' . $cy->[1] . ' ') > -1
           || index($str, ' ' . $cy->[2] . ' ') > -1;
}

# WMO-No. 306 Vol I.1, Part A, Section A, 15.6.3:
# Visibility shall be reported using the following reporting steps:
# (a) Up to 800 metres rounded down to the nearest 50 metres;
# (b) Between 800 and 5000 metres rounded down to the nearest 100 metres;
# (c) Between 5000 metres up to 9999 metres rounded down to the nearest 1000
#     metres;
# (d) With 9999 indicating 10 km and above.
#
# WMO-No. 306 Vol I.1, Part A, Section B:
# VVVV
#   Horizontal visibility at surface, in metres,
#   in increments of 50 metres up to 500 metres,             WRONG! CORRECT: 800
#   in increments of 100 metres between 500 and 5000 metres, and      WRONG! 800
#   in increments of 1000 metres between 5000 metres up to 9999 metres,
#   with 9999 indicating visibility of 10 km and above.
sub _getVisibilityM {
    my ($visM, $isEqualGreater) = @_;
    my $rp;

    return { v => 10, u => 'KM', q => 'isEqualGreater' }     if $visM == 9999;
    return { v => $visM + 0, u => 'M', q =>'isEqualGreater'} if $isEqualGreater;

    # add rp only if reported value is correctly rounded
    if ($visM =~ /0[0-7][05]0/) {
        $rp = 50;
    } elsif ($visM =~ /0\d00|[1-4]\d00/) {
        $rp = 100;
    } elsif ($visM =~ /\d000/) {
        $rp = 1000;
    } else {
        $rp = 0;
    }

    return { v => $visM + 0, u => 'M', $rp ? (rp => $rp) : () };
}

# WMO-No. 306 Vol I.1, Part A, Section B:
# VRVRVRVR
#   Runway visual range shall be reported
#   in steps of 25 metres when the runway visual range is less than 400 metres;
#   in steps of 50 metres when it is between 400 metres and 800 metres; and
#   in steps of 100 metres when the runway visual range is more than 800 metres.
sub _setVisibilityMRVROffset {
    my $vis = shift;

    # add rp only if reported value is correctly rounded
    if ($vis->{v} =~ /0[0-3][0257][05]/) {
        $vis->{rp} = 25;
    } elsif ($vis->{v} =~ /0[4-7][05]0/) {
        $vis->{rp} = 50;
    } elsif ($vis->{v} =~ /\d\d00/) {
        $vis->{rp} = 100;
    }
}

# FMH-1 12.6.7c
sub _setVisibilityFTRVROffset {
    my $vis = shift;

    # add rp only if reported value is correctly rounded
    if ($vis->{v} =~ /0\d00/) {
        $vis->{rp} = 100;
    } elsif ($vis->{v} =~ /[12][02468]00/) {
        $vis->{rp} = 200;
    } elsif ($vis->{v} =~ /[3-5][05]00|6000/) {
        $vis->{rp} = 500;
    }
}

sub _parseFraction {
    my ($val, $unit) = @_;
    my $q;

    $q = 'isLess'         if $val =~ s/^M//;
    $q = 'isEqualGreater' if $val =~ s/^P//;
    if ($val =~ m@^([1357])/([1-8]+)$@) {
        $val = $1 / $2;
    } elsif ($val =~ m@^(\d)(?: ?(\d*)/(\d*))?$@) {
        $val = $1;
        $val += $2 / $3 if $2;
    }
    return { v => $val, u => $unit, $q ? (q => $q) : () };
}

# FMH-1 table 6-1
sub _setVisibilitySMOffsetUS {
    my ($visSM, $is_auto) = @_;

    return if defined $visSM->{q};
    # TODO: add rp only if reported value is correctly rounded
    if ($is_auto) {
        if ($visSM->{v} < 2) {
            $visSM->{rp} = 1/4;
        } elsif ($visSM->{v} < 3) {
            $visSM->{rp} = 1/2;
        } else {
            $visSM->{rp} = 1;
        }
    } else {
        if ($visSM->{v} < 3/8) {
            $visSM->{rp} = 1/16;
        } elsif ($visSM->{v} < 2) {
            $visSM->{rp} = 1/8;
        } elsif ($visSM->{v} < 3) {
            $visSM->{rp} = 1/4;
        } elsif ($visSM->{v} < 15) {
            $visSM->{rp} = 1;
        } else {
            $visSM->{rp} = 5;
        }
    }
}

sub _getVisibilitySM {
    my ($vis, $is_auto, $cy) = @_;
    my $r;

    $r = _parseFraction $vis, 'SM';
    if (_cyInString $cy, ' cUS ') {
        _setVisibilitySMOffsetUS $r, $is_auto;
    } elsif (_cyInString $cy, ' cCA ') {
        if (!exists $r->{q}) {
            # http://ops.8wing.ca/downloads/WX_Decode_card.doc
            if ($r->{v} < 3/4) {
                $r->{rp} = 1/8;
            } elsif ($r->{v} < 2.5) {
                $r->{rp} = 1/4;
            } elsif ($r->{v} < 3) {
                $r->{rp} = 1/2;
            } elsif ($r->{v} < 15) {
                $r->{rp} = 1;
            } else {
                $r->{rp} = 5;
            }
        }
    } elsif (_cyInString $cy, ' EQ MM TK TN ') {
        # TODO: what about EQ(YR,YS) MM MY TKPK TNCM TX(KF)?
    }
    return { distance => $r };
}

sub _parseWeather {
    my ($weather, $mode) = @_;           # $mode: RE = recent, NI = no intensity
    my ($w, $int, $vc, $desc, $phen, $weather_str);

    if (defined $mode && $mode eq 'RE') {
        $weather_str = 'RE' . $weather;
    } else {
        $weather_str = $weather;
    }
    return { notAvailable => undef, s => $weather_str } if $weather eq '//';
    return { NSW => undef, s => $weather_str } if $weather eq 'NSW';

    return { tornado => undef, s => $weather_str } if $weather eq '+FC';

    # for $re_phenomenon4, only
    $weather = 'FU' if $weather eq 'SMOKE';
    $weather = 'HZ' if $weather eq 'HAZE';
    $weather = 'TS' if $weather eq 'TSTMS';

    ($int, $vc, $desc, $phen) =
                      $weather =~ m@([+-])?(VC)?($re_weather_desc+)?([A-Z/]+)@o;
    $w->{s} = $weather_str;
    if (defined $int) {
        $w->{phenomDescr} = ($int eq '-' ? 'isLight' : 'isHeavy');
    } else {
        # weather in vicinity _can_ have intensity, but it is an EXTENSION
        # BLSN, BR, TS _can_ have intensity, but it is an EXTENSION
        # (TODO: Canada uses +BLSN, +BLSA, +BLDU (MANOBS chapter 20))
        $w->{phenomDescr} = 'isModerate'
            unless    $vc
                   || $mode
                   || $weather =~ /^$re_weather_wo_i$/
                   || $weather =~ /^(?:BLSN|BR|TS|FC)$/;
    }
    $w->{inVicinity}    = undef if defined $vc;
    @{$w->{descriptor}} = $desc =~ /../g if defined $desc;
    @{$w->{phenomenon}} = $phen =~ /../g if defined $phen;
    return $w;
}

sub _parseOpacityPhenom {
    my ($r, $clds) = @_;

    for ($clds =~ /[A-Z]+\d/g) {
        my ($phenom, $oktas) = /(.*)(.)/;
        if ($phenom =~ /$re_weather/) {
            push @{$r->{opacityPhenomArr}}, { opacityWeather => {
                oktas   => $oktas,
                weather => _parseWeather $phenom, 'NI'
            }};
        } else {
            push @{$r->{opacityPhenomArr}}, { opacityCloud => {
                oktas     => $oktas,
                cloudType => $phenom
            }};
        }
    }
}

sub _parseRwyVis {
    my ($metar, $msg) = @_;
    my $r;

    if ($$msg =~ s@^(R($re_rwy_des)/(?:////|($re_rwy_vis)(?:V($re_rwy_vis))?(FT)?)(?:/)?([UDN])?) @@o)
    {
        my $v;

        $v->{s}        = $1;
        $v->{rwyDesig} = $2;
        $v->{visTrend} = $6 if defined $6;
        if (!defined $3) {
            $v->{RVR}{notAvailable} = undef;
        } else {
            $v->{RVR}{distance} = { v => $3, u => (defined $5 ? $5 : 'M') };
            if (defined $4) {
                $v->{RVRVariations}{distance} =
                                       { v => $4, u => $v->{RVR}{distance}{u} };
                if ($v->{RVRVariations}{distance}{v} =~ s/^M//) {
                    $v->{RVRVariations}{distance}{q} = 'isLess'
                } elsif ($v->{RVRVariations}{distance}{v} =~ s/^P//) {
                    $v->{RVRVariations}{distance}{q} = 'isEqualGreater'
                } elsif ($v->{RVRVariations}{distance}{u} eq 'M') {
                    _setVisibilityMRVROffset $v->{RVRVariations}{distance};
                } else {
                    _setVisibilityFTRVROffset $v->{RVRVariations}{distance};
                }
                $v->{RVRVariations}{distance}{v} += 0;
            }
            # corrections postponed because pattern matching changes $4
            if ($v->{RVR}{distance}{v} =~ s/^M//) {
                $v->{RVR}{distance}{q} = 'isLess';
            } elsif ($v->{RVR}{distance}{v} =~ s/^P//) {
                $v->{RVR}{distance}{q} = 'isEqualGreater';
            } elsif ($v->{RVR}{distance}{u} eq 'M') {
                _setVisibilityMRVROffset $v->{RVR}{distance};
            } else {
                _setVisibilityFTRVROffset $v->{RVR}{distance};
            }
            $v->{RVR}{distance}{v} += 0;
        }
        push @{$metar->{visRwy}}, $v;
        return 1;
    }
    return 0;
}

sub _parseRwyState {
    my ($metar, $msg) = @_;
    my $r;

    # (RRRR|RDRDR/?)ERCReReRBRBR
    # EXTENSION: allow missing /
    if ($$msg =~ s@^(((?:R/)?SNOCLO)|(88|99|$re_rwy_des2|R(?:88|99|$re_rwy_des)/?)(?:(CLRD)|([\d/])([\d/])(\d\d|//))(\d\d|//)) @@o)
    {
        $r->{s} = $1;
        if (defined $2) {
            $r->{SNOCLO} = undef;
        } else {
            if (defined $4) {
                $r->{cleared} = undef;
            } else {
                if ($5 eq '/') {
                    $r->{depositType}{notAvailable} = undef;
                } else {
                    $r->{depositType}{depositTypeVal} = $5;
                }
                if ($6 eq '/') {
                    $r->{depositExtent}{notAvailable} = undef;
                } elsif (   $6 eq '0' || $6 eq '1' || $6 eq '2'
                         || $6 eq '5' || $6 eq '9')
                {
                    $r->{depositExtent}{depositExtentVal} = $6;
                } else {
                    $r->{depositExtent}{invalidFormat} = $6;
                }
                if ($7 eq '//') {
                    $r->{depositDepth}{notAvailable} = undef;
                } elsif ($7 != 91) {
                    $r->{depositDepth}{depositDepthVal} = $7 + 0;
                } else {
                    $r->{depositDepth}{invalidFormat} = $7;
                }
            }

            if ($8 eq '//') {
                $r->{friction}{notAvailable} = undef;
            } elsif ($8 >= 1 && $8 <= 90) {
                $r->{friction}{coefficient} = $8 + 0;
            } elsif ($8 == 99) {
                $r->{friction}{unreliable} = undef;
            } elsif ($8 >= 91 && $8 <= 95) {
                $r->{friction}{brakingAction} = $8;
            } else {
                $r->{friction}{invalidFormat} = $8;
            }

            if ($3 eq '88' || $3 eq 'R88' || $3 eq 'R88/') {
                $r->{rwyDesigAll} = undef;
            } elsif ($3 eq '99' || $3 eq 'R99' || $3 eq 'R99/') {
                $r->{rwyDesigRep} = undef;
            } else {
                my $rwy_des = $3;

                if ($rwy_des =~ m@^R($re_rwy_des)/?$@) {
                    $r->{rwyDesig} = $1;
                } elsif ($rwy_des > 50) {
                    $r->{rwyDesig} = sprintf '%02dR', $rwy_des - 50;
                } else {
                    $r->{rwyDesig} = $rwy_des;
                }
            }
        }
        push @{$metar->{rwyState}}, $r;
        return 1;
    }
    return 0;
}

sub _parseWind {
    my ($wind, $dir_is_rounded) = @_;
    my ($w, $dir, $speed, $gustSpeed, $unitSpeed);

    return { notAvailable => undef } if $wind eq '/////';
    ($dir, $speed, $gustSpeed, $unitSpeed) = $wind =~ m@^$re_wind$@o;
    if ($dir eq '///' && $speed eq '//') {
        $w->{notAvailable} = undef;
    } elsif ($dir eq '000' && $speed eq '00' && !defined $gustSpeed) {
        $w->{isCalm} = undef;
    } else {
        my $isGreater;

        if ($dir eq '///') {
            $w->{dirNotAvailable} = undef;
        } elsif ($dir eq 'VRB') {
            $w->{dirVariable} = undef;
        } else {
            $w->{dir}{v} = $dir + 0; # true, not magnetic
            if ($dir_is_rounded && $dir =~ /0$/) {
                $w->{dir}{rp} = 4;
                $w->{dir}{rn} = 5;
            }
        }
        $unitSpeed =~ s/TK/KT/;
        $unitSpeed =~ s/KTS/KT/;
        $unitSpeed =~ s/ //;
        if ($speed eq '//') {
            $w->{speedNotAvailable} = undef;
        } else {
            $isGreater = $speed =~ s/^P//;
            $w->{speed} = { v => $speed + 0, u => $unitSpeed };
            $w->{speed}{q} = 'isGreater' if $isGreater;
        }
        if (defined $gustSpeed) {
            $isGreater = $gustSpeed =~ s/^P//;
            $w->{gustSpeed} = { v => $gustSpeed + 0, u => $unitSpeed };
            $w->{gustSpeed}{q} = 'isGreater' if $isGreater;
        }
    }
    return $w;
}

sub _parseWindAtLoc {
    my ($s, $location, $wind, $windVarLeft, $windVarRight) = @_;
    my $r;

    $r = _parseWind $wind;
    $r->{windVarLeft} = $windVarLeft + 0 if defined $windVarLeft;
    $r->{windVarRight} = $windVarRight + 0 if defined $windVarRight;
    return { windAtLoc => {
        s            => $s,
        windLocation => $location,
        wind         => $r
    }};
}

sub _parseCloud {
    my $cloud = shift;
    my $c;

    $c->{s} = $cloud;
    if ($cloud eq '//////') {
        $c->{notAvailable} = undef;
    } elsif ($cloud =~ m@^$re_cloud_cov$re_cloud_base(?: ?$re_cloud_type)?@o){
        $cloud =~ '(...)(?:(...) ?(.+)?)?';
        $c->{cloudCover} = $1;
        if (defined $2) {
            if ($2 eq '///') {
                $c->{baseBelowStation} = undef;
            } else {
                $c->{cloudBase} = { v => $2 * 100, u => 'FT' }; # AGL
            }
        }
        if (defined $3) {
            if ($3 eq '///') {
                $c->{cloudTypeNotAvailable} = undef;
            } else {
                $c->{cloudType} = $3;
            }
        }
    } else {
        $cloud =~ '(...)(.+)?';
        $c->{cloudCover} = $1;
        $c->{cloudType}  = $2 if defined $2;
    }
    return $c;
}

sub _parseQNH {
    my $qnh = shift;
    my ($q, $descr, $dig12, $dig34);

    $q->{s} = $qnh;
    ($descr, $dig12, $dig34) = $qnh =~ '([AQNH]*)(..)([^INS]*)';
    if ("$dig12$dig34" eq '////') {
        $q->{notAvailable} = undef;
    } else {
        $dig34 = '00' if $dig34 eq '//';
        if ($descr eq 'Q') {
            $q->{hPa} = "$dig12$dig34" + 0;
        } else {
            $q->{inHg} = sprintf "%.2f", "$dig12$dig34" / 100;
        }
    }
    return $q;
}

sub _parseColourCode {
    my $colour = shift;
    my $c;

    $colour =~ /^(BLACK)?($re_colour1)? ?(.+)?/o;
    $c->{s} = $colour;
    $c->{BLACK} = undef if defined $1;
    $c->{currentColour} =
            $2 eq 'BLU+' ? 'BLUplus' : ($2 eq 'FCST CANCEL' ? 'FCSTCANCEL' : $2)
        if defined $2;
    $c->{predictedColour} =
            $3 eq 'BLU+' ? 'BLUplus' : ($3 eq 'FCST CANCEL' ? 'FCSTCANCEL' : $3)
        if defined $3;
    return $c;
}

sub _parseLocations {
    my $loc_str = shift;
    my @loc_thru;

    return { isStationary => undef } if $loc_str =~ /^ *STNRY *$/;

    for ($loc_str =~ m@(?:[/ ]| AND )?($re_phen_locdir_comp_thru|UNKN)@og) {
        my (@loc, $is_grid);
        for ($_ =~ m@(?: ?[/-] ?| THRU )?(OV?HD|(?:OVE?R |AT |NR )?A(?:P|R?PT|D)|IN VLY|O?MTNS|OMTS|(?:OVE?R|ALG) (?:LK|RIVER|MT(?:N?S)?|VLYS?)|STNRY|UNKN|(?:DISTANT |DSNT |VC )?(?:OBSCG MTNS )?(?:ALQDS|V/D|$re_phen_locdir_comp))@og)
        {
            my $l;

            if ($_ =~ m@^(?:OV?HD|(?:OVE?R |AT |NR )?A(?:P|R?PT|D)|IN VLY|O?MTNS|OMTS|(?:OVR|ALG) (?:LK|RIVER|MT(?:N?S)?|VLYS?)|STNRY|UNKN)$@) {
                s/^OVR MT(?:N?S)?/OMTNS/;
                s/O?MTNS/OMTNS/;
                s/OMTS/OMTNS/;
                s/OV?HD/OHD/;
                s/OVER/OVR/;
                s/A(?:R?PT|D)$/AP/;
                s/^AP$/AT AP/;
                s/ /_/;
                $l->{locationSpec} = $_;
            } else {
                $_ =~ m@(?:(DISTANT|DSNT|VC) )?(OBSCG MTNS )?(?:TO ?)?(?:(?:(GRID )?(?:([1-9]\d*)(KM)? ?)?($re_compass_dir16)(?: (QUAD))?)|(ALQDS|V/D))@o;
                $l->{inDistance} = undef
                    if defined $1 && ($1 eq 'DISTANT' || $1 eq 'DSNT');
                $l->{inVicinity} = undef if defined $1 && $1 eq 'VC';
                $l->{obscgMtns}  = undef if defined $2;
                if (defined $3 || $is_grid) {
                    $is_grid = 1;
                    $l->{isGrid} = undef;
                }
                $l->{distance} = { v => $4, u => (defined $5 ? 'KM' : 'SM') }
                    if defined $4;
                $l->{compassDir} = $6 if defined $6;
                $l->{isQuadrant} = undef if defined $7;
                ($l->{locationSpec} = $8) =~ tr /\/ /_/ if defined $8;
            }
            push @loc, $l;
        }
        push @loc_thru, { location => \@loc };
    }
    return { locationThru => \@loc_thru };
}

sub _determineCeiling {
    my $cloud = shift;
    my ($ceil, $idx, $ii);

    $ceil = 100000;
    $idx = -1;
    $ii = -1;
    for (@$cloud) {
        $ii++;
        if (   exists $_->{cloudBase}
            && $_->{cloudBase}{u} eq 'FT'
            && $_->{cloudBase}{v} < $ceil
            && $_->{cloudBase}{v} < 200 * 100 # TODO: actually MSL!
            && ($_->{cloudCover} eq 'BKN' || $_->{cloudCover} eq 'OVC'))
        {
            $ceil = $_->{cloudBase}{v};
            $idx = $ii;
        }
    }
    $cloud->[$idx]{isCeiling} = undef if $idx > -1;
}

sub _parseQuadrants {
    my ($q, $inDistance) = @_;

    return { locationThru => { location => {
        $inDistance ? (inDistance => undef) : (),
        quadrant => [ $q =~ /([1-4])/g ]
    }}};
}

sub _parseTemp {
    my $temp = shift;

    $temp =~ /(.)(..)(.)/;
    return { v => (($1 eq '1' ? -1 : 1) * $2) . ($3 eq '/' ? '' : ".$3"),
             u => 'C'
           };
}

sub _parsePhenomDescr {
    my ($r, $tag, $phen_descr) = @_;

    for ($phen_descr =~ /$re_phen_desc|BBLO/og) {
        s/CONTUS/CONS/;
        s/V(LGT|THN|THIN|FBL|THK|THICK)/VRY $1/;
        s/THIN/THN/;
        s/FREQ/FRQ/;
        s/THICK/THK/;
        s/^LOW?/LOW/;
        s/^MOD/MDT/;

        push @{$r->{$tag}}, {
            FRQ      => 'isFrequent',
            OCNL     => 'isOccasional',
            INTMT    => 'isIntermittent',
            CONS     => 'isContinuous',
            THK      => 'isThick',
           'PR THK'  => 'isPrettyThick',
           'VRY THK' => 'isVeryThick',
            THN      => 'isThin',
           'PR THN'  => 'isPrettyThin',
           'VRY THN' => 'isVeryThin',
            LGT      => 'isLight',
           'PR LGT'  => 'isPrettyLight',
           'VRY LGT' => 'isVeryLight',
            FBL      => 'isFeeble',
           'PR FBL'  => 'isPrettyFeeble',
           'VRY FBL' => 'isVeryFeeble',
            MDT      => 'isModerate',
            LOW      => 'isLow',
            LWR      => 'isLower',
            ISOL     => 'isIsolated',
            CVCTV    => 'isConvective',
            DSIPTD   => 'isDissipated',
           'PAST HR' => 'inPastHour',
            BBLO     => 'baseBelowStation',
            ALOFT    => 'isAloft',
            ARND     => 'isAround',
            FREEZING => 'isFreezing',
            PTCHY    => 'isPatchy',
            VERT     => 'isVertical',
            PTL      => 'isPartial',
        }->{$_};
    }
}

sub _parsePhenom {
    my ($r, $phenom) = @_;

    if ($phenom =~ /LTG$re_ltg_types/o) {
        $$r->{lightningType} = ();
        for ($phenom =~ /$re_ltg_types/og) {
            push @{$$r->{lightningType}}, $_;
        }
    } elsif ($phenom eq 'VIS HIER') {
        $$r->{otherPhenom} = 'VIS_HYR';
    } else {
        ($$r->{otherPhenom} = $phenom) =~ tr/ /_/;
        $$r->{otherPhenom} =~ s/REDUCED/RDCD/;
        $$r->{otherPhenom} =~ s/AURORA/AURBO/;
        $$r->{otherPhenom} =~ s/SHWR/SH/;
        $$r->{otherPhenom} =~ s/_VI?SBL//;
    }
}

# "supplementary" section of TAFs
sub _parseTAFsuppl {
    my ($metar, $msg, $base_metar, $cy) = @_;
    my $r;

    if ($$msg =~ s@^(([56])(\d)(\d{3})(\d)) @@) {
        my ($ti, $type);

        $type = $2 eq '5' ? 'turbulence' : 'icing';
        $r = {
            s               => $1,
            $type . 'Descr' => $3,
            layerBase       => $4 + 0,
        };
        $ti = $5;

        # if item repeated: level is layer top
        if ($$msg =~ s@^($2$3(\d{3})$5) @@) {
            $r->{s} .= ' ' . $1;
            $r->{layerTop} = $2 + 0;
        } else {
            $r->{layerThickness} = $ti;
        }
        push @{$metar->{TAFsupplArr}}, { $type => $r };
        return 1;
    }

    if ($$msg =~ s@^($re_wind_shear_lvl) @@o) {
        push @{$metar->{TAFsupplArr}}, { windShearLvl => {
            s     => $1,
            level => $2 + 0,
            wind  => _parseWind $3
        }};
        return 1;
    }

    if ($$msg =~ s@^($re_qnh) @@o) {
        push @{$metar->{TAFsupplArr}}, { QNH => _parseQNH $1 };
        return 1;
    }

    if (   _cyInString($cy, ' cUS NZ ')
        && $$msg =~ s@^((FU|(?:FZ)?FG|BR|SN|PWR PLNT(?: PLUME)?) ($re_cloud_cov$re_cloud_base)) @@o)
    {
        my ($cl, $phen);

        $r->{s} = $1;
        $r->{cloud} = _parseCloud $3;
        $phen = $2;
        if ($phen =~ m@^$re_weather$@o) {
            $r->{weather} = _parseWeather $phen, 'NI';
        } else {
            ($r->{cloudPhenom} = $phen) =~ s/ /_/g;
        }
        push @{$metar->{TAFsupplArr}}, { obscuration => $r };
    }

    return 0;
}

# additional TAF info
sub _parseTAFinfo {
    my ($metar, $msg, $is_taf_Aug2007) = @_;
    my $r;

    if (  (!$is_taf_Aug2007 && $$msg =~ s@^(T([XN])(M)?(\d\d)/()($re_hour|24)Z) @@o)
        || ($is_taf_Aug2007 && $$msg =~ s@^(T([XN])(M)?(\d\d)/($re_day)($re_hour)Z) @@o))
    {
        $r->{s} = $1;
        $r->{temp} = { v => $4 + 0, u => 'C' };
        $r->{temp}{v} *= -1 if defined $3;
        $r->{day} = $5 if $5;
        $r->{hour} = $6;
        push @{$metar->{TAFinfoArr}},
                              { ($2 eq 'N' ? 'tempMinAt' : 'tempMaxAt') => $r };
        return 1;
    }

    if ($$msg =~ s@^(T(M)?(\d\d)/($re_hour)Z) @@o) {
        $r->{s} = $1;
        $r->{temp} = { v => $3 + 0, u => 'C' };
        $r->{temp}{v} *= -1 if defined $2;
        $r->{hour} = $4;
        push @{$metar->{TAFinfoArr}}, { tempAt => $r };
        return 1;
    }

    if ($$msg =~ s@^((COR|AMD) ($re_hour)($re_min)Z?) @@o) {
        push @{$metar->{TAFinfoArr}},
            { $2 eq 'COR' ? 'correctedAt' : 'amendedAt' => {
                s      => $1,
                hour   => $3,
                minute => $4
        }};
        return 1;
    }

    if ($$msg =~ s@^(LIMITED METWATCH ($re_day)($re_hour) TIL ($re_day)($re_hour)) @@o)
    {
        push @{$metar->{TAFinfoArr}}, { limMetwatch => {
            s         => $1,
            timeFrom => { day => $2, hour => $3 },
            timeTill => { day => $4, hour => $5 },
        }};
        return 1;
    }

    if ($$msg =~ s@^(AUTOMATED SENSOR(?:ED)? METWATCH(?: ($re_day)($re_hour) TIL ($re_day)($re_hour))?) @@o)
    {
        $r->{s} = $1;
        if (defined $2) {
            $r->{timeFrom} = { day => $2, hour => $3 };
            $r->{timeTill} = { day => $4, hour => $5 };
        }
        push @{$metar->{TAFinfoArr}}, { autoMetwatch => $r };
        return 1;
    }

    if ($$msg =~ s@^(AMD (?:LTD TO CLD VIS AND WIND(?: (?:TIL ($re_hour|24)Z|(${re_hour})Z-($re_hour|24)Z))?|(NOT SKED))) @@o)
    {
        $r->{s} = $1;
        if (defined $5) {
            $r->{isNotScheduled} = undef;
        } else {
            $r->{isLtdToCldVisWind} = undef;
            $r->{hourTill} = $2 if defined $2;
            $r->{hourFrom} = $3 if defined $3;
            $r->{hourTill} = $4 if defined $4;
        }
        push @{$metar->{TAFinfoArr}}, { amendment => $r };
        return 1;
    }

    return 0;
}

sub _getHumidity {
    my ($t, $d) = @_;

    return (
        # FGFS metar
        _rnd(100 * 10 ** (7.5 * ($d / ($d + 237.7) - $t / ($t + 237.7))), 0.01),

        # http://www.bragg.army.mil/www-wx/wxcalc.htm
        # http://www.srh.noaa.gov/bmx/tables/rh.html
        _rnd(100 * ((112 - (0.1 * $t) + $d) / (112 + (0.9 * $t))) ** 8, 0.01),

        # http://www.mattsscripts.co.uk/mweather.htm
        # http://ingrid.ldeo.columbia.edu/dochelp/QA/Basic/dewpoint.html
        _rnd(
          100 * (6.11 * exp(5417.118093 * (1 / 273.15 - (1 / ($d + 273.15)))))
              / (6.11 * exp(5417.118093 * (1 / 273.15 - (1 / ($t + 273.15))))),
          0.01),

        # http://de.wikipedia.org/wiki/Taupunkt
        _rnd(100 * (6.11213 * exp(17.5043 * $d / (241.2 + $d)))
                 / (6.11213 * exp(17.5043 * $t / (241.2 + $t))), 0.01)
        );
}

# WMO-No. 306 Vol I.1, Part A, Section D.a:
sub _IIiii2region {
    my $IIiii = shift;

    my %IIiii_region = (
         I        => [ [ 60001, 69998 ] ],
         II       => [ [ 20001, 20099 ], [ 20200, 21998 ], [ 23001, 25998 ],
                       [ 28001, 32998 ], [ 35001, 36998 ], [ 38001, 39998 ],
                       [ 40350, 48599 ], [ 48800, 49998 ], [ 50001, 59998 ] ],
         III      => [ [ 80001, 88998 ] ],
         IV       => [ [ 70001, 79998 ] ],
         V        => [ [ 48600, 48799 ], [ 90001, 98998 ] ],
         VI       => [ [ 00001, 19998 ], [ 20100, 20199 ], [ 22001, 22998 ],
                       [ 26001, 27998 ], [ 33001, 34998 ], [ 37001, 37998 ],
                       [ 40001, 40349 ] ],
        Antarctic => [ [ 89001, 89998 ] ],
    );
    for my $region (keys %IIiii_region) {
        for (@{$IIiii_region{$region}}) {
            return $region if $IIiii >= $_->[0] && $IIiii <= $_->[1];
        }
    }
    return '';
}

# determine the country that operates a station to select correct decoding rules
sub _IIiii2country {
    my $IIiii = shift;

    # ISO-3166 coded
    my %IIiii_country2 = (
        CH => [ [  6600,  6799 ] ],                  # Switzerland
        FR => [ [  7000,  7999 ] ],                  # France
        AT => [ [ 11000, 11399 ] ],                  # Austria
        BD => [ [ 41850, 41999 ] ],                  # Bangladesh
        US => [ [ 70001, 70998 ],                    # USA (US-AK)
                [ 72001, 72998 ],
                [ 74001, 74998 ],
                [ 91066, 91066 ],                    # (UM-71)
                [ 91101, 91199 ], [ 91280, 91299 ],  # (US-HI)
                [ 91210, 91219 ],                    # (GU)
                [ 91220, 91239 ],                    # (MP)
                [ 91250, 91259 ], [ 91365, 91379 ],  # (MH)
                [ 91764, 91768 ],                    # (AS)
                [ 91901, 91901 ],                    # (UM-86)
                [ 91902, 91903 ],                    # (KI-L)
              ],
        CA => [ [ 71001, 71998 ] ],                  # Canada
    );
    my %IIiii_country = (
        US => [ 61902,                       # USA (SH-AC)
                61967,                       # (IO)
                78367,                       # (CU-14)
                91245,                       # (UM-79)
                91275,                       # (UM-67)
                91334, 91348, 91356, 91413,  # (FM)
                91442,                       # (MH)
                89009, 89049, 89061,
                89083, 89175, 89528, 89598, 89627, 89628, 89637, 89664, 89674,
                89108, 89208, 89257, 89261, 89262, 89264, 89266, 89269, 89272,
                89314, 89324, 89327, 89332, 89345, 89371, 89376, 89377, 89643,
                89667, 89734, 89744, 89768, 89769, 89799, 89828, 89832, 89834,
                89847, 89864, 89865, 89866, 89867, 89868, 89869, 89872, 89873,
                89879,
              ],
    );
    for my $country (keys %IIiii_country2) {
        for (@{$IIiii_country2{$country}}) {
            return $country if $IIiii >= $_->[0] && $IIiii <= $_->[1];
        }
    }
    for my $country (keys %IIiii_country) {
        for (@{$IIiii_country{$country}}) {
            return $country if $IIiii == $_;
        }
    }
    return '';
}

# a3 standard isobaric surface for which the geopotential is reported
sub _codeTable0264 {
    my $idx = shift;

    return {
        1 => '1000',
        2 =>  '925',
        5 =>  '500',
        7 =>  '700',
        8 =>  '850'
    }->{$idx};
}

# C  genus of cloud
# C  genus of cloud predominating in the layer
# C' genus of cloud whose base is below the level of the station
sub _codeTable0500 {
    my ($r, $idx) = @_;

    if ($idx eq '/') {
        $r->{cloudTypeNotAvailable} = undef;
    } else {
        $r->{cloudType} = ( 'CI', 'CC', 'CS', 'AC', 'AS',
                            'NS', 'SC', 'ST', 'CU', 'CB' )[$idx];
    }
}

# WMO-No. 306 Vol I.1, Part A, code table 0700:
# D  true direction from which surface wind is blowing
# D  true direction towards which ice has drifted in the past 12 hours
# DH true direction from which CH clouds are moving
# DK true direction from which swell is moving
# DL true direction from which CL clouds are moving
# DM true direction from which CM clouds are moving
# Da true direction in which orographic clouds or clouds with vertical development are seen
# Da true direction in which the phenomenon indicated is observed or in which conditions specified in the same group are reported
# De true direction towards which an echo pattern is moving
# Dp true direction from which the phenomenon indicated is coming
# Ds true direction of resultant displacement of the ship during the three hours preceding the time of observation
# D1 true direction of the point position from the station
sub _codeTable0700 {
    my ($r, $type, $idx) = @_;

    if ($idx eq '/') {
        $r->{"${type}NA"} = undef;
    } elsif ($idx == 0) {
        $r->{"${type}None"} = undef;
    } elsif ($idx == 9) {
        $r->{"${type}Invisible"} = undef;
    } else {
        $r->{"${type}Dir"} = {
            1 => 'NE',
            2 => 'E',
            3 => 'SE',
            4 => 'S',
            5 => 'SW',
            6 => 'W',
            7 => 'NW',
            8 => 'N',
        }->{$idx};
    }
}

# dc duration and character of precipitation given by RRR
sub _codeTable0833 {
    my $idx = shift;

    return { hoursTill => { v => 1, q => 'isLess' }} if $idx == 0 || $idx == 4;
    return { hoursFrom => 1, hoursTill => 3 }        if $idx == 1 || $idx == 5;
    return { hoursFrom => 3, hoursTill => 6 }        if $idx == 2 || $idx == 6;
    return { hoursFrom => { v => 6, q => 'isGreater' }};
}

# eC elevation angle of the top of the cloud indicated by C
# e' elevation angle of the top of the phenomenon above horizon
sub _codeTable1004 {
    my ($r, $idx) = @_;

    if ($idx == 0) {
        $r->{topsInvisible} = undef;
    } elsif ($idx == 1) {
        $r->{elevationAngle}{v} = 45;
        $r->{elevationAngle}{q} = 'isEqualGreater';
    } elsif ($idx == 9) {
        $r->{elevationAngle}{v} = 5;
        $r->{elevationAngle}{q} = 'isLess';
    } else {
        $r->{elevationAngle} = {
            2 => 30,
            3 => 20,
            4 => 15,
            5 => 12,
            6 => 9,
            7 => 7,
            8 => 6,
        }->{$idx};
    }
}

# h height above surface of the base of the lowest cloud seen
sub _codeTable1600 {
    my $idx = shift;
    my @v;

    return { notAvailable => undef } if $idx eq '/';
    return { from => { v => 2500, u => 'M', q => 'isEqualGreater' }}
        if $idx == 9;
    @v = (0, 50, 100, 200, 300, 600, 1000, 1500, 2000, 2500)[$idx, $idx + 1];
    return { from => { v => $v[0], u => 'M' },
             to   => { v => $v[1], u => 'M' }};
}

# FMH-2 table 4-3
sub _codeTable1600US {
    my $idx = shift;
    my $v;

    return { notAvailable => undef } if $idx eq '/';
    return { from => { v => 8500, u => 'FT', q => 'isEqualGreater' }}
        if $idx == 9;
    $v = ([    0,  100 ],
          [  200,  300 ],
          [  400,  600 ],
          [  700,  900 ],
          [ 1000, 1900 ],
          [ 2000, 3200 ],
          [ 3300, 4900 ],
          [ 5000, 6500 ],
          [ 7000, 8000 ])[$idx];

    return { from => { v => $v->[0], u => 'FT' },
             to   => { v => $v->[1], u => 'FT' }};
}

# hshs height of base of cloud layer or mass whose genus is indicated by C
# htht height of the tops of the lowest clouds or height of the lowest cloud layer or fog
sub _codeTable1677 {
    my $idx = shift;

    return { v => 30,    q => 'isLess'         } if $idx == 0;
    return { v => 30 * $idx                    } if $idx <= 50;
    return { v => 300 * ($idx - 50)            } if $idx <= 80;
    return { v => 1500 * ($idx - 74)           } if $idx <= 88;
    return { v => 21000, q => 'isGreater'      } if $idx == 89;
    return { v => 50,    q => 'isLess'         } if $idx == 90;
    return { v => 2500,  q => 'isEqualGreater' } if $idx == 99;
    return [(50, 100, 200, 300, 600, 1000, 1500, 2000, 2500)
                                                        [$idx - 91, $idx - 90]];
}

# N  total cloud cover
# Nh amount of all the CL cloud present or, if no CL cloud is present, the amount of all the CM cloud present
# Ns amount of individual cloud layer or mass whose genus is indicated by C
# N' amount of cloud whose base is below the level of the station
sub _codeTable2700 {
    my $idx = shift;

    return { oktasNotAvailable => undef } if $idx eq '/';
    return { skyObscured       => undef } if $idx eq '9';
    return { oktas             => $idx  };
}

# Rt time at which precipitation given by RRR began or ended
sub _codeTable3552 {
    my $idx = shift;

    return { hoursTill => { v => 1, q => 'isLess' }}     if $idx == 1;
    return { hoursFrom => $idx - 1, hoursTill => $idx }  if $idx <= 6;
    return { hoursFrom => 6, hoursTill => 12 }           if $idx == 7;
    return { hoursFrom => { v => 12, q => 'isGreater' }} if $idx == 8;
    return { notAvailable => undef }                     if $idx == 9;
}

# RR amount of precipitation or water equivalent of solid precipitation, or diameter of solid deposit
sub _codeTable3570 {
    my ($r, $idx, $tag) = @_;

    if ($idx <= 55) {
        $$r->{$tag} = { v => $idx + 0, u => 'MM' };
    } elsif ($idx <= 90) {
        $$r->{$tag} = { v => ($idx - 50) * 10, u => 'MM' };
    } elsif ($idx <= 96) {
        $$r->{$tag} = { v => ($idx - 90) / 10, u => 'MM' };
    } elsif ($idx == 97) {
        $$r->{traces} = undef;
    } elsif ($idx == 98) {
        $$r->{$tag} = { v => 400, u => 'MM', q => 'isGreater' };
    } else {
        $$r->{notAvailable} = undef;
    }
}

# RRR amount of precipitation which has fallen during the period preceding the time of observation, as indicated by tR
sub _codeTable3590 {
    my $idx = shift;

    return { notAvailable => undef }                       if $idx eq '///';

    return { precipAmount => { v => $idx + 0, u => 'MM' }} if $idx <= 988;
    return { precipAmount => { v => 989,      u => 'MM',
                               q => 'isEqualGreater' }}    if $idx == 989;
    return { traces => undef }                             if $idx == 990;
    return { precipAmount => { v => ($idx - 990) / 10, u => 'MM' }};
}

# ss depth of newly fallen snow
sub _codeTable3870 {
    my ($r, $idx) = @_;

    if ($idx <= 55) {
        $$r->{precipAmount} = { v => $idx * 10, u => 'MM' };
    } elsif ($idx <= 90) {
        $$r->{precipAmount} = { v => ($idx - 50) * 100, u => 'MM' };
    } elsif ($idx <= 96) {
        $$r->{precipAmount} = { v => $idx + 0, u => 'MM' };
    } elsif ($idx == 97) {
        $$r->{precipAmount} = { v => 1, q => 'isLess', u => 'MM' };
    } elsif ($idx == 98) {
        $$r->{precipAmount} = { v => 4000, q => 'isGreater', u => 'MM' };
    } else {
        $$r->{noMeasurement} = undef;
    }
}

# tR duration of period of reference for amount of precipitation, ending at the time of the report
sub _codeTable4019 {
    my $idx = shift;
    return ('6h', '12h', '18h', '24h', '1h', '2h', '3h', '9h', '15h')[$idx - 1];
}

# tt time before observation or duration of phenomena (00-69)
# zz variation, location or intensity of phenomena (76-99)
# for both: 70-75
sub _codeTable4077 {
    my ($idx, $when) = @_;

    return {} unless defined $idx;
    return { $when => { hoursFrom => $idx / 10 }}
        if $idx <= 60;
    return { $when => { hoursFrom => $idx - 55, hoursTill => $idx - 54}}
        if $idx <= 66;
    return { $when => { hoursFrom => 12, hoursTill => 18 }}
        if $idx == 67;
    return { $when => { hoursFrom => { v => 18, q => 'isGreater' }}}
        if $idx == 68;
    return { $when => { hoursUnknown => undef }}
        if $idx == 69;
    return { phenomVariation =>
             ('beganDuringObs',      'endedDuringObs',
              'beganEndedDuringObs', 'changedDuringObs',
              'beganAfterObs',       'endedAfterObs' )[$idx - 70]
    } if $idx <= 75;
    return { phenomDescr =>
        ('atStation',                'atStationNotInDistance',
         'allDirections',            'allDirectionsNotAtStation',
         'approchingStation',        'recedingFromStation',
         'passingStationInDistance', 'seenInDistance',
         'inVicinityNotAtStation',   'aloftNotNearGround',
         'nearGroundNotAloft',       'isOccasional',
         'isIntermittent',           'isFrequent',
         'isSteady',                 'isIncreasing',
         'isDecreasing',             'isVariable',
         'isContinuous',             'isVeryLight',
         'isLight',                  'isModerate',
         'isHeavy',                  'isVeryHeavy')[$idx - 76]
    };
}

# VV   horizontal visibility at surface
# VsVs visibility towards the sea
# WMO-No. 306 Vol I.1, Part A, Section A, 12.2.1.3.2:
#   In reporting visibility at sea, the decile 90-99 shall be used for VV.
# WMO-No. 488:
#   3.2.2.3 Observations at sea stations
#   3.2.2.3.4 Visibility
#   The requirements ... are ... low, ... decade 90-99 of code table 4377 ...
sub _codeTable4377 {
    my ($r, $idx, $station_type) = @_;
    my (%vis, $vis_type);

    if ($idx eq '//') {
        $r->{visPrev}{s} = $idx;
        $r->{visPrev}{notAvailable} = undef;
        return;
    }
    if ($idx >= 51 && $idx <= 55) {
        $r->{visPrev}{s} = $idx;
        $r->{visPrev}{invalidFormat} = $idx;
        return;
    }

    $vis_type = 'visPrev';
    $vis{distance}{u} = 'KM';
    if ($idx == 0) {
        $vis{distance}{v} = 100;
        $vis{distance}{u} = 'M';
        $vis{distance}{q} = 'isLess';
    } elsif ($idx <= 50) {
        $vis{distance}{v} = $idx * 100;
        $vis{distance}{rp} = 100;
        $vis{distance}{u} = 'M';
    } elsif ($idx <= 80) {
        $vis{distance}{v} = $idx - 50;
        $vis{distance}{rp} = 1;
    } elsif ($idx <= 88) {
        $vis{distance}{v} = 5 * ($idx - 74);
        $vis{distance}{rp} = 4;
    } elsif ($idx == 89) {
        $vis{distance}{v} = 70;
        $vis{distance}{q} = 'isGreater';
    } elsif ($idx == 90) {
        $vis_type = 'visibilityAtLoc';
        $vis{distance}{v} = 50;
        $vis{distance}{u} = 'M';
        $vis{distance}{q} = 'isLess';
    } elsif ($idx == 99) {
        $vis_type = 'visibilityAtLoc';
        $vis{distance}{v} = 50;
        $vis{distance}{q} = 'isEqualGreater';
    } else {
        $vis_type = 'visibilityAtLoc';
        @{$vis{distance}}{'v', 'rp', 'u'} = @{(
            [   50, 150,  'M' ],   # 91
            [  200, 300,  'M' ],   # 92
            [  500, 500,  'M' ],   # 93
            [    1,   1, 'KM' ],   # 94
            [    2,   2, 'KM' ],   # 95
            [    4,   6, 'KM' ],   # 96
            [   10,  10, 'KM' ],   # 97
            [   20,  30, 'KM' ],   # 98
        )[$idx - 91]};
    }
    if ($vis_type eq 'visPrev' || $station_type =~ /^[AO]/) {
        $r->{visPrev} = \%vis;
        $r->{visPrev}{s} = $idx;
    } else {
        $r->{visibilityAtLoc}{visibility} = \%vis;
        $r->{visibilityAtLoc}{locationAt} = 'MAR';
        $r->{visibilityAtLoc}{s} = $idx;
    }
}

sub _codeTable4377US {
    my ($r, $idx, $station_type, $is_auto) = @_;
    my (%vis, $vis_type);

    if ($idx eq '//') {
        $r->{visPrev}{s} = $idx;
        $r->{visPrev}{notAvailable} = undef;
        return;
    }

    $vis_type = 'visPrev';
    $vis{distance}{u} = 'SM';
    if ($idx == 0) {
        $vis{distance}{v} = 1/16;
        $vis{distance}{q} = 'isLess';
    } elsif ($idx =~ /^(?:0[1-68]|[12][02468]|3[026]|4[048])$/) {
        $vis{distance}{v} = $idx / 16;
    } elsif ($idx =~ /^(?:5[68]|6[0134689]|7[134]|8[02457])$/) {
        $vis{distance}{v} = {
            56 =>  4, 58 =>  5,
            60 =>  6, 61 =>  7, 63 =>  8, 64 => 9, 66 => 10, 68 => 11, 69 => 12,
            71 => 13, 73 => 14, 74 => 15,
            80 => 20, 82 => 25, 84 => 30, 85 => 35, 87 => 40,
        }->{$idx};
    } elsif ($idx == 89) {
        $vis{distance}{v} = 45;
        $vis{distance}{q} = 'isEqualGreater';
    } elsif ($idx == 90) {
        $vis_type = 'visibilityAtLoc';
        $vis{distance}{v} = 1/16;
        $vis{distance}{q} = 'isLess';
        $vis{distance}{u} = 'NM';
    } elsif ($idx == 99) {
        $vis_type = 'visibilityAtLoc';
        $vis{distance}{v} = 10;
        $vis{distance}{q} = 'isGreater';
        $vis{distance}{u} = 'NM';
    } elsif ($idx > 90) {
        $vis_type = 'visibilityAtLoc';
        # FMH-2 table 4-5
        @{$vis{distance}}{'v', 'rp'} = @{(
           [ 1/16, 1/16 ],   # 91
           [ 1/8,  1/8  ],   # 92
           [ 1/4,  1/4  ],   # 93
           [ 1/2,  1/2  ],   # 94
           [ 1,    1    ],   # 95
           [ 2,    5    ],   # 96
           [ 5,    9    ],   # 97
           [ 9,    11   ],   # 98
        )[$idx - 91]};
        $vis{distance}{u} = 'NM';
    } else {
        $r->{visPrev}{s} = $idx;
        $r->{visPrev}{invalidFormat} = $idx;
        return;
    }
    _setVisibilitySMOffsetUS $vis{distance}, $is_auto
        if $vis{distance}{u} eq 'SM';
    if ($vis_type eq 'visPrev' || $station_type =~ /^[AO]/) {
        $r->{visPrev} = \%vis;
        $r->{visPrev}{s} = $idx;
    } else {
        $r->{visibilityAtLoc}{visibility} = \%vis;
        $r->{visibilityAtLoc}{locationAt} = 'MAR';
        $r->{visibilityAtLoc}{s} = $idx;
    }
}

# MANOBS 12.3.1.4
sub _codeTable4377CA {
    my ($r, $idx) = @_;
    my (%vis, $vis_type);

    if ($idx eq '//') {
        $r->{visPrev}{s} = $idx;
        $r->{visPrev}{notAvailable} = undef;
        return;
    }

    $vis_type = 'visPrev';
    $vis{distance}{u} = 'SM';
    if ($idx =~ /^(?:0[02468]|1[026]|2[048]|3[26]|4[08])$/) {
        $vis{distance}{v} = $idx / 16;
    } elsif ($idx =~ /^(?:5[689]|6[124679]|7[024]|8[0-8])$/) {
        $vis{distance}{v} = {
            56 =>  4, 58 =>  5, 59 =>  6,
            61 =>  7, 62 =>  8, 64 =>  9, 66 => 10, 67 => 11, 69 => 12,
            70 => 13, 72 => 14, 74 => 15,
            80 => 19, 81 => 22, 82 => 25, 83 => 28, 84 => 32, 85 => 35,
                      86 => 38, 87 => 41, 88 => 44,
        }->{$idx};
    } elsif ($idx == 89) {
        $vis{distance}{v} = 44;
        $vis{distance}{q} = 'isGreater';
    } else {
        $r->{visPrev}{s} = $idx;
        $r->{visPrev}{invalidFormat} = $idx;
        return;
    }

    if (!exists $vis{distance}{q}) {
        if ($vis{distance}{v} < 3/4) {
            $vis{distance}{rp} = 1/8;
        } elsif ($vis{distance}{v} < 2.5) {
            $vis{distance}{rp} = 1/4;
        } elsif ($vis{distance}{v} < 3) {
            $vis{distance}{rp} = 1/2;
        } elsif ($vis{distance}{v} < 15) {
            $vis{distance}{rp} = 1;
        } elsif ($vis{distance}{v} == 15 || $vis{distance}{v} == 28) {
            $vis{distance}{rp} = 4;
        } elsif ($vis{distance}{v} != 44) {
            $vis{distance}{rp} = 3;
        }
    }

    $r->{visPrev} = \%vis;
    $r->{visPrev}{s} = $idx;
}

# w1w1 Present weather phenomenon not specified in Code table 4677, or
#      specification of present weather phenomenon in addition to group 7wwW1W2
sub _codeTable4687 {
    my $idx = shift;
    my $r;

    return { weatherPresent1 => $idx }
        unless (   ($idx >= 47 && $idx <= 57)
                || ($idx >= 60 && $idx <= 67)
                || ($idx >= 70 && $idx <= 77));

    $r = ( { weatherSynopFG => { # 47
             visVertFrom => { distance => { v => 60, u => 'M' }},
             visVertTo   => { distance => { v => 90, u => 'M' }}}},
         { weatherSynopFG => { # 48
             visVertFrom => { distance => { v => 30, u => 'M' }},
             visVertTo   => { distance => { v => 60, u => 'M' }}}},
         { weatherSynopFG => { # 49
             visVert     => { distance => { v => 30, u => 'M',
                                            q => 'isLess' }}}},
         { weatherSynopPrecip => { # 50
             rateOfFall     => { v => '0.10', u => 'MMH', q => 'isLess' }}},
         { weatherSynopPrecip => { # 51
             rateOfFallFrom => { v => '0.10', u => 'MMH' },
             rateOfFallTo   => { v => 0.19, u => 'MMH' }}},
         { weatherSynopPrecip => { # 52
             rateOfFallFrom => { v => '0.20', u => 'MMH' },
             rateOfFallTo   => { v => 0.39, u => 'MMH' }}},
         { weatherSynopPrecip => { # 53
             rateOfFallFrom => { v => '0.40', u => 'MMH' },
             rateOfFallTo   => { v => 0.79, u => 'MMH' }}},
         { weatherSynopPrecip => { # 54
             rateOfFallFrom => { v => '0.80', u => 'MMH' },
             rateOfFallTo   => { v => 1.59, u => 'MMH' }}},
         { weatherSynopPrecip => { # 55
             rateOfFallFrom => { v => '1.60', u => 'MMH' },
             rateOfFallTo   => { v => 3.19, u => 'MMH' }}},
         { weatherSynopPrecip => { # 56
             rateOfFallFrom => { v => '3.20', u => 'MMH' },
             rateOfFallTo   => { v => 6.39, u => 'MMH' }}},
         { weatherSynopPrecip => { # 57
             rateOfFall     => { v => 6.4, u => 'MMH',
                                 q => 'isEqualGreater' }}},
         { }, # 58
         { }, # 59
         { weatherSynopPrecip => { # 60
             rateOfFall     => { v => '1.0', u => 'MMH', q => 'isLess' }}},
         { weatherSynopPrecip => { # 61
             rateOfFallFrom => { v => '1.0', u => 'MMH' },
             rateOfFallTo   => { v => 1.9, u => 'MMH' }}},
         { weatherSynopPrecip => { # 62
             rateOfFallFrom => { v => '2.0', u => 'MMH' },
             rateOfFallTo   => { v => 3.9, u => 'MMH' }}},
         { weatherSynopPrecip => { # 63
             rateOfFallFrom => { v => '4.0', u => 'MMH' },
             rateOfFallTo   => { v => 7.9, u => 'MMH' }}},
         { weatherSynopPrecip => { # 64
             rateOfFallFrom => { v => '8.0', u => 'MMH' },
             rateOfFallTo   => { v => 15.9, u => 'MMH' }}},
         { weatherSynopPrecip => { # 65
             rateOfFallFrom => { v => '16.0', u => 'MMH' },
             rateOfFallTo   => { v => 31.9, u => 'MMH' }}},
         { weatherSynopPrecip => { # 66
             rateOfFallFrom => { v => '32.0', u => 'MMH' },
             rateOfFallTo   => { v => 63.9, u => 'MMH' }}},
         { weatherSynopPrecip => { # 67
             rateOfFall     => { v => '64.0', u => 'MMH',
                                 q => 'isEqualGreater' }}},
         { }, # 68
         { }, # 69
         { weatherSynopPrecip => { # 70
             rateOfFall     => { v => '1.0', u => 'CMH', q => 'isLess' }}},
         { weatherSynopPrecip => { # 71
             rateOfFallFrom => { v => '1.0', u => 'CMH' },
             rateOfFallTo   => { v => 1.9, u => 'CMH' }}},
         { weatherSynopPrecip => { # 72
             rateOfFallFrom => { v => '2.0', u => 'CMH' },
             rateOfFallTo   => { v => 3.9, u => 'CMH' }}},
         { weatherSynopPrecip => { # 73
             rateOfFallFrom => { v => '4.0', u => 'CMH' },
             rateOfFallTo   => { v => 7.9, u => 'CMH' }}},
         { weatherSynopPrecip => { # 74
             rateOfFallFrom => { v => '8.0', u => 'CMH' },
             rateOfFallTo   => { v => 15.9, u => 'CMH' }}},
         { weatherSynopPrecip => { # 75
             rateOfFallFrom => { v => '16.0', u => 'CMH' },
             rateOfFallTo   => { v => 31.9, u => 'CMH' }}},
         { weatherSynopPrecip => { # 76
             rateOfFallFrom => { v => '32.0', u => 'CMH' },
             rateOfFallTo   => { v => 63.9, u => 'CMH' }}},
         { weatherSynopPrecip => { # 77
             rateOfFall     => { v => '64.0', u => 'CMH',
                                 q => 'isEqualGreater' }}}
       )[$idx - 47];

    $r->{weatherSynopPrecip}{phenomenon} = 'DZ' if $idx >= 50 && $idx <= 57;
    $r->{weatherSynopPrecip}{phenomenon} = 'RA' if $idx >= 60 && $idx <= 67;
    $r->{weatherSynopPrecip}{phenomenon} = 'SN' if $idx >= 70 && $idx <= 77;

    return $r;
}

sub _radiationType {
    my $idx = shift;

    return {
        0 => 'rad0PosNet',
        1 => 'rad1NegNet',
        2 => 'rad2GlobalSolar',
        3 => 'rad3DiffusedSolar',
        4 => 'rad4DownwardLongWave',
        5 => 'rad5UpwardLongWave',
        6 => 'rad6ShortWave',
    }->{$idx};
}

sub _getTempCity {
    my ($prefix, $sn, $temp, $tempAirF) = @_;
    my $r;

    $r->{s} = "$prefix$sn$temp";
    $r->{temp}{v} = $temp + 0;
    if ($sn == 1) {
        $r->{temp}{v} *= -1;
    } elsif (defined $tempAirF && $r->{temp}{v} + 50 < $tempAirF) {
        $r->{temp}{v} += 100;
    }
    $r->{temp}{u} = 'F';
    return $r;
}

sub _getRecordTemp {
    my $record = shift;

    $record =~ /(..)(.)(..)/;
    return {
        s            => "$1$2$3",
        recordPeriod => $3,
        recordType    => {
            v => $1,
            $2 eq 'X' ? (q => $1 eq 'LO' ? 'isLess' : 'isGreater') : ()
        }
    };
}

# FM18:
# FM18 (buoy):        ZZYY Abnnn YYMMJ    GGggi       QLLLLL LLLLLL (6QQQ/)
# Ab: maritime zone, nnn: 001..499, 500 added if drifting buoy

########################################################################
# _parseSynop
########################################################################
sub _parseSynop {
    my ($msg, $msg_hdr) = @_;
    my (%report, $corr_msg, $period, $windUnit, $windEST, $region, $country,
        $is_auto);

    my $re_D____D = '[A-Z\d]{3,}';
    my $re_A1bw   = '(?:1[1-7]|2[1-6]|3[1-4]|4[1-8]|5[1-6]|6[1-6]|7[1-4])';
    my $re_MMM    = '(?:\d{3})'; # 001..623, 901..936
    my $re_ULaULo = '(?:\d\d)';
    my $re_IIiii  = '(?:\d{5})';

    $report{msg} = $msg_hdr . $msg;
    $report{isSynop} = undef;

=pod

=head2 Parsing of SYNOP messages

=cut

    $msg .= ' '; # this makes parsing much easier

    $corr_msg = $msg;

=for comment ###########################################################

=head3 Section 0: information about the observation and the observing station

=for comment ###########################################################

    MiMiMjMj
    FM12 (fixed land):  AAXX                     YYGGiw IIiii
    FM13 (sea):         BBXX (D....D|A1bwnbnbnb) YYGGiw 99LaLaLa QcLoLoLoLo
    FM14 (mobile land): OOXX D....D              YYGGiw 99LaLaLa QcLoLoLoLo MMMULaULo h0h0h0h0im

=over

=begin html

=item

group M<sub>i</sub>M<sub>i</sub>M<sub>j</sub>M<sub>j</sub> <!--

=end html

=item group MiMiMjMj

=for html --><dd>

station type

=for html </dd>

=cut

    if ($msg !~ s/^((?:AA|BB|OO)XX) //) {
        $report{ERROR} = _makeErrorMsg 'obsStationType', $corr_msg, $msg;
        return %report;
    }
    $report{obsStationType} = { s => $1, stationType => $1 };

=begin html

=item

group D....D or A<sub>1</sub>b<sub>w</sub>n<sub>b</sub>n<sub>b</sub>n<sub>b</sub> <!--

=end html

=item group D....D or A1bwnbnbnb

=for html --><dd>

for sea stations: call sign or station id

for mobile land stations: call sign

=for html </dd>

=cut

    $region = '';
    $country = '';

    # group (D....D|A1bwnbnbnb)
    if ($report{obsStationType}{stationType} eq 'BBXX') {
        if ($msg !~ s@^(?:(${re_A1bw}\d{3})|($re_D____D)) @@o) {
            $report{ERROR} = _makeErrorMsg 'stationId', $corr_msg, $msg;
            return %report;
        }
        if (defined $1) {
            $report{callSign} = { s => $1, id => $1 };
        } else {
            $report{callSign} = { s => $2, id => $2 };
        }
    }
    if ($report{obsStationType}{stationType} eq 'OOXX') {
        if ($msg !~ s@^($re_D____D) @@o) {
            $report{ERROR} = _makeErrorMsg 'callSign', $corr_msg, $msg;
            return %report;
        }
        $report{callSign} = { s => $1, id => $1 };
    }

=begin html

=item

group YYGGi<sub>w</sub> <!--

=end html

=item group YYGGiw

=for html --><dd>

day and hour of observation, indicator for wind speed (unit)

=for html </dd>

=cut

    # group YYGGiw
    if ($msg !~ s@^(($re_day)($re_hour)([0134/])) @@o) {
        $report{ERROR} = _makeErrorMsg 'obsTimeWindInd', $corr_msg, $msg;
        return %report;
    }
    $report{obsTime}{s} = "$2$3";
    @{$report{obsTime}}{'day', 'hour', 'minute'} = ($2, $3, '00');

    # WMO-No. 306 Vol I.1, Part A, code table 1855:
    $report{windIndicator}{s} = $4;
    if ($4 ne '/') {
        $windUnit = $4 < 2 ? 'MPS' : 'KT';
        $windEST = $4 == 0 || $4 == 3;
        $report{windIndicator}{windUnit} = $windUnit;
        $report{windIndicator}{isEstimated} = undef if $windEST;
    } else {
        $report{windIndicator}{notAvailable} = undef;
    }

    if ($report{obsTime}{hour} =~ /00|06|12|18/) {
        $period = 6;
    } elsif ($report{obsTime}{hour} =~ /03|09|15|21/) {
        $period = 3;
    } else {
        $period = 1;
    }

=begin html

=item

group IIiii or 99L<sub>a</sub>L<sub>a</sub>L<sub>a</sub> Q<sub>c</sub>L<sub>o</sub>L<sub>o</sub>L<sub>o</sub>L<sub>o</sub> <!--

=end html

=item group IIiii or 99LaLaLa QcLoLoLoLo

=for html --><dd>

for fixed land stations: station identification

for sea or mobile land stations: position of the station

=for html </dd>

=cut

    if ($report{obsStationType}{stationType} eq 'AAXX') {
        # group IIiii
        if ($msg !~ m@^($re_IIiii) (.*)@o || !($region = _IIiii2region $1)) {
            $report{ERROR} = _makeErrorMsg 'stationId', $corr_msg, $msg;
            return %report;
        }
        $msg = $2;
        $report{obsStationId} = { s => $1, id => $1, region => $region };
        $country = _IIiii2country $1;
    } else {
        my ($lat_unit, $lon_unit);

        # group 99LaLaLa QcLoLoLoLo
        if ($msg =~ s@^(99/// /////) @@) {
            $report{stationPosition}{s} = $1;
            $report{stationPosition}{notAvailable} = undef;
            $lat_unit = '';
            $lon_unit = '';
        } elsif ($msg =~ s@^(99(\d(\d)\d) ([1357])(\d\d(\d)\d)) @@) {
            $report{stationPosition}{s} = $1;
            $report{stationPosition}{lat} = $2 / 10;
            $report{stationPosition}{lat} *= -1 if $4 eq '3' || $4 eq '5';
            $report{stationPosition}{lon} = $5 / 10;
            $report{stationPosition}{lon} *= -1 if $4 eq '5' || $4 eq '7';
            $lat_unit = $3;
            $lon_unit = $6;
        } else {
            $report{ERROR} = _makeErrorMsg 'stationPosition', $corr_msg, $msg;
            return %report;
        }

=begin html

=item

group MMMU<sub>La</sub>U<sub>Lo</sub> h<sub>0</sub>h<sub>0</sub>h<sub>0</sub>h<sub>0</sub>i<sub>m</sub> <!--

=end html

=item group MMMULaULo h0h0h0h0im

=for html --><dd>

for mobile land stations: position of the station (Marsden square, height)

=for html </dd>

=back

=cut

        if ($report{obsStationType}{stationType} eq 'OOXX') {
            # group MMMULaULo h0h0h0h0im
            if (   $msg !~ m@^((?:///|($re_MMM))(?://|($re_ULaULo)) (?:////|(\d{4}))([1-8])) (.*)@o
                || (   defined $2
                    && !(($2 >= 1 && $2 <= 623) || ($2 >= 901 && $2 <= 936)))
                || (defined $3 && "$lat_unit$lon_unit" ne $3))
            {
                $report{ERROR} = _makeErrorMsg 'stationPosition',$corr_msg,$msg;
                return %report;
            }
            $msg = $6;
            $report{stationPosition}{s} .= " $1";
            $report{stationPosition}{marsdenSquare} = $2 + 0 if defined $2;
            if (defined $4) {
                $report{stationPosition}{elevation} = {
                    v => $4 + 0,
                    u => ($5 <= 4 ? 'M' : 'FT'),
                    q => 'confidenceIs' .
                         { 1 => 'Excellent', 2 => 'Good',
                           3 => 'Fair',      0 => 'Poor' }->{$5 % 4}
                };
            }
        }
    }

=for comment ###########################################################

=head3 Section 1: Land Observations (data for global exchange common for all code forms)

=for comment ###########################################################

    iRixhVV Nddff (00fff) 1snTTT (2snTdTdTd|29UUU) 3P0P0P0P0 (4PPPP|4a3hhh) 5appp 6RRRtR (7wwW1W2|7wawaWa1Wa2) 8NhCLCMCH 9GGgg

=over

=begin html

=item

group i<sub>R</sub>i<sub>x</sub>hVV <!--

=end html

=item group iRixhVV

=for html --><dd>

precipitation indicator, weather indicator, base of lowest cloud

=for html </dd>

=cut

    # group iRixhVV
    # EXTENSION: iR can be '/': notAvailable
    # EXTENSION: ix can be '/': notAvailable
    if ($msg !~ s@^(([0-4/])([1-7/])([\d/])(\d\d|//)) @@) {
        $report{ERROR} = _makeErrorMsg 'indicatorCloudVis', $corr_msg, $msg;
        return %report;
    }

    # WMO-No. 306 Vol I.1, Part A, code table 1819:
    $report{precipInd}{s} = $2;
    if ($2 eq '/') {
        $report{precipInd}{notAvailable} = undef;
    } else {
        # ( 1 + 3, 1, 3, omitted (amount=0), omitted (NA) )[$2];
        $report{precipInd}{precipIndVal} = $2;
    }

    # WMO-No. 306 Vol I.1, Part A, code table 1860:
    # 1 Manned Included
    # 2 Manned Omitted (no significant phenomenon to report)
    # 3 Manned Omitted (no observation, data not available)
    # 4 Automatic Included using Code tables 4677 and 4561 (US: FMH-2 4-12/4-14)
    # 5 Automatic Omitted (no significant phenomenon to report)
    # 6 Automatic Omitted (no observation, data not available)
    # 7 Automatic Included using Code tables 4680 and 4531 (US: FMH-2 4-13/4-15)
    $report{wxInd}{s} = $3;
    if ($3 eq '/') {
        $report{wxInd}{notAvailable} = undef;
    } else {
        $report{wxInd}{wxIndVal} = $3;
    }
    if (   exists $report{wxInd}{wxIndVal}
        && $report{wxInd}{wxIndVal} >= 4)
    {
        $report{reportModifier}{s} = 'AUTO';
        $report{reportModifier}{modifierType} = $report{reportModifier}{s};
        $is_auto = 1;
    }

    if ($country eq 'US') {
        $report{baseLowestCloud} = _codeTable1600US $4;
    } else {
        $report{baseLowestCloud} = _codeTable1600 $4;
    }
    $report{baseLowestCloud}{s} = $4;

    if ($country eq 'US') {
        _codeTable4377US \%report, $5, $report{obsStationType}{stationType},
                         $is_auto;
    } elsif ($country eq 'CA') {
        _codeTable4377CA \%report, $5;
    } else {
        _codeTable4377 \%report, $5, $report{obsStationType}{stationType};
    }

=begin html

=item

group Nddff (00fff) <!--

=end html

=item group Nddff (00fff)

=for html --><dd>

total cloud cover, wind direction and speed

=cut

    # group Nddff (00fff)
    if ($msg !~ s@^([\d/])($re_wind_dir|//|99)(\d\d|//) @@o) {
        $report{ERROR} = _makeErrorMsg 'cloudWind', $corr_msg, $msg;
        return %report;
    }
    $report{totalCloudCover} = _codeTable2700 $1;
    $report{totalCloudCover}{s} = $1;
    $report{sfcWind}{s} = "$2$3";
    if ("$2$3" eq '////') {
        $report{sfcWind}{wind}{notAvailable} = undef;
    } elsif ($2 eq '00') {
        $report{sfcWind}{wind}{isCalm} = undef;
        $report{sfcWind}{wind}{isEstimated} = undef if $windEST;
    } else {
        if ($2 eq '//') {
            $report{sfcWind}{wind}{dirNotAvailable} = undef;
        } elsif ($2 eq '99') {
            $report{sfcWind}{wind}{dirVarAllUnk} = undef;
        } else {
            $report{sfcWind}{wind}{dir} = { v => $2 * 10, rp => 4, rn => 5 };
        }
        if ($3 eq '//') {
            $report{sfcWind}{wind}{speedNotAvailable} = undef;
        } else {
            $report{sfcWind}{wind}{speed} = { v => $3 + 0, u => $windUnit };
        }
        $report{sfcWind}{wind}{isEstimated} = undef if $windEST;
    }

    if (   exists $report{sfcWind}{wind}{speed}
        && $report{sfcWind}{wind}{speed} == 99)
    {
        if ($msg !~ s@^(00([1-9]\d\d)) @@) {
            $report{ERROR} = _makeErrorMsg 'wind', $corr_msg, $msg;
            return %report;
        }
        $report{sfcWind}{s} .= " $1";
        $report{sfcWind}{wind}{speed}{v} = $2 + 0;
    }

=item

=begin html

group 1s<sub>n</sub>TTT <!--

=end html

=item group 1snTTT

=for html --><dd>

temperature

=for html </dd>

=cut

    # group 1snTTT
    if ($msg =~ s@^(1(?:////|([01]\d\d[\d/]))) @@) {
        $report{temperature}{s} = $1;
        if (defined $2) {
            $report{temperature}{air}{temp} = _parseTemp $2;
        } else {
            $report{temperature}{air}{notAvailable} = undef;
        }
    }

=begin html

=item

group 2s<sub>n</sub>T<sub>d</sub>T<sub>d</sub>T<sub>d</sub> or 29UUU <!--

=end html

=item group 2snTdTdTd or 29UUU

=for html --><dd>

dewpoint or relative humidity

=for html </dd>

=cut

    # group 2snTdTdTd|29UUU
    if ($msg =~ s@^(2(?:////|([01]\d\d[\d/])|9(100|0\d\d))) @@) {
        $report{temperature}{s} .= ' ' if exists $report{temperature};
        $report{temperature}{s} .= $1;
        if (defined $2) {
            $report{temperature}{dewpoint}{temp} = _parseTemp $2;
            if (   exists $report{temperature}{air}
                && exists $report{temperature}{air}{temp})
            {
                @{$report{temperature}}{'relHumid1', 'relHumid2',
                                        'relHumid3', 'relHumid4'}
                    = _getHumidity $report{temperature}{air}{temp}{v},
                                   $report{temperature}{dewpoint}{temp}{v};
            }
        } elsif (defined $3) {
            $report{temperature}{relHumid1} = $3 + 0;
        } else {
            $report{temperature}{dewpoint}{notAvailable} = undef;
        }
    }

=begin html

=item

group 3P<sub>0</sub>P<sub>0</sub>P<sub>0</sub>P<sub>0</sub> <!--

=end html

=item group 3P0P0P0P0

=for html --><dd>

station level pressure

=for html </dd>

=cut

    # group 3P0P0P0P0
    if ($msg =~ s@^(3(?:(\d{4})|////)) @@) {
        $report{stationPressure}{s} = $1;
        if (defined $2) {
            $report{stationPressure}{hPa} = $2 / 10;
            $report{stationPressure}{hPa} += 1000
                if $report{stationPressure}{hPa} < 500;
            # force .0
            $report{stationPressure}{hPa} =
                                  sprintf '%.1f', $report{stationPressure}{hPa};
        } else {
            $report{stationPressure}{notAvailable} = undef;
        }
    }
    if (   !exists $report{stationPressure}
        && $report{obsStationType}{stationType} eq 'AAXX')
    {
        push @{$report{warning}}, { warningType => 'pressureMissing' };
    }

=begin html

=item

group 4PPPP or 4a<sub>3</sub>hhh <!--

=end html

=item group 4PPPP or 4a3hhh

=for html --><dd>

sea level pressure or geopotential height of an agreed standard isobaric surface

=for html </dd>

=cut

    # group 4PPPP|4a3hhh
    if ($msg =~ s@^(4////) @@) {
        $report{SLPhPa}{s} = $1;
        $report{SLPhPa}{notAvailable} = undef;
    } elsif ($msg =~ s@^(4([09]\d\d)([\d/])) @@) {
        $report{SLPhPa}{s} = $1;
        $report{SLPhPa}{hPa} = $2;
        $report{SLPhPa}{hPa} += 1000 if $2 < 500;
        $report{SLPhPa}{hPa} .= ".$3" unless $3 eq '/';
    } elsif ($msg =~ s@^(4([1-8])(\d{3})) @@) {
        my ($surface, $height) = ($2, $3);

        $report{gpSurface}{s} = $1;
        if ($surface =~ /[1278]/) {
            $report{gpSurface}{surface} = _codeTable0264 $surface;
            # hhh geopotential of an agreed standard isobaric surface given by a3, in standard geopotential metres, omitting the thousands digit.
            #    1 (1000):           100 gpm ->    0 ...  999
            #    2  (925):           800 gpm ->  300 ... 1299
            #    5  (500): 5000 ... 5500 gpm -> 4500 ... 5999 !
            #    7  (700):          3000 gpm -> 2500 ... 3499
            #    8  (850):          1500 gpm -> 1000 ... 1999
            if ($surface == 2) {
                $height += 1000 if $height < 300;
            } elsif ($surface == 7) {
                $height += ($height < 500 ? 3000 : 2000);
            } elsif ($surface == 8) {
                $height += 1000;
            }
            $report{gpSurface}{geopotential} = $height;
#       } elsif (   $surface == 6
#           && exists $report{obsStationId}
#           && $report{obsStationId}{id} == 84735)
#       {
#           $height += ($height < 500 ? 4000 : 3000);
#           $report{gpSurface}{geopotential} = $height;
        } else {
            $report{gpSurface}{invalidFormat} = $surface;
        }
    }

=begin html

=item

group 5appp <!--

=end html

=item group 5appp

=for html --><dd>

pressure tendency and amount of change

=cut

    # group 5appp
    if ($msg =~ s@^(5([0-8])(\d{3})) @@) {
        $report{pressureChange} = {
            s                 => $1,
            period            => '3h',
            pressureTendency  => $2,
            pressureChangeVal => {
                v => $3 / 10 * ($2 >= 5 ? -1 : 1),
                u => 'hPa'
        }};
    } elsif ($msg =~ s@^(5////) @@) {
        $report{pressureChange} = {
            s            => $1,
            period       => '3h',
            notAvailable => undef
        };
    }

=begin html

=item

group 6RRRt<sub>R</sub> <!--

=end html

=item group 6RRRtR

=for html --><dd>

amount of precipitation for given period

=for html </dd>

=cut

    # group 6RRRtR
    if ($msg =~ s@^(6(\d{3}|///)([\d/])) @@) {
        if (   exists $report{precipInd}{precipIndVal}
            && $report{precipInd}{precipIndVal} != 0
            && $report{precipInd}{precipIndVal} != 1)
        {
            push @{$report{warning}}, { warningType => 'precipNotOmitted1' };
        }

        $report{precipHourly} = _codeTable3590 $2;
        $report{precipHourly}{s} = $1;
        if ($3 eq '/') {
            $report{precipHourly}{periodNotAvailable} = undef;
        } elsif ($3 == 0) {
            $report{precipHourly}{periodNotAvailable} = undef;
        } else {
            $report{precipHourly}{period} = _codeTable4019 $3;
        }
    } elsif (   exists $report{precipInd}{precipIndVal}
             && (   $report{precipInd}{precipIndVal} == 0
                 || $report{precipInd}{precipIndVal} == 1))
    {
        push @{$report{warning}}, { warningType => 'precipOmitted1' };
    }

=begin html

=item

group 7wwW<sub>1</sub>W<sub>2</sub> or 7w<sub>a</sub>w<sub>a</sub>W<sub>a1</sub>W<sub>a2</sub> <!--

=end html

=item group 7wwW1W2 or 7wawaWa1Wa2

=for html --><dd>

present and past weather

=for html </dd>

=cut

    # group (7wwW1W2|7wawaWa1Wa2)
    if ($msg =~ s@^(7(\d\d|//)([\d/])([\d/])) @@) {
        if (   exists $report{wxInd}{wxIndVal}
            && $report{wxInd}{wxIndVal} != 1
            && $report{wxInd}{wxIndVal} != 4
            && $report{wxInd}{wxIndVal} != 7)
        {
            push @{$report{warning}}, { warningType => 'weatherNotOmitted' };
        }

        $report{weatherSynop}{s} = $1;
        if ($2 eq '//') {
            $report{weatherSynop}{weatherPresentNotAvailable} = undef;
        } else {
            $report{weatherSynop}{weatherPresent} = $2;
        }
        $report{weatherSynop}{period} = $period . 'h';
        if ($3 eq '/') {
            $report{weatherSynop}{weatherPast1NotAvailable} = undef;
        } else {
            $report{weatherSynop}{weatherPast1} = $3;
        }
        if ($4 eq '/') {
            $report{weatherSynop}{weatherPast2NotAvailable} = undef;
        } else {
            $report{weatherSynop}{weatherPast2} = $4;
        }
    } elsif (   exists $report{wxInd}{wxIndVal}
             && (   $report{wxInd}{wxIndVal} == 1
                 || $report{wxInd}{wxIndVal} == 4
                 || $report{wxInd}{wxIndVal} == 7))
    {
        push @{$report{warning}}, { warningType => 'weatherOmitted' };
    }

=begin html

=item

group 8N<sub>h</sub>C<sub>L</sub>C<sub>M</sub>C<sub>H</sub> <!--

=end html

=item group 8NhCLCMCH

=for html --><dd>

cloud type for each level and cloud cover of lowest reported cloud type

=for html </dd>

=cut

    # group 8NhCLCMCH
    if ($msg =~ s@^(8([/9])///|8([0-8/])([\d/])([\d/])([\d/])) @@) {
        if (defined $3) {
            $report{cloudTypes} = {
                $4 eq '/' ? ('cloudTypeLowNA', undef)   :('cloudTypeLow', $4),
                $5 eq '/' ? ('cloudTypeMiddleNA', undef):('cloudTypeMiddle',$5),
                $6 eq '/' ? ('cloudTypeHighNA', undef)  :('cloudTypeHigh', $6),
            };
            if ($3 ne '/') {
                if ($4 ne '/' && $4 != 0) {
                    $report{cloudTypes}{oktasLow} = $3;
                } elsif ($5 ne '/' && $5 != 0) {
                    $report{cloudTypes}{oktasMiddle} = $3;
                } else {
                    $report{cloudTypes}{oktas} = $3;
                }
            }
        } else {
            $report{cloudTypes}{$2 eq '/' ? 'notAvailable' : 'skyObscured'}
                                                                        = undef;
        }
        $report{cloudTypes}{s} = $1;
    } elsif ($msg =~ s@^(8[\d/]{4}) @@) {
        $report{cloudTypes} = { s => $1, invalidFormat => $1 };
    }

    # EXTENSION: tolerate multiple 8NhCLCMCH groups
    while ($msg =~ s@^(8[\d/]{4}) @@) {
        push @{$report{warning}}, { warningType => 'multCloudTypes', s => $1 };
    }

=begin html

=item

group 9GGgg <!--

=end html

=item group 9GGgg

=for html --><dd>

exact observation time

=back

=cut

    # group 9GGgg
    if ($msg =~ s@^(9($re_hour)($re_min)) @@o) {
        @{$report{exactObsTime}}{'s', 'hour', 'minute'} = ($1, $2, $3);
    }

=for comment ###########################################################

=head3 Section 2: Sea Surface Observations (maritime data for global exchange, optional)

(not implemented)

=for comment ###########################################################

=cut

# TODO: 222

    if ($msg =~ s@^(222[\d/]{2}(?: (?:[\d/]{5}|ICE))+) @@) {
        push @{$report{warning}}, { warningType => 'notProcessed', s => $1 };
    }

=for comment ###########################################################

=head3 Section 3: Climatological Data (data for regional exchange, optional)

=for comment ###########################################################

The WMO regions are:

=over

=item

I (Africa)

=item

II (Asia)

=item

III (South America)

=item

IV (North and Central Amerika)

=item

V (South-West Pacific)

=item

VI (Europe)

=item

Antarctic

=back

    region I: 0TgTgRcRt 1snTxTxTx 2snTnTnTn 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp 80000 0LnLcLdLg (1sLdLDLve)

    region II: 0EsnT'gT'g 1snTxTxTx 2snTnTnTn 3EsnTgTg 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp

    region III: 1snTxTxTx 2snTnTnTn 3EsnTgTg 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp

    region IV: 0CsDLDMDH 1snTxTxTx 2snTnTnTn 3E/// 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp TORNADO/ONE-MINUTE MAXIMUM x KNOTS AT x UTC

    region V: 1snTxTxTx 2snTnTnTn 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp

    region VI: 1snTxTxTx 2snTnTnTn 3EsnTgTg 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7R24R24R24R24 8NsChshs 9SPSPspsp

    Antarctic: 0dmdmfmfm (00200) 1snTxTxTx 2snTnTnTn 4E'sss 5j1j2j3j4 (j5j6j7j8j9) 6RRRtR 7DmDLDMDH 8NsChshs 9SPSPspsp

=cut

    if ($msg =~ s@^333 @@) {
        my (@s3, $r);

        @s3  = ();

=over

=begin html

=item

region I: group 0T<sub>g</sub>T<sub>g</sub>R<sub>c</sub>R<sub>t</sub> <!--

=end html

=item region I: group 0TgTgRcRt

=for html --><dd>

minumum ground temperature last night, character and start/end of precipitation

=for html </dd>

=cut

        # region I: group 0TgTgRcRt
        if ($region eq 'I' && $msg =~ s@^(?:0(//|\d\d)([\d/])([\d/])) @@) {
            my $tag;

            $r->{s} = "0$1";
            if ($1 eq '//') {
                $r->{notAvailable} = undef;
            } else {
                $r->{period} = 'n';
                $r->{temp}{v} = $1 > 50 ? -($1 - 50) : $1 + 0;
                $r->{temp}{u} = 'C';
            }
            push @s3, { tempMinGround => $r };
            $r = undef;

            # WMO-No. 306 Vol II, Chapter I, code table 167:
            $r->{s} = $2;
            if ($2 eq '/') {
                $r->{notAvailable} = undef
            } else {
                if ($2 == 1 || $2 == 5) {
                    $r->{phenomDescr} = 'isLight';
                } elsif ($2 == 2 || $2 == 6) {
                    $r->{phenomDescr} = 'isModerate';
                } elsif ($2 == 3 || $2 == 7) {
                    $r->{phenomDescr} = 'isHeavy';
                } elsif ($2 == 4 || $2 == 8) {
                    $r->{phenomDescr} = 'isVeryHeavy';
                } elsif ($2 == 9) {
                    $r->{phenomDescr} = 'isVariable';
                }
                if ($2 == 0) {
                    $r->{noPrecip} = undef;
                } elsif ($2 < 5) {
                    $r->{phenomDescr2} = 'isIntermittent';
                } elsif ($2 < 9) {
                    $r->{phenomDescr2} = 'isContinuous';
                }
            }
            push @s3, { precipCharacter => $r };
            $r = undef;

            # WMO-No. 306 Vol II, Chapter I, code table 168:
            # Rt time of beginning or end of precipitation
            $r->{s} = $3;
            if ($3 eq '/') {
                $r->{notAvailable} = undef
            } else {
                if ($3 == 0) {
                    $r->{noPrecip} = undef;
                } elsif ($3 == 1) {
                    $r->{hoursTill} = { v => 1, q => 'isLess' };
                } elsif ($3 < 7) {
                    $r->{hoursFrom} = $3 - 1;
                    $r->{hoursTill} = $3;
                } elsif ($3 < 9) {
                    $r->{hoursFrom} = ($3 - 7) * 2 + 6;
                    $r->{hoursTill} = ($3 - 7) * 2 + 8;
                } else {
                    $r->{hoursFrom} = { v => 10, q => 'isGreater' };
                }
            }
            # WMO-No. 306 Vol I.1, Part A, code table 4677:
            #   00..49 is not precipitation
            $tag =      exists $report{weatherSynop}
                     && exists $report{weatherSynop}{weatherPresent}
                     && $report{weatherSynop}{weatherPresent} >= 50
                   ? 'beginPrecip' : 'endPrecip';
            push @s3, { $tag => $r };
            $r = undef;
        }

=begin html

=item

region II: group 0Es<sub>n</sub>T'<sub>g</sub>T'<sub>g</sub> <!--

=end html

=item region II: group 0EsnT'gT'g

=for html --><dd>

state of ground, minimum ground temperature last night

=for html </dd>

=cut

        # EXTENSION: allow E for state of the ground
        if ($region eq 'II' && $msg =~ s@^(0([\d/E]))(///|[01]\d\d) @@) {
            $r->{s} = $1;
            if ($2 eq '/' || $2 eq 'E') {
                $r->{notAvailable} = undef;
            } else {
                $r->{stateOfGroundVal} = $2;
            }
            push @s3, { stateOfGround => $r };
            $r = undef;

            $r->{s} = $3;
            if ($3 eq '///') {
                $r->{notAvailable} = undef;
            } else {
                $r->{period} = 'n';
                $r->{temp} = _parseTemp "${3}0";
                $r->{temp}{v} += 0;
            }
            push @s3, { tempMinGround => $r };
            $r = undef;
        }

=begin html

=item

region IV: group 0C<sub>g</sub>D<sub>L</sub>D<sub>M</sub>D<sub>H</sub> <!--

=end html

=item region IV: group 0CsDLDMDH

=for html --><dd>

state of sky in tropics

=for html </dd>

=cut

        # region IV: group 0CsDLDMDH
        if ($region eq 'IV' && $msg =~ s@^0([\d/])([\d/])([\d/])([\d/]) @@) {
            # WMO-No. 306 Vol II, Chapter IV, code table 430
            $r->{s} = "0$1";
            if ($1 eq '/') {
                $r->{notAvailable} = undef;
            } else {
                $r->{stateOfSkyVal} = $1;
            }
            push @s3, { stateOfSky => $r };
            $r = undef;

            $r->{s} = "$2$3$4";
            _codeTable0700 $r, 'cloudTypeLow', $2;
            _codeTable0700 $r, 'cloudTypeMiddle', $3;
            _codeTable0700 $r, 'cloudTypeHigh', $4;
            push @s3, { cloudTypesDrift => $r };
            $r = undef;
        }

=begin html

=item

Antarctic: group 0d<sub>m</sub>d<sub>m</sub>f<sub>m</sub>f<sub>m</sub> <!--

=end html

=item Antarctic: group 0dmdmfmfm (00200)

=for html --><dd>

maximum wind during the preceding six hours

=for html </dd>

=cut

        # Antarctic: group 0dmdmfmfm (00200)
        if ($region eq 'Antarctic' && $msg =~ s@^(0([05][1-9]|[1267]\d|[38][0-6])(\d\d)( 00200)?) @@)
        {
            $r->{s} = $1;
            $r->{wind} = { speed => { v => $3, u => 'KT' }};
            $r->{wind}{speed}{v} += 200 if defined $4;

            # WMO-No. 306 Vol I.1, Part A, code table 0877:
            $r->{wind}{dir} = $2 * 10;
            if ($r->{wind}{dir} > 500) {
                $r->{wind}{dir} -= 500;
                $r->{wind}{speed}{v} += 100;
            }
            $r->{measurePeriod} = '1m';
            $r->{timeBeforeObsSince} = { hoursFrom => 6 };
            push @s3, { highestMeanSpeed => $r };
            $r = undef;
        }

=begin html

=item

group 1s<sub>n</sub>T<sub>x</sub>T<sub>x</sub>T<sub>x</sub> <!--

=end html

=item group 1snTxTxTx

=for html --><dd>

maximum temperature (regions I, II, VI: last 12 hours day-time; region III: day-time; region IV: at 00:00 and 18:00 last 12 hours, at 06:00 last 24 hours, at 12:00 previous day; region V: last 24 hours, Antarctic: last 12 hours)

=for html </dd>

=cut

        # group 1snTxTxTx
        if ($region && $msg =~ s@^(1(?:([01]\d{3})|////)) @@) {
            $r->{s} = $1;
            if (defined $2) {
                $r->{temp} = _parseTemp $2;
            } else {
                $r->{notAvailable} = undef;
            }
            if ($region eq 'V') {
                $r->{period} = '24h';
                push @s3, { tempMaxPeriod => $r };
            } elsif ($region eq 'IV') {
                if ($report{obsTime}{hour} =~ /00|18/) {
                    $r->{period} = '12h';
                } elsif ($report{obsTime}{hour} == 6) {
                    $r->{period} = '24h';
                } elsif ($report{obsTime}{hour} == 12) {
                    $r->{period} = '1d';
                } else {
                    $r->{periodNotAvailable} = undef;
                }
                push @s3, { tempMaxPeriod => $r };
            } elsif ($region eq 'Antarctic') {
                $r->{period} = '12h';
                push @s3, { tempMaxPeriod => $r };
            } else { # I, II, III, VI
                $r->{period} = '12h' unless $region eq 'III';
                push @s3, { tempMaxDaytime => $r };
            }
            $r = undef;
        }

=begin html

=item

group 2s<sub>n</sub>T<sub>n</sub>T<sub>n</sub>T<sub>n</sub> <!--

=end html

=item group 2snTnTnTn

=for html --><dd>

minimum temperature (regions I, II, VI: night-time in the last 12 hours; region III: night-time; region IV: at 00:00 last 18 hours, at 06:00 and 18:00 last 24 hours, at 12:00 last 12 hours; region V: in the last 24 hours, Antarctic: last 12 hours)

=for html </dd>

=cut

        # group 2snTnTnTn
        if ($region && $msg =~ s@^(2(?:([01]\d{3})|////)) @@) {
            $r->{s} = $1;
            if (defined $2) {
                $r->{temp} = _parseTemp $2;
            } else {
                $r->{notAvailable} = undef;
            }
            if ($region eq 'V') {
                $r->{period} = '24h',
                push @s3, { tempMinPeriod => $r };
            } elsif ($region eq 'IV') {
                if ($report{obsTime}{hour} =~ /06|18/) {
                    $r->{period} = '24h';
                } elsif ($report{obsTime}{hour} == 0) {
                    $r->{period} = '18h';
                } elsif ($report{obsTime}{hour} == 12) {
                    $r->{period} = '12h';
                } else {
                    $r->{periodNotAvailable} = undef;
                }
                push @s3, { tempMinPeriod => $r };
            } elsif ($region eq 'Antarctic') {
                $r->{period} = '12h';
                push @s3, { tempMinPeriod => $r };
            } else { # I, II, III, VI
                $r->{period} = '12h' unless $region eq 'III';
                push @s3, { tempMinNighttime => $r };
            }
            $r = undef;
        }

=begin html

=item

regions II, VI: group 3Es<sub>n</sub>T<sub>g</sub>T<sub>g</sub>, region IV: 3E/// <!--

=end html

=item regions II, VI: group 3EsnTgTg, region IV: 3E///

=for html --><dd>

state of ground, minimum ground temperature last night (regions II, VI)

=for html </dd>

=cut

        # region II: group 3EsnTgTg
        # region VI: group 3EsnTgTg
        if (   (   ($region eq 'II' || $region eq 'VI')
                && $msg =~ s@^(?:3([\d/])(///|[01]\d\d)) @@)
            || (   $region eq 'IV'
                && $msg =~ s@^(?:3([\d/])(///)) @@))
        {
            $r->{s} = "3$1";
            $r->{s} .= $2 if $region eq 'IV';
            if ($1 eq '/') {
                $r->{notAvailable} = undef;
            } else {
                $r->{stateOfGroundVal} = $1;
            }
            push @s3, { stateOfGround => $r };
            $r = undef;

            if ($region ne 'IV') {
                $r->{s} = $2;
                if ($2 eq '///') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{period} = 'n';
                    $r->{temp} = _parseTemp "${2}0";
                    $r->{temp}{v} += 0;
                }
                push @s3, { tempMinGround => $r };
                $r = undef;
            }
        }

=begin html

=item

group 4E'sss <!--

=end html

=item group 4E'sss

=for html --><dd>

state of ground if covered with snow or ice, snow depth

=cut

        # group 4E'sss
        if ($msg =~ s@^(?:4([\d/])(///|\d{3})) @@) {
            $r->{s} = "4$1";
            if ($1 eq '/') {
                $r->{notAvailable} = undef;
            } else {
                $r->{stateOfGroundSnowVal} = $1;
            }
            push @s3, { stateOfGroundSnow => $r };
            $r = undef;

            # WMO-No. 306 Vol I.1, Part A, code table 3889:
            $r->{s} = $2;
            if ($2 eq '///') {
                $r->{notAvailable} = undef;
            } elsif ($2 eq '000') {
                $r->{invalidFormat} = undef;
            } elsif ($2 eq '997') {
                $r->{precipAmount} = { v => 0.5, u => 'CM', q => 'isLess' };
            } elsif ($2 eq '998') {
                $r->{coverNotCont} = undef;
            } elsif ($2 eq '999') {
                $r->{noMeasurement} = undef;
            } else {
                $r->{precipAmount} = { v => $2 + 0, u => 'CM' };
            }
            push @s3, { snowDepth => $r };
            $r = undef;
        }

=begin html

=item

group 5j<sub>1</sub>j<sub>2</sub>j<sub>3</sub>j<sub>4</sub> (j<sub>5</sub>j<sub>6</sub>j<sub>7</sub>j<sub>8</sub>j<sub>9</sub>) <!--

=end html

=item group 5j1j2j3j4 (j5j6j7j8j9)

=for html --><dd>

evaporation, temperature change, duration of sunshine, radiation type and amount, direction of cloud drift, direction and elevation of cloud, pressure change

=for html </dd>

=cut

        {
            my ($msg5, $had_evapo, $had_pChg, $had_sun_1d, $had_sun_1h,
                %had_rad, $had_drift);

            # group 5j1j2j3j4 (j5j6j7j8j9)

            # determine all 5xxxx groups with 0xxxx .. 6xxxx suppl. groups:
            # 1. use only consecutive groups 0xxxx .. 6xxxx
            $msg5 = ($msg =~ /^(5[\d\/]{4} (?:[0-6][\d\/]{4} )*)/)[0];
            $msg5 = '' unless defined $msg5;
            # 2. remove trailing 6xxxx if precipitation indicator is 0 or 2
            if (   exists $report{precipInd}{precipIndVal}
                && (   $report{precipInd}{precipIndVal} == 0
                    || $report{precipInd}{precipIndVal} == 2))
            {
                $msg5 =~ s/6[\d\/]{4} $//;
            }

            # strip $msg5 from $msg, use $msg5 for group(s) 5j1j2j3j4
            $msg = substr $msg, length $msg5;

            # allow any order, but check for duplicates/impossible combinations
            while ($msg5 ne '') {
                my $match_found;

                $match_found = 0;

                if ($msg5 =~ s@^(5///[\d/]) @@) {
                    push @s3, { group5xxxxNA => { s => $1 }};
                    $match_found = 1;
                }

                # group 5EEEiE
                if (!$had_evapo && $msg5 =~ s@^(5([0-3]\d\d)(\d)) @@) {
                    push @s3, { evapo => {
                        s              => $1,
                        evapoAmount    => $2 / 10,
                        evapoIndicator => $3
                    }};
                    $match_found = 1;
                    $had_evapo = 1;
                }

                # group 54g0sndT
                if ($msg5 =~ s@^(54([0-5])([01])(\d)) @@) {
                    $r->{s} = $1;
                    $r->{hoursFrom} = $2;
                    $r->{hoursTill} = $2 + 1;

                    # WMO-No. 306 Vol I.1, Part A, code table 0822:
                    $r->{temp}{v} = $4 < 5 ? $4 + 10 : $4 + 0;
                    $r->{temp}{v} *= -1 if $3 == 1;
                    $r->{temp}{u} = 'C';
                    push @s3, { tempChange => $r };
                    $match_found = 1;
                    $r = undef;
                }

                # group 55SSS (j5j6j7j8j9)*, SSS = 000..240
                if (   !$had_sun_1d
                    && $msg5 =~ s@^(55(?:///|(${re_hour}\d|240))) @@o)
                {
                    $r->{s} = $1;
                    $r->{sunshinePeriod} = '1d';
                    if (defined $2) {
                        $r->{sunshine} = { v => $2 / 10, u => 'hour' };
                    } else {
                        $r->{sunshineNotAvailable} = undef;
                    }
                    for ($msg5 =~ m@^(0(?:////|\d{4}) )?(1(?:////|\d{4}) )?(2(?:////|\d{4}) )?(3(?:////|\d{4}) )?(4(?:////|\d{4}) )?(5[0-4]\d{3} )?(6(?:////|\d{4}) )?@)
                    {
                        next unless defined $_;

                        $msg5 = substr $msg5, 6;

                        /(.)(.*) /;
                        $r->{s} .= " $1$2";
                        $r->{radiationPeriod} = '24h';
                        if ($2 eq '////') {
                            $r->{_radiationType $1}{notAvailable} = undef;
                        } else {
                            $r->{_radiationType $1}{radiationValue} =
                                                      { v => $2 + 0, u => 'Jcm2' };
                        }
                    }
                    push @s3, { radiationSun => $r };
                    $match_found = 1;
                    $had_sun_1d = 1;
                    $r = undef;
                }

                # group 553SS (j5j6j7j8j9)*, SS = 00..10
                if (!$had_sun_1h && $msg5 =~ s@^(553(?://|(0\d|10))) @@) {
                    $r->{s} = $1;
                    $r->{sunshinePeriod} = '1h';
                    if (defined $2) {
                        $r->{sunshine} = { v => $2 / 10, u => 'hour' };
                    } else {
                        $r->{sunshineNotAvailable} = undef;
                    }
                    for ($msg5 =~ m@^(0(?:////|\d{4}) )?(1(?:////|\d{4}) )?(2(?:////|\d{4}) )?(3(?:////|\d{4}) )?(4(?:////|\d{4}) )?(5[0-4]\d{3} )?(6(?:////|\d{4}) )?@)
                    {
                        next unless defined $_;

                        $msg5 = substr $msg5, 6;

                        /(.)(.*) /;
                        $r->{s} .= " $1$2";
                        $r->{radiationPeriod} = '1h';
                        if ($2 eq '////') {
                            $r->{_radiationType $1}{notAvailable} = undef;
                        } else {
                            $r->{_radiationType $1}{radiationValue} =
                                                   { v => $2 + 0, u => 'kJm2' };
                        }
                    }
                    push @s3, { radiationSun => $r };
                    $match_found = 1;
                    $had_sun_1h = 1;
                    $r = undef;
                }

                if (   $msg5 =~ m@^(55([45])0([78])) \2(\d{4}|////) (.*)@
                    && !exists $had_rad{"$2$3"})
                {
                    $msg5 = $5;
                    $r->{s} = "$1 $2$4";
                    $r->{radiationPeriod} = $2 eq '4' ? '1h' : '24h';
                    if ($4 eq '////') {
                        $r->{notAvailable} = undef;
                    } else {
                        $r->{radiationValue} =
                              { v => $4 + 0, u => $2 eq '4' ? 'kJm2' : 'Jcm2' };
                    }
                    push @s3,
                        { ($3 eq '7' ? 'radShortWave' : 'radDirectSolar'), $r };
                    $match_found = 1;
                    $had_rad{"$2$3"} = undef;
                    $r = undef;
                }

                # group 56DLDMDH, direction of cloud drift
                if (!$had_drift && $msg5 =~ s@^(56([\d/])([\d/])([\d/])) @@) {
                    $r->{s} = $1;
                    _codeTable0700 $r, 'cloudTypeLow', $2;
                    _codeTable0700 $r, 'cloudTypeMiddle', $3;
                    _codeTable0700 $r, 'cloudTypeHigh', $4;
                    push @s3, { cloudTypesDrift => $r };
                    $match_found = 1;
                    $had_drift = 1;
                    $r = undef;
                }

                # group 57CDaeC, direction and elevation of cloud
                if ($msg5 =~ s@^(57([\d/])(\d)(\d)) @@) {
                    $r->{s} = $1;
                    _codeTable0500 $r, $2;
                    _codeTable0700 $r, 'cloud', $3;
                    _codeTable1004 $r, $4;
                    push @s3, { cloudLocation => $r };
                    $match_found = 1;
                    $r = undef;
                }

                # group 58p24p24p24
                if (!$had_pChg && $msg5 =~ s@^(58(?:(\d{3})|///)) @@) {
                    push @s3, { pressureChange => {
                        s      => $1,
                        period => '24h',
                        defined $2
                           ? (pressureChangeVal => { v => $2 / 10, u => 'hPa' })
                           : ('notAvailable', undef)
                    }};
                    $match_found = 1;
                    $had_pChg = 1;
                }

                # group 59p24p24p24
                if (!$had_pChg && $msg5 =~ s@^(59(?:(\d{3})|///)) @@) {
                    push @s3, { pressureChange => {
                        s      => $1,
                        period => '24h',
                        defined $2 ? (pressureChangeVal => {
                                               v => (-$2 / 10) * 1, # avoid "-0"
                                               u => 'hPa'
                                     })
                                   : ('notAvailable', undef)
                    }};
                    $match_found = 1;
                    $had_pChg = 1;
                }
                last unless $match_found; # no match but $msg5 was not empty
            }

            # EXTENSION: allow un-announced 6xxxx group
            if ($msg5 =~ m@^6(?:\d{3}|///)\d $@) {
                $msg = $msg5 . $msg;
                $msg5 = '';
            }
            if ($msg5 ne '') {
                $msg = $msg5 . $msg;
                $report{ERROR} = _makeErrorMsg 'invalid333-5xxxx',
                                               $corr_msg, $msg;
                $report{section3} = \@s3;
                return %report;
            }
        }

=begin html

=item

group 6RRRt<sub>R</sub> <!--

=end html

=item group 6RRRtR

=for html --><dd>

amount of precipitation for given period

=for html </dd>

=cut

        # group 6RRRtR
        if (   $msg =~ m@^(6(\d{3}|///)([\d/])) (.*)@
            && ($country eq 'BD' || $3 ne '/'))
        {
            if (   exists $report{precipInd}{precipIndVal}
                && $report{precipInd}{precipIndVal} != 0
                && $report{precipInd}{precipIndVal} != 2)
            {
                push @{$report{warning}},
                                     { warningType => 'precipNotOmitted3' };
            }

            $msg = $4;
            $r = _codeTable3590 $2;
            $r->{s} = $1;
            if ($3 eq '/') {
                # WMO-No. 306 Vol II, Chapter II, Section D:
                $r->{period} = '3h'; # BD
            } elsif ($3 > 0) {
                $r->{period} = _codeTable4019 $3;
            } else {
                $r->{periodNotAvailable} = undef;
            }
            push @s3, { precipHourly => $r };
            $r = undef;
        } elsif (   exists $report{precipInd}{precipIndVal}
                 && (   $report{precipInd}{precipIndVal} == 0
                     || $report{precipInd}{precipIndVal} == 2))
        {
            push @{$report{warning}}, { warningType => 'precipOmitted3' };
        }

=begin html

=item

regions I..VI: group 7R<sub>24</sub>R<sub>24</sub>R<sub>24</sub>R<sub>24</sub> <!--

=end html

=item regions I..VI: group 7R24R24R24R24

=for html --><dd>

amount of precipitation in the last 24 hours

=for html </dd>

=cut

        # EXTENSION: allow 7////
        # regions I..VI: group 7R24R24R24R24
        if ($region =~ /^(?:I|II|III|IV|V|VI)$/) {
            if ($msg =~ s@^(7(\d{4})) @@) {
                $r->{s} = $1;
                $r->{period} = '24h';
                if ($2 == 9999) {
                    $r->{traces} = undef;
                } elsif ($2 == 9998) {
                    $r->{precipAmount} = { v => 999.8, u => 'MM',
                                           q => 'isEqualGreater' };
                } else {
                    $r->{precipAmount} = { v => $2 / 10, u => 'MM' };
                }
                push @s3, { precipHourly => $r };
                $r = undef;
            } elsif ($msg =~ s@^(7////) @@) {
                push @s3, { precipHourly => {
                    s            => $1,
                    period       => '24h',
                    notAvailable => undef
                }};
            }
        }

=begin html

=item

Antarctic: group 7D<sub>m</sub>D<sub>L</sub>D<sub>M</sub>D<sub>H</sub> <!--

=end html

=item Antarctic: group 7DmDLDMDH

=for html --><dd>

maximum wind during the preceding six hours

=for html </dd>

=cut

        # Antarctic: group 7DmDLDMDH
        if (   $region eq 'Antarctic'
            && $msg =~ s@^7([1-8])([\d/])([\d/])([\d/]) @@)
        {
            $r->{s} = "7$1";
            _codeTable0700 $r, 'compass', $1;
            push @s3, { windDir6h => $r };
            $r = undef;

            $r->{s} = "$2$3$4";
            _codeTable0700 $r, 'cloudTypeLow', $2;
            _codeTable0700 $r, 'cloudTypeMiddle', $3;
            _codeTable0700 $r, 'cloudTypeHigh', $4;
            push @s3, { cloudTypesDrift => $r };
            $r = undef;
        }

=begin html

=item

group 8N<sub>s</sub>Ch<sub>s</sub>h<sub>s</sub> <!--

=end html

=item group 8NsChshs

=for html --><dd>

cloud cover and height for cloud layers

=for html </dd>

=cut

        # group 8NsChshs (but not 80000!)
        while (   $msg =~ m@^(8([\d/])([\d/])([0-46-9]\d|5[06-9]|//)) (.*)@
               && $1 ne '80000')
        {
            my ($tag, $dist);

            $msg = $5;
            $r->{s} = $1;
            $r->{cloudOktas} = _codeTable2700 $2;
            _codeTable0500 $r, $3;
            $tag = $2 eq '9' && $3 eq '/' ? 'visVert': 'cloudBase';
            if ($4 eq '//') {
                $r->{cloudBaseNotAvailable} = undef;
            } else {
                $dist = _codeTable1677 $4;
                if (ref $dist eq 'HASH') {
                    if ($tag eq 'visVert') {
                        $r->{$tag}{distance} = $dist;
                        $r->{$tag}{distance}{u} = 'M';
                        $r->{$tag}{s} = $4;
                    } else {
                        $r->{$tag} = $dist;
                        $r->{$tag}{u} = 'M';
                    }
                } else {
                    if ($tag eq 'visVert') {
                        $r->{"${tag}From"}{distance}{v} = $dist->[0];
                        $r->{"${tag}From"}{distance}{u} = 'M';
                        $r->{"${tag}To"}{distance}{v} = $dist->[1];
                        $r->{"${tag}To"}{distance}{u} = 'M';
                    } else {
                        $r->{"${tag}From"}{v} = $dist->[0];
                        $r->{"${tag}From"}{u} = 'M';
                        $r->{"${tag}To"}{v} = $dist->[1];
                        $r->{"${tag}To"}{u} = 'M';
                    }
                }
            }
            push @s3, { cloudInfo => $r };
            $r = undef;
        }

=begin html

=item

group 9S<sub>P</sub>S<sub>P</sub>s<sub>p</sub>s<sub>p</sub> <!--

=end html

=item group 9SPSPspsp

=for html --><dd>

supplementary information

=cut

        # WMO-No. 306 Vol I.1, Part A, code table 3778:
        # group 9SPSPspsp
        while ($msg =~ m@^9\d{4} @) {
            $r = undef;

            # group 90(0(tt|zz)|[15]tt)
            if ($msg =~ s@^(90(?:0\d\d|[15]$re_synop_tt)) @@o) {
                my ($s, $type);
                $s = $1;
                $s =~ /..(.)(..)/;
                $type = { 0 => 'timeBeforeObsBegin',
                          1 => 'timeBeforeObsEnd',
                          5 => 'timeBeforeObsSince'}->{$1};
                $r = _codeTable4077 $2, $type;
                $r->{s} = $s;
                push @s3, { weatherSynopInfo => $r };
            # group 909Rtdc
            } elsif ($msg =~ s@^909([1-9])([0-79]) @@) {
                my $tag;

                $r = _codeTable3552 $1;
                $r->{s} = "909$1";
                # WMO-No. 306 Vol I.1, Part A, code table 4677:
                #   00..49 is not precipitation
                $tag =      exists $report{weatherSynop}
                         && exists $report{weatherSynop}{weatherPresent}
                         && $report{weatherSynop}{weatherPresent} >= 50
                       ? 'beginPrecip' : 'endPrecip';
                push @s3, { $tag => $r };
                $r = undef;

                $r = _codeTable0833 $2;
                $r->{s} = $2;
                if ($2 <= 3) {
                    $r->{onePeriod} = undef;
                } elsif ($2 <= 7) {
                    $r->{morePeriods} = undef;
                } else {
                    $r->{periodsNA} = undef;
                }
                push @s3, { precipPeriods => $r };
            # group 910ff (00fff)
            } elsif ($msg =~ s@^(910(\d\d)) @@) {
                $r->{s} = $1;
                $r->{period}         = '10m';
                $r->{wind}{speed}{u} = $windUnit;
                $r->{wind}{speed}{v} = $2 + 0;
                if ($r->{wind}{speed}{v} == 99) {
                    if ($msg !~ s@^(00([1-9]\d\d)) @@) {
                        $report{ERROR} = _makeErrorMsg 'wind', $corr_msg, $msg;
                        $report{section3} = \@s3;
                        return %report;
                    }
                    $r->{s} .= " $1";
                    $r->{wind}{speed}{v} = $2 + 0;
                }
                push @s3, { highestGust => $r };
            # TODO group 906tt (913ff (00fff) or/and 915dd
            # group (90[247]tt) 91[1-4]ff (00fff) (903tt)
            } elsif ($msg =~ s@^((?:90([247])($re_synop_period) )?91([1-4])(\d\d)) @@o)
            {
                my $type;

                if (defined $2) {
                    $r = _codeTable4077 $3, (
                        { 2 => 'timeBeforeObsBegin',
                          4 => 'timeBeforeObs',
                          7 => 'timeBeforeObsSince'}->{$2});
                    $r->{measurePeriod} = '10m'
                        if $2 == 4 || $4 == 2 || $4 == 4;
                } else {
                    $r->{timeBeforeObsSince} = { hoursFrom => $period };
                }
                $r->{s} = $1;
                $type = { 1 => 'highestGust',
                          2 => 'highestMeanSpeed',
                          3 => 'meanSpeed',
                          4 => 'lowestMeanSpeed' }->{$4};
                $r->{wind}{speed}{u} = $windUnit;
                $r->{wind}{speed}{v} = $5 + 0;
                if ($r->{wind}{speed}{v} == 99) {
                    if ($msg !~ s@^(00([1-9]\d\d)) @@) {
                        $report{ERROR} = _makeErrorMsg 'wind', $corr_msg, $msg;
                        $report{section3} = \@s3;
                        return %report;
                    }
                    $r->{s} .= " $1";
                    $r->{wind}{speed}{v} = $2 + 0;
                }
                if ($msg =~ s@^903($re_synop_period) @@o) {
                    $r->{s} .= " 903$1";
                    $r->{timeBeforeObsEnd} =
                            (_codeTable4077 $1,
                                    'timeBeforeObsEnd')->{timeBeforeObsEnd};
                }
                push @s3, { $type => $r };
            # group (907tt) 930RR
            } elsif ($msg =~ s@^((?:907($re_synop_tt) )?930(\d\d)) @@o) {
                if (defined $2) {
                    $r = _codeTable4077 $2, 'timeBeforeObsSince';
                } else {
                    $r->{period} = $period . 'h';
                }
                $r->{s} = $1;
                _codeTable3570 \$r, $3, 'precipAmount';
                push @s3, { precipHourly => $r };
            # group (907tt) 931ss or 931s's'
            } elsif ($msg =~ s@^((?:907($re_synop_tt) )?931(\d\d)) @@o) {
                if (defined $2) {
                    if (   $country eq 'CH'
                        && (   ($2 == 68 && $report{obsTime}{hour} ==  6)
                            || ($2 == 66 && $report{obsTime}{hour} == 18)))
                    {
                        $r->{timeBeforeObsSince} =
                                        { hoursFrom => $2 == 68 ? 24 : 12 };
                    } else {
                        $r = _codeTable4077 $2, 'timeBeforeObsSince';
                    }
                } else {
                    $r->{period} = $period . 'h';
                }
                $r->{s} = $1;
                # WMO-No. 306 Vol II, Chapter VI, Section D:
                if ($country eq 'FR') { # 931s's'
                    $r->{precipAmount} = { v => $3 + 0, u => 'CM' };
                    $r->{precipAmount}{q} = 'isEqualGreater' if $3 == 99;
                } elsif ($country eq 'AT') {
                    $r->{period} = '24h' if $report{obsTime}{hour} == 6;
                    _codeTable3870 \$r, $3;
                    $r->{precipAmount}{v} = 5 if $3 == 97;
                } else {
                    _codeTable3870 \$r, $3;
                }
                push @s3, { snowFall => $r };
            # group 96[024]ww
            } elsif ($msg =~ s@^(96([024])(\d\d)) @@) {
                push @s3, { { 0 => 'weatherSynopAdd',
                              2 => 'weatherSynopAmplPast',
                              4 => 'weatherSynopAmpl' }->{$2} => {
                                s              => $1,
                                $2 == 4 ? ('period', $period . 'h') : (),
                                weatherPresent => $3
                }};
            # group (90[2467]tt) 966ww (903tt)
            } elsif (   $msg =~ m@^((?:90([2467])($re_synop_period) )?966(\d\d)(?: 903($re_synop_period))?) (.*)@o
                     && (defined $2 || defined $5))
            {
                $msg = $6;
                if (defined $2) {
                    $r = _codeTable4077 $3, (
                            { 2 => 'timeBeforeObsBegin',
                              4 => 'timeBeforeObs',
                              6 => 'timeBeforeObsSince',
                              7 => 'timeBeforeObsSince'}->{$2});
                    $r->{timeBeforeObsEnd} =
                            (_codeTable4077 $5,
                                     'timeBeforeObsEnd')->{timeBeforeObsEnd}
                        if defined $5;
                } else {
                    $r = _codeTable4077 $5, 'timeBeforeObsEnd';
                }
                $r->{s} = $1;
                $r->{weatherPresent} = $4;
                push @s3, { weatherSynopPast => $r };
            # group 96[135]w1w1
            } elsif ($msg =~ s@^(96([135])($re_synop_w1w1)) @@o) {
                push @s3, { { 1 => 'weatherSynopAdd',
                              3 => 'weatherSynopAmplPast',
                              5 => 'weatherSynopAmpl' }->{$2} => {
                                s              => $1,
                                $2 == 5 ? ('period', $period . 'h') : (),
                                %{ _codeTable4687 $3 }
                }};
            # group (90[2467]tt) 967w1w1 (903tt)
            } elsif (   $msg =~ m@^((?:90([2467])($re_synop_period) )?967($re_synop_w1w1)(?: 903($re_synop_period))?) (.*)@o
                     && (defined $2 || defined $5))
            {
                $msg = $6;
                if (defined $2) {
                    $r = _codeTable4077 $3, (
                            { 2 => 'timeBeforeObsBegin',
                              4 => 'timeBeforeObs',
                              6 => 'timeBeforeObsSince',
                              7 => 'timeBeforeObsSince'}->{$2});
                    $r->{timeBeforeObsEnd} =
                            (_codeTable4077 $5,
                                     'timeBeforeObsEnd')->{timeBeforeObsEnd}
                        if defined $5;
                } else {
                    $r = _codeTable4077 $5, 'timeBeforeObsEnd';
                }
                push @s3, { weatherSynopPast => {
                                s              => $1,
                                %{$r},
                                %{ _codeTable4687 $4 }
                }};
            # group 950Nmn3
            } elsif ($msg =~ s@^(950(\d)(\d)) @@o) {
                push @s3, { conditionMountain => {
                                s             => $1,
                                condMountainLoc => {
                                    cloudMountain => $2,
                                    cloudEvol     => $3
                }}};
            # group 951Nvn4
            } elsif ($msg =~ s@^(951(\d)(\d)) @@o) {
                push @s3, { conditionValley => {
                                s             => $1,
                                condValleyLoc => {
                                    cloudValley    => $2,
                                    cloudBelowEvol => $3
                }}};
            # TODO: more 9xxxx stuff
            # } elsif () {
            } else {
                last;
            }
        }

        # region I: group 80000 0LnLcLdLg (1sLdLDLve)
        if ($region eq 'I' && $msg =~ m@^80000 (.*)@) {
           # TODO
           #while ($msg =~ m@^(?:0(\d{4}))(?: 1(\d{4}))? (.*)@) {
           #}
        }

        # region IV: group TORNADO/ONE-MINUTE MAXIMUM x KNOTS AT x UTC
        if ($region eq 'IV') {
            $msg =~ s@(?<=^TORNADO)/(?=ONE-MINUTE MAXIMUM [1-9]\d+ KNOTS AT $re_hour:$re_min UTC )@ @o;

            # region IV: group TORNADO
            if ($msg =~ s@^(TORNADO) @@) {
                push @s3, { weather => { s => $1, tornado => undef }};
            }

            # region IV: group ONE-MINUTE MAXIMUM x KNOTS AT x UTC
            if ($msg =~ s@^(ONE-MINUTE MAXIMUM ([1-9]\d+) KNOTS AT ($re_hour):($re_min) UTC) @@o)
            {
                push @s3, { highestMeanSpeed => {
                    s             => $1,
                    hour          => $3,
                    minute        => $4,
                    measurePeriod => '1m',
                    wind          => { speed => { v => $2, u => 'KT' }}}};
            }
        }

        $report{section3} = \@s3;
    }

=back

=for comment ###########################################################

=head3 Section 4: Clouds with base below station level (data for national use, optional)

=for comment ###########################################################

    N'C'H'H'Ct

=over

=begin html

=item

group N'C'H'H'C<sub>t</sub> <!--

=end html

=item group N'C'H'H'Ct

=for html --><dd>

data for clouds with base below station level

=for html </dd>

=cut

    if ($msg =~ s@^444 @@) {
        my (@s4, $r);

        @s4  = ();

        # group N'C'H'H'Ct
        while ($msg =~ s@^(([\d/])([\d/])(\d\d)(\d)) @@) {
            $r->{s} = $1;
            $r->{cloudOktas} = _codeTable2700 $2;
            _codeTable0500 $r, $3;

            # WMO-No. 306 Vol I.1, Part A, Section B:
            # H'H' altitude of the upper surface of clouds reported by C', in hundreds of metres
            $r->{cloudTops} = { v => $4 * 100, u => 'M' };
            $r->{cloudTops}{q} = 'isEqualGreater' if $4 == 99;

            # WMO-No. 306 Vol I.1, Part A, Section B:
            # Ct description of the top of cloud whose base is below the level of the station. (code table 0552)
            $r->{cloudTopDescr} = $5;

            push @s4, { cloudBelowStation => $r };
        }

        $report{section4} = \@s4;
    }

=back

=for comment ###########################################################

=head3 Section 5: data for national use (optional)

(not implemented)

=for comment ###########################################################

    CA:      1ssss 2swswswsw 3dmdmfmfm 4fhftftfi
    US land: 0ittDtDtD RECORD* 1snTT snTxTxsnTnTn RECORD* 2R24R24R24R24 44snTwTw 9YYGG
    US sea:  11fff 22fff 3GGgg 4ddfmfm 9GGgg

=cut

    if ($country eq 'CA' && $msg =~ s@^555 @@) {         # MANOBS 12.5
        my (@s5, $r);

        @s5  = ();

        # group 1ssss
        if ($msg =~ s@^(1(?:(\d{4})|////)) @@) {
            $r->{s} = $1;
            $r->{period} = '1d';
            if (!defined $2) {
                $r->{noMeasurement} = undef;
            } elsif ($2 == 9999) {
                $r->{traces} = undef;
            } else {
                $r->{precipAmount} = { v => $2 / 10, u => 'CM' };
            }
            push @s5, { snowFall => $r };
            $r = undef;
        }

        # group 2swswswsw
        if ($msg =~ s@^(2(?:(\d{4})|////)) @@) {
            $r->{s} = $1;
            $r->{period} = '1d';
            if (!defined $2) {
                $r->{noMeasurement} = undef;
            } elsif ($2 == 9999) {
                $r->{traces} = undef;
            } else {
                $r->{precipAmount} = { v => $2 / 10, u => 'MM' };
            }
            push @s5, { waterEquivOfSnow => $r };
            $r = undef;
        }

        # group 3dmdmfmfm
        if ($msg =~ s@^(3(?:(\d\d|//)(\d\d)|////)) @@) {
            $r->{s} = $1;
            $r->{period} = '1d';
            if (!defined $2) {
                $r->{wind}{notAvailable} = undef;
            } else {
                if ($2 eq '//') {
                    $r->{wind}{dirNotAvailable} = undef;
                } else {
                    $r->{wind}{dir} = $2 * 10;
                }
                $r->{wind}{speed} = { v => $3, u => 'KT' };
            }
            push @s5, { highestWind => $r };
            $r = undef;

            # group 4fhftftfi
            if ($msg =~ s@^4([01/])($re_hour|//)([0-3/]) @@) {
                if ($1 eq '/') {
                    $r->{notAvailable} = undef;
                } else {
                    $r->{speed} = { v => 100, u => 'KT',
                                   $1 == 0 ? ( q => 'isLess' )
                                           : ( q => 'isEqualGreater') };
                }
                push @s5, { highestWind => {
                    s      => "4$1$2",
                    wind   => $r,
                    period => '1d',
                    $2 ne '//' ? (hour => $2, minute => '00') : ()
                }};
                $r = undef;

                if ($3 eq '/') {
                    $r->{notAvailable} = undef;
                } elsif ($3 == 0) {
                    $r->{speed} = { v => 17, u => 'KT', q => 'isLess' };
                } elsif ($3 == 1) {
                    $r->{speed} = { v => 17, rp => 11, u => 'KT' };
                } elsif ($3 == 2) {
                    $r->{speed} = { v => 28, rp => 6, u => 'KT' };
                } else {
                    $r->{speed} = { v => 34, u => 'KT',
                                    q => 'isEqualGreater' };
                }
                push @s5, { highestMeanSpeed => {
                    s             => $3,
                    period        => '1d',
                    measurePeriod => '2m',
                    wind          => $r,
                }};
                $r = undef;
            }
        }
        $report{section5} = \@s5;
    }

    if ($country eq 'US' && $msg =~ s@^555 @@) {         # FMH-2 7.
        my (@s5, $r);

        @s5  = ();

        if ($report{obsStationType}{stationType} eq 'AAXX') {
            # group RECORD
            while ($msg =~ s@^($re_record_temp) @@) {
                push @s5, { recordTemp => _getRecordTemp $1 };
            }

            # group 0ittDtDtD
            if ($msg =~ s@^(0(?:([134679])(\d{3})|([258])000|0///)) @@o) {
                $r->{s} = $1;
                if (!defined $2 && !defined $4) {
                    $r->{notAvailable} = undef;
                } else {
                    my $type = (($2 || $4) - 1) / 3;
                    $r->{tideType} = ('low', 'neither', 'high')[$type];
                    if (defined $2) {
                        $r->{tideDeviation} = { v => $3 + 0, u => 'FT' };
                        if ($2 % 3 == 1) {
                            $r->{tideDeviation}{v} *= -1;
                            $r->{tideLevel} = 'below';
                        } else {
                            $r->{tideLevel} = 'above';
                        }
                    } else {
                        $r->{tideDeviation} = { v => 0, u => 'FT' };
                        $r->{tideLevel} = 'equal';
                    }
                }
                push @s5, { tideData => $r };
                $r = undef;
            }

            # groups for CITY data: 1snTT snTxTxsnTnTn RECORD* 2R24R24R24R24
            if ($msg =~ s@^1([01])(\d\d) ([01])(\d\d)([01])(\d\d)((?: $re_record_temp)*) 2(\d{4}) @@o)
            {
                my $tempAirF;

                $tempAirF = $report{temperature}{air}{temp}{v} * 1.8 + 32
                    if    exists $report{temperature}
                       && exists $report{temperature}{air}
                       && exists $report{temperature}{air}{temp};

                # ... group 1snTT
                push @s5, { tempCity => _getTempCity("1", $1, $2, $tempAirF) };

                # ... group snTxTx
                $r = _getTempCity "", $3, $4, $tempAirF;
                if ($report{obsTime}{hour} == 0) {
                    $r->{period} = '12h';
                } elsif ($report{obsTime}{hour} == 12) {
                    $r->{period} = '1d';
                }
                push @s5, { tempCityMax => $r };

                # ... group snTnTn
                $r = _getTempCity "", $5, $6, $tempAirF;
                if ($report{obsTime}{hour} == 0) {
                    $r->{period} = '18h';
                } elsif ($report{obsTime}{hour} == 12) {
                    $r->{period} = '12h';
                }
                push @s5, { tempCityMin => $r };

                # ... group 2R24R24R24R24
                push @s5, { precipCity => {
                    s            => "2$8",
                    period       => '24h',
                    precipAmount => { v => $8 / 100, u => 'IN' }
                }};

                # ... group RECORD
                for ($7 =~ /$re_record_temp/og) {
                    push @s5, { recordTempCity => _getRecordTemp $_ };
                }
            }

            # group 44snTwTw
            if ($msg =~ s@^(44([01]\d\d)) @@o) {
                push @s5, { waterTemp => {
                    s    => $1,
                    temp => _parseTemp "$2/"
                }};
            }

            # group 9YYGG
            if ($msg =~ s@^(9($re_day)($re_hour)) @@o) {
                push @s5, { obsTime => {
                    s      => $1,
                    day    => $2,
                    hour   => $3,
                    minute => '00'
                }};
            }
        } elsif ($report{obsStationType}{stationType} eq 'BBXX') {
        }

        $report{section5} = \@s5;
    }

    push @{$report{warning}}, { warningType => 'notProcessed', s => $msg }
        if $msg ne '';

    return %report;
}

########################################################################
# _parseMetarTaf
########################################################################
sub _parseMetarTaf {
    my ($msg, $msg_hdr, $is_speci, $is_taf) = @_;
    my (%metar, $corr_msg);
    my ($day, $hour, $minute);
    my ($curr_day, $curr_hour, $curr_minute);
    my ($ii, $td, @cy, $delim, $tt, $is_taf_Aug2007, $is_auto);

    $metar{msg} = $msg_hdr . $msg;
    $metar{isSpeci} = undef if $is_speci;
    $metar{isTaf} = undef if $is_taf;

=pod

=head2 Parsing of METAR and TAF messages

The message is checked for typical errors and corrected. Errors for METARs could
be:

=over

=item *

invalid characters (everything except capital letters, digits, dot, slash,
dollar sign, space, plus sign, minus sign)

=item *

QNH with spaces

=item *

temperature or dew point with spaces

=item *

misspelled key words, or with missing or additional spaces

=item *

missing key words

=item *

wrong order of keywords

=item *

removal of slashes before and after some components

=back

If the message is modified there will be a C<warning>.

=cut

    $msg .= ' '; # this makes parsing much easier

    # EXTENSION: preprocessing
    # remove trailing =
    $msg =~ s/= $/ /s;

    # replace invalid characters
    $msg =~ s/[^A-Z\d.,\/\$ +-]/?/gos;

    # QNH with spaces
    $msg =~ s/(?<= A) (?=[23]\d{3} )//;
    $msg =~ s/(?<= Q) (?=[01]\d{3} )//;
    $msg =~ s/(?<= (?:A[23]|Q[01])) (?=\d{3} )//;
    $msg =~ s/(?<= (?:A[23]|Q[01])\d) (?=\d\d )//;
    $msg =~ s/(?<= (?:A[23]|Q[01])\d\d) (?=\d )//;

    # misspelled key words, or with missing or additional spaces
    $msg =~ s/ (?:CAMOK|CC?VOK|CAVOC) / CAVOK /;
    $msg =~ s/ (?:NO(?: S)?IG|NOSI(?: G)?|N[L ]OSIG|NSI?G|(?:MO|N0)SIG|NOS I ?G) / NOSIG /;
    $msg =~ s/(?<! )(?=(?:TEMPO|BECMG) )/ /; # BLU+BLU+(TEMPO|BECMG)
    $msg =~ s/ (?:R MK|RMKS|RRMK)\.? / RMK /;
    $msg =~ s@(?<= RMK)(?=[^ /])@ @;
    $msg =~ s@ 0VC(?=$re_cloud_base$re_cloud_type?|$re_cloud_type )@ OVC@og;
    $msg =~ s@ BNK(?=$re_cloud_base$re_cloud_type?|$re_cloud_type )@ BKN@og;
    $msg =~ s@( $re_cloud_type| FULYR)[/-](?=$re_compass_dir16 )@$1 @og;
    $msg =~ s@(?<= VR) (?=B$re_wind_speed$re_wind_speed_unit )@@og;
    $msg =~ s@(?<= R) (?=WY)@@g;
    $msg =~ s@(?<= RW) (?=Y)@@g;
    $msg =~ s@(?<= RWY$re_rwy) (?=[LCR] )@@og;
    $msg =~ s@(?<= RWY$re_rwy) (?=(?:LL|RR) )@@og;
    $msg =~ s@ RNW @ RWY @g;
    $msg =~ s@ ?/(?=AURBO )@ @;
    $msg =~ s@(?<= A)0(?=[12]A? )@O@;
    $msg =~ s@(?<= ALQ)(?=S )@D@g;
    $msg =~ s@(?<= O)VH(?= )@HD@g;
    $msg =~ s@(?<= OCNL)Y(?= )@@g;
    $msg =~ s@ LIGHTNING @ LTG @g;
    $msg =~ s@ LTNG @ LTG @g;
    $msg =~ s@( $re_phen_desc L)GT(?=$re_ltg_types+)@$1TG@og;
    $msg =~ s@(?<= LTG) (?=$re_ltg_types{2,})@@og;
    $msg =~ s@(?<= LTG) (?=(?:C[ACGW])+)@@g; # IC could be weather
    $msg =~ s@(?<= I)N(?=OVC )@@g;
    $msg =~ s@(?<= SP)O(?=TS )@@g;
    $msg =~ s@(?<= W) (?=IND )@@g;
    $msg =~ s@(?<= WIND RWY) (?=$re_rwy_des )@@og;
    $msg =~ s@(?<= (?:RWY|THR))($re_rwy_des ${re_wind_dir}0$re_wind_speed) (?=$re_wind_speed_unit )@$1@og;
    $msg =~ s@(?<= N)ORTH(?= )@@g;
    $msg =~ s@(?<= S)OUTH(?= )@@g;
    $msg =~ s@(?<= E)AST(?= )@@g;
    $msg =~ s@(?<= W)EST(?= )@@g;
    $msg =~ s@(?<= B)A(?=NK )@@g;
    $msg =~ s@(?<= F)(?= BNK )@G@g;
    $msg =~ s@( [+-]?)(RA|SN)SH @$1SH$2 @g;
    $msg =~ s@(?<= ISOL)D(?= )@@g;
    $msg =~ s@(?<= H)A(?=ZY )@@g;
    $msg =~ s@(?<= DS)(?=T )@N@g;
    $msg =~ s@(?<= D)IS(?=T )@SN@g;
    $msg =~ s@(?<= PL)(?=ME )@U@g;
    $msg =~ s@(?<= BL)K @ACK@g;
    $msg =~ s@(?<= INVIS)T(?= )@@g;
    $msg =~ s@(?<= FROIN)/ ?@ @g;
    $msg =~ s@(?<= AS)(?=CTD )@O@g;
    $msg =~ s@(?<= EMBD)D(?= )@@g;
    $msg =~ s@(?<=^[A-Z][A-Z\d]{3} $re_day$re_hour$re_min)(?= )@Z@o
        unless $is_taf;
    $msg =~ s@(?<=^[A-Z][A-Z\d]{3} $re_day$re_hour$re_min)KT(?= )@Z@o;
    $msg =~ s@(?<=^[A-Z][A-Z\d]{3} $re_day$re_hour${re_min}Z )(${re_wind_dir}0)/(?=${re_wind_speed}$re_wind_speed_unit )@$1@o;
    $msg =~ s@(?<= [0-2]) (?=\d0${re_wind_speed}$re_wind_speed_unit )@@og;
    $msg =~ s@(?<= 3) (?=[0-6]0${re_wind_speed}$re_wind_speed_unit )@@og;
    $msg =~ s@(?<=M)/(?=S )@P@;
    $msg =~ s@(?<=[+]) (?=$re_weather_w_i )@@o;
    $msg =~ s@(?<=$re_cloud_base )(M?\d\d/)/(?=M?\d\d )@$1@o;
    $msg =~ s@ TMPO @ TEMPO @g;
    $msg =~ s@(?<= PR)0(?=B[34]0 )@O@g;
    $msg =~ s@(?<= BECMG| TEMPO)(?=$re_hour(?:$re_hour|24) )@ @og;
    $msg =~ s@(?<= (?:BECMG|TEMPO|INTER) $re_day$re_hour)(?: /|/ )(?=$re_day(?:$re_hour|24) )@/@og
        if $is_taf;
    $msg =~ s@ PRECIP @ PCPN @;
    $msg =~ s@( T[XN]M?\d\d/(?:$re_hour|24)Z)(?=T[XN]M?\d\d/(?:$re_hour|24)Z )@$1 @o;
    $msg =~ s@/($re_snw_cvr_title)@ $1@o;
    $msg =~ s@(?<= RWY0) (?=[1-9] )@@g;
    $msg =~ s@(?<= RWY[12]) (?=\d )@@g;
    $msg =~ s@(?<= RWY3) (?=[0-6] )@@g;
    $msg =~ s@ DRFTG SNW @ DRSN @g;
    $msg =~ s@(?<= T)O(?=\d{3}[01]\d{3} )@0@g;
    $msg =~ s@/(?=BLN VISBL )@ @g;

    # NSW should not be in initial TAF section, only in trends
    if ($is_taf) {
        my $idx = index $msg, ' NSW ';

        $msg =~ s@NSW @@
            unless    $idx > -1
               && substr($msg, 0, $idx) =~ / FM\d| (?:BECMG|TEMPO|INTER|PROB[34]0) /;
    }

    # missing key words
    $msg =~ s@ (?:TEMPO|INTER) ($re_hour$re_min)/($re_hour$re_min)(?= )@ TEMPO FM$1 TL$2@og
        unless $is_taf;
    $msg =~ s@(?<! (?:BECMG|TEMPO) )(?=FM$re_hour$re_min )@BECMG @og
        unless $is_taf;

    # wrong order of keywords
    $msg =~ s@( DSNT| DISTANT)( (?:$re_weather|SMOKE|HAZE|TSTMS|$re_cloud_type|LTG)(?: TO)?)(?=$re_phen_locdirs )@$2$1@og;
    $msg =~ s@( $re_weather| SMOKE| HAZE| TSTMS| $re_cloud_type| LTG)((?: TO)?$re_phen_locdirs|$re_num_quadrant) (DSNT|DISTANT) @$1 $3$2 @og;

    # komma removed
    $msg =~ s@^((?:U|ZM).* RMK.*QFE\d{3}) (?=\d )@$1,@;
    $msg =~ s@^(FQ.* RMK.*TX/\d\d) (?=\d )@$1,@;

    $corr_msg = $msg;
    if ($metar{msg} . ' ' ne $msg_hdr . $msg) {
        # we appended a blank, don't show it in warning
        my $msg_mod;
        $msg_mod = "$msg_hdr$msg";
        $msg_mod =~ s/ $//;
        push @{$metar{warning}}, { warningType => 'msgModified', s => $msg_mod};
    }

=pod

Then the components of the message are checked starting with the observation
station and the observation time (or issue time for TAF).

=cut

    push @{$metar{reportModifier}}, { s => $1, modifierType => $1 }
        if $msg =~ s/^(COR|AMD) //;

    if ($msg !~ s/^([A-Z][A-Z\d]{3}) //) {
        $metar{ERROR} = _makeErrorMsg 'obsStation', $corr_msg, $msg;
        return %metar;
    }
    $metar{obsStationId}{id} = $1;
    $metar{obsStationId}{s} = $1;

    @cy = $metar{obsStationId}{id} =~ /(((.).)..)/;
    $cy[2] = 'cCA' if _cyInString \@cy, ' C ';
    $cy[2] = 'cJP' if _cyInString \@cy, ' RJ RO ';
    $cy[2] = 'cUS' if _cyInString \@cy, ' FHAW FJDG K MUGM NSTU PA PFYU PG PH '
                                 . 'PJON PK PLPA PMDY POLI PPIZ PT PWAK TJ TI ';

    $tt = $is_taf ? 'issueTime' : 'obsTime';
    if ($msg =~ s/^($re_day)($re_hour)($re_min)Z //o) {
        $metar{$tt}{s} = "$1$2$3Z";
        @{$metar{$tt}}{'day', 'hour', 'minute'} = ($1, $2, $3);
    } elsif ($msg =~ s/^($re_day$re_hour${re_min}KT) //o) {
        $metar{$tt} = { s => $1, invalidFormat => $1 };
    } elsif ($is_taf && $msg =~ /^$re_day$re_hour$re_min /o) {
        $metar{issueTime}{s} = '';
        $metar{issueTime}{invalidFormat} = '';
    } elsif ($msg =~ /^(NIL) $/) {
        # EXTENSION: NIL instead of issueTime
        push @{$metar{reportModifier}}, { s => $1, modifierType => $1 };
        return %metar;
    } else {
        $metar{ERROR} = _makeErrorMsg 'obsTime', $corr_msg, $msg;
        return %metar;
    }

=pod

For METAR messages the next component may optionally be the report modifier to
indicate a missing message (C<NIL>), a message created by an automated station
(C<AUTO>), or an corrected (C<COR>) or retarded (C<RTD>) message. For Canada,
the component BBB may appear to indicate that the report has been retarded
(C<RR?>), corrected (C<CC?>), amended (C<AA?>), or segmented (C<P??>).

=cut

    # EXTENSION: BBB (for Canada)
    while (!$is_taf && $msg =~ s/^(NIL|AUTO|COR|RTD|$re_bbb) //o) {
        my $r;

        $r->{s} = $1;
        if ($r->{s} =~ 'NIL|AUTO|COR|RTD') {
            $r->{modifierType} = $r->{s};
            $is_auto = $r->{modifierType} eq 'AUTO';
        } else {
            $r->{s} =~ '(.)(.)(.)';
            if ($1 eq 'P') {
                $r->{modifierType} = $1;
                $r->{segment} = "$2$3";
                $r->{isLastSegment} = undef if $2 eq 'Z';
            } else {
                $r->{modifierType} = "$1$2";
                if ($3 eq 'Z') {
                    $r->{over24hLate}  = undef;
                } elsif ($3 eq 'Y') {
                    $r->{sequenceLost} = undef;
                } else {
                    $r->{bulletinSeq} = $3;
                }
            }
        }
        push @{$metar{reportModifier}}, $r;
        return %metar if $r->{modifierType} eq 'NIL';
    }

=pod

For TAF messages there may follow the modifier C<NIL> (which terminates the
message), C<COR> (corrected), or C<AMD> (amended). Then follows the forecast
period and optionally C<CNL> to indicate a cancelled forecast.

=cut

    if ($is_taf) {
        if ($msg =~ s/^(NIL|COR|AMD) //) {
            push @{$metar{reportModifier}}, { s => $1, modifierType => $1 };
            return %metar if $1 eq 'NIL';
        }

        if ($msg =~ s/^($re_day)($re_hour)($re_hour|24) //o) {
            $metar{fcstPeriod}{s} = "$1$2$3";
            @{$metar{fcstPeriod}}{'day', 'hourFrom', 'hourTill'} = ($1, $2, $3);
        } elsif ($msg =~ s@^($re_day)($re_hour)/($re_day)($re_hour|24) @@o) {
            $metar{fcstPeriod}{s} = "$1$2/$3$4";
            @{$metar{fcstPeriod}}{'dayFrom', 'hourFrom', 'dayTill', 'hourTill'}
                                                             = ($1, $2, $3, $4);
            $is_taf_Aug2007 = 1;
        } else {
            $metar{ERROR} = _makeErrorMsg 'fcstPeriod', $corr_msg, $msg;
            return %metar;
        }

        if ($msg =~ s/^(CNL) //) {
            $metar{fcstCancelled}{s} = "$1";
            return %metar;
        }

        # EXTENSION: NIL after fcstPeriod
        if ($msg =~ s/^(NIL) //) {
            push @{$metar{reportModifier}}, { s => $1, modifierType => $1 };
            return %metar;
        }

        # EXTENSION: NOT AVBL after fcstPeriod. Remarks may follow
        if ($msg =~ s/^(FCST NOT AVBL DUE NO OBS) //) {
            $metar{fcstNotAvbl}{s} = $1;
            $metar{fcstNotAvbl}{fcstNotAvblReason} = 'NOOBS';
        }
        if ($msg =~ s/^(FCST NOT AVBL DUE INSUFFICIENT OBS) //) {
            $metar{fcstNotAvbl}{s} = $1;
            $metar{fcstNotAvbl}{fcstNotAvblReason} = 'INSUFFICIENTOBS';
        }
    }

=pod

The first weather group is the surface wind. It may include a gust speed; the
wind speed and/or the wind direction may also be missing. There may be a second
group to report a varying wind direction.

=cut

    # EXTENSION: allow ///// (wind not available)
    # EXTENSION: allow /// for wind direction (not available)
    # EXTENSION: allow // for wind speed (not available)
    if ($msg =~ s@^(/////|(?:$re_wind)) @@o) {
        $metar{sfcWind}{s} = $1;
        $metar{sfcWind}{wind} = _parseWind $1, 1;
    # EXTENSION: allow invalid formats
    } elsif ($msg =~ s@^(${re_wind_dir}0$re_wind_speed|\d{6}$re_wind_speed_unit) @@o)
    {
        $metar{sfcWind}{s} = $1;
        $metar{sfcWind}{wind}{invalidFormat} = $1;
    # EXTENSION: allow missing wind
    } else {
        push @{$metar{warning}}, { warningType => 'windMissing' };
    }

    # EXTENSION:
    # Annex 3 Appendix 3 4.1.4.2.b.1 requires wind speed >=3 kt and
    #   variation 60-180 for this group: not checked
    if ($msg =~ s/^($re_wind_dir\d)V($re_wind_dir\d) //o) {
        if (exists $metar{sfcWind}) {
            $metar{sfcWind}{s} .= ' ';
        } else {
            # TODO?: actually not "not available" but missing
            $metar{sfcWind}{wind}{dirNotAvailable} = undef;
            $metar{sfcWind}{wind}{speedNotAvailable} = undef;
            $metar{sfcWind}{s} = '';
        }
        $metar{sfcWind}{wind}{windVarLeft} = $1 + 0;
        $metar{sfcWind}{wind}{windVarRight} = $2 + 0;
        $metar{sfcWind}{s} .= "$1V$2";
    }

    while ($is_taf && $msg =~ s@^($re_wind_shear_lvl) @@o) {
        push @{$metar{TAFsupplArr}}, { windShearLvl => {
            s     => $1,
            level => $2 + 0,
            wind  => _parseWind $3
        }};
    }

=pod

If the weather conditions allow (visbility more than 10 km, no relevant weather,
no clouds below 5000 ft) this may be indicated by the keyword C<CAVOK>. The next
component after that should be the temperature (see below).

=cut

    # Annex 3 Appendix 3 2.2:
    #   vis >10km, no relevant weather, no cloud below 5000 ft
    if ($msg =~ s/^CAVOK //) {
        $metar{CAVOK} = undef; # cancels visibility, weather and cloud
    }

    if (!exists $metar{CAVOK}) {

=pod

Otherwise, the prevailing visibilty follows. It can have a compass direction
attached or C<NDV> if no directional variations can be given. There may be an
additional group for the minimum visibilty.

=cut

        # EXTENSION: allow 16 compass directions
        # EXTENSION: allow 'M' after re_vis_m
        # EXTENSION: allow P6000 (UERP ULOO UMKK UMOO URSS UTAK)
        if (   $msg =~ /^((?:(P)?($re_vis_m)M?(NDV)?|($re_vis_km))($re_compass_dir16)?)(?: ($re_vis_m)($re_compass_dir))? (.*)/o
            && !(defined $2 && $3 == 9999))
        {
            $msg = $9;
            $metar{visPrev}{s} = $1;
            $metar{visPrev}{compassDir} = $6 if defined $6;
            if ($7) {
                $metar{visMin}{s} = $7 . $8;
                $metar{visMin}{distance} = _getVisibilityM $7;
                $metar{visMin}{compassDir} = $8;
            }
            if (defined $3) {
                $metar{visPrev}{distance} = _getVisibilityM $3, $2;
                $metar{visPrev}{NDV}      = undef if defined $4;
            } else {
                $metar{visPrev}{distance} = { v => $5, rp => 1, u => 'KM' };
                $metar{visPrev}{distance}{v} =~ s/KM//;
            }
        } elsif ($msg =~ s@^($re_vis_sm)SM @@o) {
            $metar{visPrev} = _getVisibilitySM $1, $is_auto, \@cy;
            $metar{visPrev}{s} = "$1SM";
        } elsif ($is_taf && $msg =~ s@^(P?[1-9]\d*)SM @@) {
            $metar{visPrev} = _getVisibilitySM $1, $is_auto, \@cy;
            $metar{visPrev}{s} = "$1SM";
        # EXTENSION: allow //// and misformatted entry
        } elsif ($msg =~ s@^//// @@) {
            $metar{visPrev}{notAvailable} = undef;
            $metar{visPrev}{s} = '////';
        } elsif ($msg =~ s@^(\d{3})(?: ($re_vis_m)($re_compass_dir))? @@o) {
            $metar{visPrev}{invalidFormat} = $1;
            $metar{visPrev}{s} = $1;
            if ($2) {
                $metar{visMin}{s} = $2 . $3;
                $metar{visMin}{distance} = _getVisibilityM $2;
                $metar{visMin}{compassDir} = $3;
            }
        } else {
            # EXTENSION: allow missing visibility even without CAVOK
            push @{$metar{warning}}, { warningType => 'visibilityMissing' };
        }
    }

=pod

After that may follow one or more runway visibilty ranges or C<RVRNO> if they
are not available.

=cut

    while (_parseRwyVis \%metar, \$msg) {};

    # EXTENSION: allow RVRNO (not available) (KADW, RKSG, PAED)
    while ($msg =~ s@^RVRNO @@) {
        $metar{RVRNO} = undef;
    }

    if (!exists $metar{CAVOK}) {

=pod

Then there can be up to 3 groups to describe the present weather.

=cut

        # EXTENSION: allow // (not available)
        while ($msg =~ s@^($re_weather|//|([+-]?(?:RATS|BRHZ|RESH|RETS|VC|VCRA|RABR))) @@o) {
            if (defined $2) {
                push @{$metar{weather}}, {
                    s             => $2,
                    invalidFormat => $2
                };
            } else {
                push @{$metar{weather}}, _parseWeather $1;
            }
        }

=pod

Then follow up to 3 groups for the sky condition (cloud cover and base or
vertical visibilty) optionally with cloud type. The keywords C<CLR>, C<SKC>,
C<NSC>, or C<NCD> may indicate different sky conditions if no cloud cover is
given.

=cut

        # EXTENSION: allow ////// (not available) without CB, TCU
        # WMO-No. 306 Vol I.1, Part A, Section A, 15.9.1.1:
        #   CLR: no clouds below 10000 (FMH-1: 12000) ft detected by autom. st.
        #   SKC: no clouds + VV not restricted but not CAVOK
        #   NSC: no significant clouds + no CB + VV not restr. but not CAVOK,SKC
        #   NCD: no clouds + CB, TCU detected by automatic observation system
        while ($msg =~ s@^((SKC|NSC|CLR|NCD)|VV($re_cloud_base)|(/{6}|$re_cloud_cov(?:$re_cloud_base(?: ?$re_cloud_type)?|$re_cloud_type))|($re_cloud_cov\d{1,2}|$re_cloud_cov(?: \d{1,3})?)) @@o)
        {
            if (defined $5) {
                push @{$metar{cloud}}, {
                    s             => $1,
                    invalidFormat => $1
                };
            } elsif (defined $2) {
                push @{$metar{cloud}}, {
                    s        => $1,
                    noClouds => $2
                };
            } elsif (defined $3) {
                $metar{visVert}{s} = $1;
                if ($3 eq '///') {
                    $metar{visVert}{notAvailable} = undef;
                } else {
                    $metar{visVert}{distance} = { v => $3 * 100, u => 'FT' };
                }
            } else {
                push @{$metar{cloud}}, _parseCloud $4;
            }
        }
        _determineCeiling $metar{cloud} if exists $metar{cloud};
    }

=pod

The following group contains the current air temperature and optionally the dew
point. If both are given the humidity can be determined.

=cut

    # EXTENSION: FMH-1: dew point is optional
    # EXTENSION: allow // for temperature and dew point
    # EXTENSION: allow XX for temperature and dew point
    if (!$is_taf && $msg =~ s@^((?:(M)?(\d ?\d)|(?://|XX))/((M)? ?(\d ?\d)|(?://|XX))?) @@)
    {
        $metar{temperature}{s} = $1;
        if (defined $3) {
            $metar{temperature}{air}{temp} = { v => $3, u => 'C' };
            $metar{temperature}{air}{temp}{v} =~ tr / //d;
            $metar{temperature}{air}{temp}{v} += 0;
            $metar{temperature}{air}{temp}{v} *= -1 if defined $2;
        } else {
            $metar{temperature}{air}{notAvailable} = undef;
        }
        if (defined $4) {
            if (defined $6) {
                $metar{temperature}{dewpoint}{temp} = { v => $6, u => 'C' };
                $metar{temperature}{dewpoint}{temp}{v} =~ tr / //d;
                $metar{temperature}{dewpoint}{temp}{v} += 0;
                $metar{temperature}{dewpoint}{temp}{v} *= -1 if defined $5;
            } else {
                $metar{temperature}{dewpoint}{notAvailable} = undef;
            }
        }
        if (   exists $metar{temperature}{air}
            && exists $metar{temperature}{air}{temp}
            && exists $metar{temperature}{dewpoint}
            && exists $metar{temperature}{dewpoint}{temp})
        {
            @{$metar{temperature}}{'relHumid1', 'relHumid2',
                                   'relHumid3', 'relHumid4'}
                = _getHumidity $metar{temperature}{air}{temp}{v},
                               $metar{temperature}{dewpoint}{temp}{v};
        }
    } else {
        push @{$metar{warning}}, { warningType => 'tempMissing'} unless $is_taf;
    }

=pod

The last regular item is the QNH given in hectopascal or in. Hg.

=cut

    if ($msg =~ s@^($re_qnh) @@o) {
        $metar{QNH} = _parseQNH $1;
    }
    # EXTENSION: allow Altimeter after QNH
    if (   _cyInString(\@cy, ' MN MS MT MZ OI OM RP VT ')
        && $msg =~ s@^($re_qnh) @@o)
    {
        my $qnh;

        $qnh = _parseQNH $1;
        for (keys %$qnh) {
            if ($_ eq 's') {
                $metar{QNH}{s} .= ' '. $qnh->{$_};
            } else {
                $metar{QNH}{$_} .= $qnh->{$_};
            }
        }
    }
    push @{$metar{warning}}, { warningType => 'QNHMissing' }
        unless $is_taf || exists $metar{QNH};

=pod

After that may follow country or station specific information: pressure, worst
cloud cover. There may also be groups indicating recent weather, wind shear,
runway conditions and runway winds.

=cut

    # EXTENSION: allow some (sea level?) pressure (OIxx, OPxx)
    if ($cy[1] =~ /O[IP]/ && $msg =~ s@^([789]\d\d[./]\d) @@) {
        $metar{somePressure}{s} = $1;
        $metar{somePressure}{hPa} = $1;
        $metar{somePressure}{hPa} =~ s@/@.@;
    }

    # EXTENSION: allow worst cloud cover
    if (_cyInString(\@cy, ' LO ') && $msg =~ s@^($re_cloud_cov|SKC) @@o) {
        $metar{cloudMaxCover}{s} = $1;
        if ($1 eq 'SKC') {
            $metar{cloudMaxCover}{SKC} = undef;
        } else {
            $metar{cloudMaxCover}{cloudCover} = $1;
        }
    }

    while ($msg =~ s@^RE($re_weather_re) @@o) {
        push @{$metar{recentWeather}}, _parseWeather($1, 'RE');
    }

    if ($msg =~ s/^(WS ALL RWY) //) {
        push @{$metar{windShear}}, {
            s           => $1,
            rwyDesigAll => undef
        };
    }
    while ($msg =~ s@^(WS RWY($re_rwy_des)) @@o) {
        push @{$metar{windShear}}, {
            s        => $1,
            rwyDesig => $2
        };
    }

    # sea surface temperature, state of the sea WTsTs/SS'
    # TODO

    while (_parseRwyState \%metar, \$msg) {};

    # EXTENSION: allow colour code
    if ($msg =~ s@^($re_colour) @@o) {
        $metar{colourCode} = _parseColourCode $1;
    }

    # EXTENSION: country specific things
    if (   _cyInString(\@cy, ' SCSE ')
        && $msg =~ s@^(NEFO PLAYA (?:([1-9]\d+0)FT|(SKC))) @@)
    {
        $metar{NEFO_PLAYA}{s} = $1;
        if (defined $2) {
            $metar{NEFO_PLAYA}{cloudBase} = { v => $2 + 0, u => 'FT'};
        } else {
            $metar{NEFO_PLAYA}{SKC} = undef;
        }
    }

    # EXTENSION:
    while (_cyInString(\@cy, ' LG ') && $msg =~ s@^($re_rwy_wind) @@o) {
        my $r;

        $r->{s} = $1;
        $r->{wind} = _parseWind $4, 1;
        $r->{wind}{windVarLeft} = $5 + 0 if defined $5;
        $r->{wind}{windVarRight} = $6 + 0 if defined $6;
        $r->{rwyDesig} = defined $2 ? $2 : $3;
        push @{$metar{rwyWind}}, $r;
    }

    # EXTENSION:
    while (_cyInString(\@cy, ' LPMA ') && $msg =~ s@^($re_rs_rwy_wind) @@o) {
        my $r;
        $r->{s} = $1;
        $r->{wind} = _parseWind $3 . '0' . $4, 1;
        if ($2 eq 'RS') {
            $r->{windLocation} = $2;
            $metar{windAtLoc} = $r;
        } else {
            $r->{rwyDesig} = $2;
            push @{$metar{rwyWind}}, $r;
        }
    }

=pod

Finally, there may be trends and remarks.

=cut

    if ($msg =~ s/^(NOSIG) //) {
        my $td;

        $td->{s} = $1;
        $td->{trendType} = $1;
        push @{$metar{trend}}, $td;

        # EXTENSION: allow rwyState after NOSIG. Not considered trend data!
        while (_parseRwyState \%metar, \$msg) {};
    }

    # EXTENSION: allow RH after NOSIG. Not considered trend data!
    if (_cyInString(\@cy, ' OOSA ') && $msg =~ s@^(RH(\d\d)) @@) {
        $metar{RH}{s} = $1;
        $metar{RH}{relHumid} = $2 + 0;
    }

    # "supplementary" section of TAFs
    while ($is_taf && _parseTAFsuppl \%metar, \$msg, \%metar, \@cy) {};

    # additional info in TAFs
    while ($is_taf && _parseTAFinfo \%metar, \$msg, $is_taf_Aug2007) {};

    # changes at midnight: 0000 with FM/AT, 2400 with TL
    # EXTENSION: allow missing period
    # EXTENSION: FM: allow time with Z
    # EXTENSION: FM: allow time without minutes
    # EXTENSION: FM: allow 24Z? or 2400Z?
    while (   (!$is_taf && $msg =~ s/^(BECMG|TEMPO) //)
           || ($is_taf && !$is_taf_Aug2007 && $msg =~ s@^(FM(?:($re_hour)($re_min)?|(24)(?:00)?)Z?|(?:(BECMG|TEMPO|INTER)|(?:PROB([34]0)(?: (TEMPO|INTER))?))(?: ($re_hour|24)($re_hour|24))?) @@o)
           || ($is_taf_Aug2007 && $msg =~ s@^(FM(?:($re_day)($re_hour)($re_min))|(?:(BECMG|TEMPO|INTER)|(?:PROB([34]0)(?: (TEMPO|INTER))?)) ($re_day)($re_hour)/($re_day)($re_hour|24)) @@o))
    {
        my $td;

        if ($is_taf && !$is_taf_Aug2007) {
            $td->{s} = $1;
            if (defined $2) {
                $td->{trendType} = 'BECMG';
                @{$td->{trendTime1}}{'timeSpec', 'hour'} = ('FM', $2);
                $td->{trendTime1}{minute} = defined $3 ? $3 : '00';
            } elsif (defined $4) {
                $td->{trendType} = 'BECMG';
                @{$td->{trendTime1}}{'timeSpec', 'hour', 'minute'}
                                                             = ('FM', $4, '00');
            } else {
                if (defined $5) {
                    $td->{trendType} = $5;
                } else {
                    $td->{probability} = $6;
                    $td->{trendType} = defined $7 ? $7 : 'PROB';
                }
                if (defined $8) {
                    @{$td->{trendTime1}}{'timeSpec', 'hour', 'minute'}
                                                             = ('FM', $8, '00');
                    @{$td->{trendTime2}}{'timeSpec', 'hour', 'minute'}
                                                             = ('TL', $9, '00');
                } else {
                    push @{$metar{warning}},
                                    { warningType => 'periodMissing', s => $1 };
                }
            }
        } elsif ($is_taf_Aug2007) {
            $td->{s} = $1;
            if (defined $2) {
                $td->{trendType} = 'BECMG';
                @{$td->{trendTime1}}{'timeSpec', 'day', 'hour', 'minute'}
                                                           = ('FM', $2, $3, $4);
            } else {
                if (defined $5) {
                    $td->{trendType} = $5;
                } else {
                    $td->{probability} = $6;
                    $td->{trendType} = defined $7 ? $7 : 'PROB';
                }
                @{$td->{trendTime1}}{'timeSpec', 'day', 'hour', 'minute'}
                                                         = ('FM', $8, $9, '00');
                @{$td->{trendTime2}}{'timeSpec', 'day', 'hour', 'minute'}
                                                       = ('TL', $10, $11, '00');
            }
        } else {
            $td->{s} = $1;
            $td->{trendType} = $1;

            if ($msg =~ s@^((?:TL2400|(?:FM|TL|AT)$re_hour$re_min)Z?) @@o) {
                $td->{s} .= " $1";
                @{$td->{trendTime1}}{'timeSpec', 'hour', 'minute'}
                                                         = $1 =~ /(..)(..)(..)/;

                if ($msg =~ s@^TL(2400Z?|$re_hour${re_min}Z?) @@o) {
                    $td->{s} .= " TL$1";
                    @{$td->{trendTime2}}{'timeSpec', 'hour', 'minute'}
                                                     = ('TL', $1 =~ /(..)(..)/);
                }
            }
        }

        if ($msg =~ s@^($re_wind) @@o) {
            $td->{sfcWind}{s} = $1;
            $td->{sfcWind}{wind} = _parseWind $1, 1;
        # EXTENSION: allow invalid formats
        } elsif ($msg =~ s@^(${re_wind_dir}0$re_wind_speed|\d{6}$re_wind_speed_unit) @@o)
        {
            $td->{sfcWind}{s} = $1;
            $td->{sfcWind}{wind}{invalidFormat} = $1;
        }

        # EXTENSION: allow variation
        if ($msg =~ s/^($re_wind_dir\d)V($re_wind_dir\d) //o) {
            if (exists $td->{sfcWind}) {
                $td->{sfcWind}{s} .= ' ';
            } else {
                # TODO?: actually not "not available" but missing
                $td->{sfcWind}{wind}{dirNotAvailable} = undef;
                $td->{sfcWind}{wind}{speedNotAvailable} = undef;
                $td->{sfcWind}{s} = '';
            }
            $td->{sfcWind}{wind}{windVarLeft} = $1 + 0;
            $td->{sfcWind}{wind}{windVarRight} = $2 + 0;
            $td->{sfcWind}{s} .= "$1V$2";
        }

        if ($msg =~ s/^CAVOK //) {
            $td->{CAVOK} = undef;
        }

        if (!exists $td->{CAVOK}) {
            # EXTENSION: allow 'M' after re_vis_m
            if ($msg =~ s@^(($re_vis_m)M?|($re_vis_km)|($re_vis_sm)SM) @@o) {
                if (defined $2) {
                    $td->{visPrev}{distance} = _getVisibilityM $2;
                } elsif (defined $3) {
                    $td->{visPrev}{distance} = { v => $3, rp => 1, u => 'KM' };
                    $td->{visPrev}{distance}{v} =~ s/KM//;
                } else {
                    $td->{visPrev} = _getVisibilitySM $4, $is_auto, \@cy;
                }
                $td->{visPrev}{s} = $1;
            } elsif ($is_taf && $msg =~ s@^(P?[1-9]\d*)SM @@) {
                $td->{visPrev} = _getVisibilitySM $1, $is_auto, \@cy;
                $td->{visPrev}{s} = "$1SM";
            } elsif ($msg =~ s@^(\d{3}) @@) {
                $td->{visPrev}{s} = $1;
                $td->{visPrev}{invalidFormat} = $1;
            }

            # EXTENSION: allow VC..
            while ($msg =~ s@^($re_weather|NSW) @@o) {
                push @{$td->{weather}}, _parseWeather $1;
            }

            $ii = 0;
            while ($msg =~ s@^((SKC|NSC)|VV($re_cloud_base)|($re_cloud_cov$re_cloud_base$re_cloud_type?|$re_cloud_cov$re_cloud_type)) @@o)
            {
                if (defined $2) {
                    push @{$td->{cloud}}, {
                        s        => $1,
                        noClouds => $2
                    };
                } elsif (defined $3) {
                    $td->{visVert}{s} = $1;
                    if ($3 eq '///') {
                        $td->{visVert}{notAvailable} = undef;
                    } else {
                        $td->{visVert}{distance} = { v => $3 * 100, u => 'FT' };
                    }
                } else {
                    push @{$td->{cloud}}, _parseCloud $4;
                }
            }
            _determineCeiling $td->{cloud} if exists $td->{cloud};

            # EXTENSION: allow colour code
            if ($msg =~ s@^($re_colour) @@o) {
                $td->{colourCode} = _parseColourCode $1;
            }
        }

        # EXTENSION: allow rwyState after trend BECMG / TEMPO
        while (_parseRwyState $td, \$msg) {};

        # "supplementary" section of TAFs
        while ($is_taf && _parseTAFsuppl $td, \$msg, \%metar, \@cy) {};

        # additional info in TAFs
        while ($is_taf && _parseTAFinfo \%metar, \$msg, $is_taf_Aug2007) {};

        push @{$metar{trend}}, $td;
    }

    if ($msg =~ s@^RMK[ /] ?@@) {
        my ($notRecognised, $parsed);

        @{$metar{remark}} = ();

        $notRecognised = '';
        while ($msg ne '') {
            my $r;

            $parsed = 1;
            $r = {};
            if (   $notRecognised =~ /DUE(?: TO)?(?: $re_phen_desc| GROUND)?$/o
                || $notRecognised =~ /BY(?: $re_phen_desc| GROUND)?$/o)
            {
                $parsed = 0;
            } elsif (_parseRwyState $r, \$msg) {
                push @{$metar{remark}}, { rwyState => $r->{rwyState}[0] };
            } elsif (_parseRwyVis $r, \$msg) {
                push @{$metar{remark}}, { visRwy => $r->{visRwy}[0] };
            } elsif (   _cyInString(\@cy, ' LI OA ')
                     && $msg =~ s@^($re_cloud_cov) @@o)
            {
                push @{$metar{remark}}, { cloudMaxCover => {
                    s          => $1,
                    cloudCover => $1
                }};
            } elsif (   _cyInString(\@cy, ' BKPR EG EQ ET FH L OA PGUA ')
                     && $msg =~ s@^($re_colour) @@o)
            {
                push @{$metar{remark}}, { colourCode => _parseColourCode $1 };
            } elsif (   _cyInString(\@cy, ' cJP cUS MM MU NS UT ')
                     && $msg =~ s@^(8/([\d/])([\d/])([\d/])) @@)
            {
                push @{$metar{remark}}, { cloudTypes => {
                    s => $1,
                    $2 eq '/' ? ('cloudTypeLowAboveOvercast', undef)
                              : ('cloudTypeLow', $2),
                    $3 eq '/' ? ('cloudTypeMiddleAboveOvercast', undef)
                              : ('cloudTypeMiddle', $3),
                    $4 eq '/' ? ('cloudTypeHighAboveOvercast', undef)
                              : ('cloudTypeHigh', $4),
                }};
            } elsif ($msg =~ s@^(PWINO|FZRANO|TSNO|PNO|RVRNO|NO ?SPECI|VIA PHONE|RCRNR|(?:LGT |HVY )?FROIN|SD[FGP]/HD[FGP]|VRBL CONDS|ACFT MSHP)[/ ]@@)
            {
                $r->{s} = $1;
                ($r->{v} = $1) =~ tr/ /_/;
                $r->{v} =~ s/NO_?SPECI/NOSPECI/;
                $r->{v} =~ s/_VISBL$//;
                push @{$metar{remark}}, { keyword => $r };
            } elsif ($msg =~ s@^(?:/) @@){
                push @{$metar{remark}}, { keyword => {
                    s => '/',
                    v => 'slash'
                }};
            } elsif ($msg =~ s@^(?:\$) @@){
                push @{$metar{remark}}, { needMaint => { s => '$' }};
            } elsif (   _cyInString(\@cy, ' cCA cUS ')
                     && $msg =~ s@^(CIG ?(?:RAG|RGD)) @@)
            {
                $r->{s} = $1;
                ($r->{v} = $1) =~ tr/ /_/;
                $r->{v} =~ s/RGD/RAG/;
                $r->{v} =~ s/CIGRAG/CIG_RAG/;
                push @{$metar{remark}}, { keyword => $r };
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(CIG VRBL? ([1-9]\d*)-([1-9]\d*)) @@o)
            {
                $r->{s} = $1;
                $r->{visibilityFrom}{distance} = { v => $2 * 100, u => 'FT' };
                $r->{visibilityTo}{distance}   = { v => $3 * 100, u => 'FT' };
                push @{$metar{remark}}, { ceilVisVariable => $r };
            } elsif (   _cyInString(\@cy, ' cCA cJP cUS BGTL ET EG EQ LI LQ MHSC MM MU NS NZ RK TXKF ')
                     && $msg =~ s@^(SLP ?(?:(\d\d)(\d)|NO|///)) @@)
            {

                if (!defined $2) {
                    push @{$metar{remark}}, { SLPhPa => {
                        s            => $1,
                        notAvailable => undef
                    }};
                # TODO: really need QNH? but values are REALLY close:
                # KLWM 161454Z AUTO 31012G21KT 10SM CLR M10/M18 A2808 RMK SLP511
                #   -> 951.1 hPa
                # KQHT 021355Z 26003KT 8000 IC M15/M16 A3070 RMK SLP484
                #   -> 1048.4 hPa
                # higher QNHs possible:
                # ETNS 161520Z 11006KT 7000 FEW020 M02/M05 Q1078 WHT WHT
                #   -> SLP78?
                # UNWW 021400Z 00000MPS CAVOK M25/M29 Q1056 NOSIG RMK QFE759
                #   -> SLP56?
                } elsif (!exists $metar{QNH} || !exists $metar{QNH}{inHg}) {
                    if ($2 > 80 || $2 < 20) { # only if within sensible range
                        my $slp;
                        $slp = "$2.$3";
                        # threshold 55 taken from mdsplib
                        $slp += $slp < 55 ? 1000 : 900;
                        push @{$metar{remark}}, { SLPhPa => {
                            s => $1,
                            hPa => sprintf '%.1f', $slp
                        }};
                    } else {
                        push @{$metar{remark}}, { SLPhPa => {
                            s => $1,
                            invalidFormat => "no QNH, x$2.$3 hPa"
                        }};
                    }
                } else {
                    my ($slp, @slp, @qnh, $qnhHPa);
                    my $INHG2HPA = 33.86388640341;

                    @slp = ($1, $2, $3);
                    $qnhHPa = _rnd($metar{QNH}{inHg} * $INHG2HPA, 1);

                    $slp = $qnhHPa;                       # start with given QNH
                    $slp =~ s/..$//;                  # remove 2 trailing digits
                    $slp .= "$slp[1].$slp[2]";              # append given value
                    $slp += 100 if $slp + 50 < $qnhHPa;     # ensure SLP > QNH
                    push @{$metar{remark}}, { SLPhPa => {
                        s => $slp[0],
                        hPa => sprintf '%.1f', $slp
                    }};
                }
            } elsif (   _cyInString(\@cy, ' ROTM ')
                     && $msg =~ s@^(SP([23]\d\.\d{3})) @@)
            {
                push @{$metar{remark}}, { SLPinHg => {
                    s    => $1,
                    inHg => $2
                }};
            } elsif (   _cyInString(\@cy, ' x ')
                     && $msg =~ s@^(SLP(?:(\d{3})|NO)) @@)
            {
                if (!defined $2) {
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $1,
                        notAvailable => undef
                    }};
                } elsif (!exists $metar{QNH} || !exists $metar{QNH}{inHg}) {
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $1,
                        invalidFormat => "no QNH, x$2 in. Hg"
                    }};
                } else {
                    my ($qnhInHg, @slp);

                    @slp = ($1, $2);
                    $qnhInHg = $metar{QNH}{inHg} * 100;
                    $qnhInHg =~ s/...$//;
                    $qnhInHg .= "$slp[1]";
                    $qnhInHg /= 100;
                    $qnhInHg += 10 if $qnhInHg + 5 < $metar{QNH}{inHg};
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $slp[0],
                        inHg => $qnhInHg
                    }};
                }
            } elsif (   _cyInString(\@cy, ' NZ ')
                     && $msg =~ s@^GRID(${re_wind_dir}0${re_wind_speed}KT) @@o)
            {
                push @{$metar{remark}}, { gridWind => {
                    s    => "GRID$1",
                    wind => _parseWind $1
                }};
            } elsif (   _cyInString(\@cy, ' BG LC LH LT SP UL ')
                     && $msg =~ s@^($re_rwy_wind) @@o)
            {
                $r->{rwyWind}{s} = $1;
                $r->{rwyWind}{rwyDesig} = defined $2 ? $2 : $3;
                $r->{rwyWind}{wind} = _parseWind $4;
                $r->{rwyWind}{wind}{windVarLeft} = $5 + 0 if defined $5;
                $r->{rwyWind}{wind}{windVarRight} = $6 + 0 if defined $6;
                push @{$metar{remark}}, $r;
            } elsif (   _cyInString(\@cy, ' RC ')
                     && $msg =~ s@^($re_rwy_wind2) @@o)
            {
                $r->{rwyWind}{s} = $1;
                $r->{rwyWind}{wind} = _parseWind $2;
                $r->{rwyWind}{rwyDesig} = $3;
                push @{$metar{remark}}, $r;
            } elsif (   _cyInString(\@cy, ' cJP cUS BG EGUN ET LQTZ NZ RK ')
                     && $msg =~ s@^($re_rsc) @@o)
            {
                $r->{s} = $1;
                for ($1 =~ /SLR|LSR|PSR|P|SANDED|WET|DRY|IR|WR|\/\/|\d\d/g){
                    if ($_ eq '//') {
                        push @{$r->{rwySfcCondArr}}, { notAvailable => undef };
                    } elsif ($_ =~ /\d/) {
                        push @{$r->{rwySfcCondArr}}, { decelerometer => $_ };
                    } else {
                        push @{$r->{rwySfcCondArr}}, { rwySfc => $_ };
                    }
                }
                push @{$metar{remark}}, { rwySfcCondition => $r };
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(((?:$re_opacity_phenom[0-8])+)(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) @@o)
            {
                $r->{s} = $1;
                $r->{opacityPhenomArr} = ();
                _parseOpacityPhenom $r, $2;
                $r->{cloudTypeAsoctd} = $3 if defined $4;
                $r->{cloudTypeEmbd}   = $3 if defined $5;
                push @{$metar{remark}}, { opacityPhenom => $r };
            } elsif (   _cyInString(\@cy, ' BG KTTS ')
                     && $msg =~ s@^([0-8])($re_opacity_phenom) @@o)
            {
                $r->{s} = "$1$2";
                $r->{opacityPhenomArr} = ();
                _parseOpacityPhenom $r, "$2$1";
                push @{$metar{remark}}, { opacityPhenom => $r };
            } elsif (   _cyInString(\@cy, ' cJP YB YM YP ')
                     && $msg =~ s@^(([0-8])(?:((?:BL)?SN|FG)|($re_cloud_type))(\d{3})(?: TO)?($re_phen_locdirs)?) @@o)
            {
                $r->{s} = $1;
                $r->{oktas} = $2;
                $r->{weather} = _parseWeather $3, 'NI' if defined $3;
                $r->{cloudType} = $4 if defined $4;
                $r->{cloudBase} = { v => $5 * 100, u => 'FT' };
                $r->{locationAnd} = _parseLocations $6 if defined $6;
                push @{$metar{remark}}, { cloudOpacityLvl => $r };
            } elsif (   _cyInString(\@cy, ' cUS MH ')
                     && $msg =~ s@^(($re_cloud_cov$re_cloud_base?) V ($re_cloud_cov)) @@o)
            {
                $r->{cloudCoverVar} = _parseCloud $2;
                $r->{cloudCoverVar}{s} = $1;
                $r->{cloudCoverVar}{cloudCover2} = $3;
                push @{$metar{remark}}, $r;
            } elsif (   _cyInString(\@cy, ' cCA cUS EK LI LL NS ')
                     && $msg =~ s@^($re_cloud_cov$re_cloud_base) @@o)
            {
                push @{$metar{remark}}, { cloud => _parseCloud $1 };
            } elsif (   _cyInString(\@cy, ' U ')
                     && $msg =~ s@^($re_cloud_type)($re_cloud_base) @@o)
            {
                push @{$metar{remark}}, { cloudTypeLvl => {
                    s         => "$1$2",
                    cloudType => $1,
                    cloudBase => { v => $2 * 100, u => 'FT' },
                }};
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(TRS?( LWR)?(?: CLD|((?:[/ ]?$re_trace_cloud)+))($re_phen_locdirs)?) @@o)
            {
                $r->{s} = $1;
                $r->{isLower} = undef if defined $2;
                if (defined $3) {
                    for ($3 =~ /$re_trace_cloud/og) {
                        push @{$r->{cloudType}}, $_;
                    }
                } else {
                    $r->{cloudTypeNotAvailable} = undef;
                }
                $r->{locationAnd} = _parseLocations $4 if defined $4;
                push @{$metar{remark}}, { cloudTrace => $r };
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(((?:$re_trace_cloud[/ ])+)TR) @@o)
            {
                $r->{s} = $1;
                for ($2 =~ /$re_trace_cloud/og) {
                    push @{$r->{cloudType}}, $_;
                }
                push @{$metar{remark}}, { cloudTrace => $r };
            } elsif (   _cyInString(\@cy, ' cCA FQ TR LL ')
                     && $msg =~ s@^RE($re_weather_re) @@o)
            {
                push @{$metar{remark}},
                               { recentWeather => [ _parseWeather($1, 'RE') ] };
            } elsif (   _cyInString(\@cy, ' KQ ')
                     && $msg =~ s@^(SFC $re_vis (?:($re_vis_m)|($re_vis_km))) @@o)
            {
                $r->{s} = $1;
                $r->{locationAt} = 'SFC';
                if (defined $2) {
                    $r->{visibility}{distance} = _getVisibilityM $2;
                } else {
                    $r->{visibility}{distance} = { v => $3, rp => 1, u => 'KM'};
                    $r->{visibility}{distance}{v} =~ s/KM//;
                }
                push @{$metar{remark}}, { visibilityAtLoc => $r };
            } elsif (   _cyInString(\@cy, ' cUS ')
                     && $msg =~ s@^((SFC|TWR) $re_vis ($re_vis_sm)(?:SM)?) @@o)
            {
                $r->{s} = $1;
                $r->{locationAt} = $2;
                $r->{visibility} = _getVisibilitySM $3, $is_auto, \@cy;
                push @{$metar{remark}}, { visibilityAtLoc => $r };
            } elsif (   _cyInString(\@cy, ' cUS ')
                     && $msg =~ s@^($re_vis ($re_vis_sm)V($re_vis_sm)) @@o)
            {
                $r->{visVar1} = _getVisibilitySM $2, $is_auto, \@cy;
                $r->{visVar2} = _getVisibilitySM $3, $is_auto, \@cy;
                $r->{visVar1}{s} = $1;
                push @{$metar{remark}}, $r;
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^($re_vis(?: VRB)? ($re_vis_sm) ?- ?($re_vis_sm)) @@o)
            {
                $r->{visVar1} = _getVisibilitySM $2, $is_auto, \@cy;
                $r->{visVar2} = _getVisibilitySM $3, $is_auto, \@cy;
                $r->{visVar1}{s} = $1;
                push @{$metar{remark}}, $r;
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(([RS])(\d\d)(?: AFT ?($re_hour) ?(?:UTC)?)?) @@o)
            {
                # http://ops.8wing.ca/downloads/WX_Decode_card.doc
                my $hoursFrom;

                if (defined $4) {
                    $hoursFrom = $metar{obsTime}{hour} - $4;
                    $hoursFrom += 24 if $hoursFrom < 0;
                } else {
                    $hoursFrom = $metar{obsTime}{hour} % 6;
                    $hoursFrom = 6 if $hoursFrom == 0;
                }
                push @{$metar{remark}}, {
                    ($2 eq 'S' ? 'snowFall' : 'precipHourly') => {
                        s                  => $1,
                        timeBeforeObsSince => { hoursFrom => $hoursFrom },
                        precipAmount       => { v => $3 + 0,
                                                u => ($2 eq 'S' ? 'CM' : 'MM') }
                }};
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(VSBY VRBL (\d\.\d)V(\d)(\.[+\d])) @@)
            {
                $r->{visVar1} = _getVisibilitySM $2, $is_auto, \@cy;
                $r->{visVar2} = _getVisibilitySM $4 eq '.+' ? "P$3" : "$3$4", $is_auto, \@cy;
                $r->{visVar1}{s} = $1;
                push @{$metar{remark}}, $r;
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(PCPN (\d+\.\d)MM PAST HR) @@)
            {
                push @{$metar{remark}}, { precipHourly => {
                    s            => $1,
                    period       => '1h',
                    precipAmount => { v => $2 + 0, u => 'MM' }
                }};
            } elsif (   _cyInString(\@cy, ' cJP E KQ LQ NZ ')
                     && $msg =~ s@^($re_vis((?:$re_phen_locdirs $re_vis_m_km_remark)(?:(?: AND)?$re_phen_locdirs $re_vis_m_km_remark)*)) @@o)
            {
                my @visLocData;

                $r->{visListAtLoc}{s} = $1;
                $r->{visListAtLoc}{visLocData} = \@visLocData;
                for ($2 =~ m@$re_phen_locdirs $re_vis_m_km_remark@og){
                    my $v;

                    m@($re_phen_locdirs) (?:($re_vis_m)|([1-9]\d{2,3})M|($re_vis_km))@o;
                    $v->{locationAnd} = _parseLocations $1;
                    if (defined $2) {
                        $v->{visibility}{distance} = _getVisibilityM $2;
                    } elsif (defined $3) {
                        $v->{visibility}{distance} =
                                            _getVisibilityM sprintf('%04d', $3);
                    } else {
                        $v->{visibility}{distance} = { v => $4, rp => 1, u => 'KM' };
                        $v->{visibility}{distance}{v} =~ s/KM//;
                    }
                    push @visLocData, $v;
                }
                push @{$metar{remark}}, $r;
            } elsif ($msg =~ s@^($re_vis($re_phen_locdirs) LWR) @@o) {
                push @{$metar{remark}}, { phenomenonAtLoc => {
                    s           => $1,
                    otherPhenom => 'VIS_LWR',
                    locationAnd => _parseLocations $2
                }};
            } elsif (   _cyInString(\@cy, ' cJP RP ')
                     && $msg =~ s@^(A)([23]\d)(\d\d) @@)
            {
                push @{$metar{remark}}, { QNH => {
                    s    => "$1$2$3",
                    inHg => "$2.$3"
                }};
            } elsif (   _cyInString(\@cy, ' KQ ')
                     && $msg =~ s@^(Q(?:NH)?([01]\d{3})MB) @@)
            {
                push @{$metar{remark}}, { QNH => {
                    s => $1,
                    hPa => $2
                }};
            } elsif (_cyInString(\@cy, ' MM ') && $msg =~ s@^([789]\d\d) @@) {
                push @{$metar{remark}}, { QFF => {
                    s => $1,
                    hPa => $1
                }};
            } elsif ($msg =~ s@^(QNH([01]\d{3}\.\d)) @@) {
                push @{$metar{remark}}, { QNH => {
                    s => $1,
                    hPa => $2
                }};
            } elsif (   _cyInString(\@cy, ' cUS EG ET LQ ')
                     && $msg =~ s@^(COR ($re_hour)($re_min)) @@o)
            {
                push @{$metar{remark}}, { correctedAt => {
                    s      => $1,
                    hour   => $2,
                    minute => $3
                }};
            } elsif (   _cyInString(\@cy, ' cUS ')
                     && $msg =~ s@^(SNINCR (\d+)/(\d+)) @@)
            {
                push @{$metar{remark}}, { snowIncr => {
                    s        => $1,
                    pastHour => $2,
                    onGround => $3
                }};
            } elsif (   _cyInString(\@cy, ' cCA cJP cUS BG EG ET EQ LI LQ MM MU NS RKSO ')
                     && $msg =~ s@^1([01]\d{3}) @@)
            {
                push @{$metar{remark}}, { tempMaxPeriod => {
                    s      => "1$1",
                    period => '6h',
                    temp   => _parseTemp($1)
                }};
            } elsif (   _cyInString(\@cy, ' cCA cJP cUS BG EG ET EQ LI LQ MM MU NS RKSO ')
                     && $msg =~ s@^2([01]\d{3}) @@)
            {
                push @{$metar{remark}}, { tempMinPeriod => {
                    s     => "2$1",
                    period => '6h',
                    temp  => _parseTemp($1)
                }};
            } elsif (   _cyInString(\@cy, ' cCA cJP cUS BG EG ET EQ LI LQ MM MU NS RKSO ')
                     && $msg =~ s@^(5(?:([0-8])(\d{3})|////)) @@)
            {
                push @{$metar{remark}}, { pressureChange => {
                    s      => $1,
                    period => '3h',
                    defined $2 ? ('pressureTendency', $2,
                                  'pressureChangeVal', {
                                              v => $3 / 10 * ($2 >= 5 ? -1 : 1),
                                              u => 'hPa'
                                 })
                               : ('notAvailable', undef)
                }};
            } elsif (   _cyInString(\@cy, ' cCA cUS ET EQ LP LQ NZ ROTM ')
                     && $msg =~ s@^(PK[ /]?WND ?($re_wind_dir\d$re_wind_speed)/(?:($re_hour)?($re_min))) @@o)
            {
                push @{$metar{remark}}, { peakWind => {
                    s      => $1,
                    wind   => _parseWind("$2KT"),
                    defined $3 ? ('hour', $3) : (),
                    minute => $4
                }};
            } elsif (   $cy[0] eq 'KEHA'
                     && $msg =~ s@^(PK[ /]?WND ($re_wind_speed) 000) @@o)
            {
                push @{$metar{remark}}, { peakWind => {
                    s    => $1,
                    wind => _parseWind("///$2KT")
                }};
            } elsif (   $cy[0] eq 'EDLP'
                     && $msg =~ s@^RWY ?($re_rwy_des) ?(TRL[0 ]?(\d\d)) ([A-Z]{2}) ATIS ([A-Z]) @@o)
            {
                push @{$metar{remark}}, { activeRwy => {
                    s        => "RWY$1",
                    rwyDesig => $1
                }};
                push @{$metar{remark}}, { transitionLvl => {
                    s     => $2,
                    level => $3 + 0
                }};
                push @{$metar{remark}}, { obsInitials => { s => $4 }};
                push @{$metar{remark}}, { currentATIS => {
                    s    => "ATIS $5",
                    ATIS => $5
                }};
            } elsif (   _cyInString(\@cy, ' cCA cJP cUS BG EG ET EQ LI NS MU ')
                     && $msg =~ s@^(T([01]\d{3})([01]\d{3})?) @@)
            {
                $r->{s} = $1;
                $r->{air}{temp} = _parseTemp $2;
                $r->{dewpoint}{temp} = _parseTemp $3
                    if defined $3;
                push @{$metar{remark}}, { tempHourly => $r };
            } elsif (   _cyInString(\@cy, ' cUS ')
                     && $msg =~ s@^(98(?:(\d{3})|///)) @@)
            {
                $r->{s} = $1;
                $r->{sunshinePeriod} = '1d';
                if (defined $2) {
                    $r->{sunshine} = { v => $2 + 0, u => 'min' };
                } else {
                    $r->{sunshineNotAvailable} = undef;
                }
                push @{$metar{remark}}, { radiationSun => $r };
            } elsif (   _cyInString(\@cy, ' UTTP UTSS ')
                     && $msg =~ s@^(QFE([01]?\d{3})) @@)
            {
                push @{$metar{remark}}, { QFE => {
                    s => $1,
                    hPa => $2 + 0
                }};
            } elsif (   _cyInString(\@cy, ' cCA cUS ')
                     && $msg =~ s@^(/?(DA|PA)[ /]?([+-]?\d+)(?:FT)?/?) @@)
            {
                push @{$metar{remark}}, { ({ DA => 'densityAlt',
                                             PA => 'pressureAlt'
                                           }->{$2})
                                          => {
                    s        => $1,
                    altitude => { v => $3 + 0, u => 'FT' }
                }};
            } elsif (   _cyInString(\@cy, ' KNIP ')
                     && $msg =~ s@^((RH|SST|AI|OAT)/(\d\d)F?) @@)
            {
                $r->{s} = $1;
                if ($2 eq 'SST' || $2 eq 'OAT') {
                    $r->{temp} = { v => $3 + 0, u => 'F' };
                } elsif ($2 eq 'RH') {
                    $r->{relHumid} = $3 + 0;
                } else {
                    $r->{"$2Val"} = $3 + 0;
                }
                push @{$metar{remark}}, { $2 => $r };
            } elsif (   _cyInString(\@cy, ' KN ')
                     && $msg =~ s@^((RH|SST)/(\d+)) @@)
            {
                $r->{s} = $1;
                if ($2 eq 'SST') {
                    $r->{temp} = { v => $3 + 0, u => 'C' };
                } else {
                    $r->{relHumid} = $3 + 0;
                }
                push @{$metar{remark}}, { $2 => $r };
            } elsif (   _cyInString(\@cy, ' U ZM ')
                     && $msg =~ s@^(QFE ?(\d{3})(?:,(\d))?(?:/(\d{3,4}))?) @@)
            {
                my $qfe;

                $qfe->{s} = $1;
                $qfe->{mmHg} = $2 + 0;
                $qfe->{mmHg} += $3 / 10 if defined $3;
                $qfe->{hPa} = $4 + 0 if defined $4;
                push @{$metar{remark}}, { QFE => $qfe };
            } elsif ($msg =~ s@^(((?:VIS|CHI)NO) RWY ?($re_rwy_des)) @@o) {
                push @{$metar{remark}}, { $2 => {
                    s => $1,
                    rwyDesig => $3
                }};
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^($re_snw_cvr) @@o)
            {
                $r->{s} = $1;
                $r->{snowCoverType} = defined $2 ? $2 : 'NIL';
                $r->{snowCoverType} =~ s/ONE /ONE_/;
                $r->{snowCoverType} =~ s/MU?CH /MUCH_/;
                $r->{snowCoverType} =~ s/TR(?:ACE ?)? ?/TRACE_/;
                $r->{snowCoverType} =~ s/MED(?:IUM)?/MEDIUM/;
                $r->{snowCoverType} =~ s/ PACK(?:ED)?/_PACKED/;
                push @{$metar{remark}}, { snowCover => $r };
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(OBS TAKEN [+](\d+)) @@)
            {
                push @{$metar{remark}}, { obsTimeOffset => {
                    s       => $1,
                    minutes => $2
                }};
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^((?:BLN (?:DSA?PRD (\d+) ?FT\.?|VI?SBL(?: TO)? (\d+) ?FT))) @@)
            {
                $r->{s} = $1;
                if (defined $2) {
                    $r->{disappearedAt} = { distance => { v => $2, u => 'FT' }};
                } else {
                    $r->{visibleTo} = { distance => { v => $3, u => 'FT' }};
                }
                push @{$metar{remark}}, { balloon => $r };
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(RVR(?: RWY ?$re_rwy_des $re_rwy_vis ?FT\.?)+) @@o)
            {
                for ($1 =~ /((?:RVR )?RWY ?$re_rwy_des $re_rwy_vis ?FT\.?)/og) {
                    my $v;

                    $v->{s} = "$_";

                    /($re_rwy_des) ($re_rwy_vis)/;

                    $v->{rwyDesig}      = $1;
                    $v->{RVR}{distance} = { v => $2, u => 'FT' };
                    $v->{RVR}{distance}{q} = 'isLess'
                        if $v->{RVR}{distance}{v} =~ s/^M//;
                    $v->{RVR}{distance}{q} = 'isEqualGreater'
                        if $v->{RVR}{distance}{v} =~ s/^P//;
                    $v->{RVR}{distance}{v} += 0;
                    _setVisibilityFTRVROffset $v->{RVR}{distance};
                    push @{$metar{remark}}, { visRwy => $v };
                }
            } elsif (   _cyInString(\@cy, ' FQ ')
                     && $msg =~ s@^(TX/(\d\d[,.]\d)) @@)
            {
                $r->{tempMaxFQ}{s} = $1;
                $r->{tempMaxFQ}{temp} = { v => $2, u => 'C' };
                $r->{tempMaxFQ}{temp}{v} =~ s/,/./;
                push @{$metar{remark}}, $r;
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(NXT FCST BY (?:($re_hour)|($re_day)($re_hour)($re_min))Z) @@o)
            {
                if (defined $2) {
                    push @{$metar{remark}}, { nextFcstBy => {
                        s      => $1,
                        hour   => $2,
                        minute => '00'
                    }};
                } else {
                    push @{$metar{remark}}, { nextFcstBy => {
                        s      => $1,
                        day    => $3,
                        hour   => $4,
                        minute => $5
                    }};
                }
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(NXT FCST WILL BE ISSUED AT ($re_day)($re_hour)($re_min)Z) @@o)
            {
                push @{$metar{remark}}, { nextFcstAt => {
                    s      => $1,
                    day    => $2,
                    hour   => $3,
                    minute => $4
                }};
            } elsif (   _cyInString(\@cy, ' LT ')
                     && $msg =~ s@^(AMD AT ($re_day)($re_hour)($re_min)Z) @@o)
            {
                push @{$metar{remark}}, { amdAt => {
                    s      => $1,
                    day    => $2,
                    hour   => $3,
                    minute => $4
                }};
            } elsif (   _cyInString(\@cy, ' cCA ')
                     && $msg =~ s@^(FCST BASED ON AUTO OBS(?: ($re_hour)-($re_hour)Z)?\.?) @@o)
            {
                push @{$metar{remark}}, { fcstAutoObs => {
                    s => $1,
                    defined $2 ? ('hourFrom', $2,
                                  'hourTill', $3)
                               : ()
                }};
            } elsif (   _cyInString(\@cy, ' EF ')
                     && $msg =~ s@^(BASED ON AUTOMETAR) @@)
            {
                push @{$metar{remark}}, { fcstAutoMETAR => { s => $1 }};
            } elsif (   _cyInString(\@cy, ' ED ')
                     && $msg =~ s@^(ATIS ([A-Z])) @@)
            {
                push @{$metar{remark}}, { currentATIS => {
                    s    => $1,
                    ATIS => $2
                }};
            } elsif ($msg =~ s@^(((?:$re_data_estmd[ /])+)(?:VISUALLY )?$re_estmd) @@o)
            {
                $r->{s} = $1;
                for ($2 =~ /$re_data_estmd/og) {
                    s/WI?NDS?(?: DATA)?/WND/;
                    s/CEILING/CIG/;
                    s/(?:CIG )?BLN/CIG BLN/;
                    s/ /_/g;
                    push @{$r->{estimatedItem}}, $_;
                }
                push @{$metar{remark}}, { estimated => $r };
            } elsif ($msg =~ s@^($re_estmd((?:[ /]$re_data_estmd)+)) @@o) {
                $r->{s} = $1;
                for ($2 =~ /$re_data_estmd/og) {
                    s/WI?NDS?(?: DATA)?/WND/;
                    s/CEILING/CIG/;
                    s/(?:CIG )?BLN/CIG BLN/;
                    s/ /_/g;
                    push @{$r->{estimatedItem}}, $_;
                }
                push @{$metar{remark}}, { estimated => $r };
            } elsif ($msg =~ s@^((?:(THN) )?($re_cloud_cov) ABV ($re_cloud_base)) @@o)
            {
                $r->{cloudAbove}{s} = $1;
                $r->{cloudAbove}{isThin} = undef if defined $2;
                $r->{cloudAbove}{cloudCover} = $3;
                $r->{cloudAbove}{cloudBase}  = { v => $4 * 100, u => 'FT' };
                push @{$metar{remark}}, $r;
            } elsif (   _cyInString(\@cy, ' cUS NZ ')
                     && $msg =~ s@^((FU|(?:FZ)?FG|BR|SN|DU|PWR PLNT(?: PLUME)?) ($re_cloud_cov$re_cloud_base)) @@o)
            {
                my ($cl, $phen);

                $r->{s} = $1;
                $r->{cloud} = _parseCloud $3;
                $phen = $2;
                if ($phen =~ m@^$re_weather$@o) {
                    $r->{weather} = _parseWeather $phen, 'NI';
                } else {
                    ($r->{cloudPhenom} = $phen) =~ s/ /_/g;
                }
                push @{$metar{remark}}, { obscuration => $r };
            } elsif (   _cyInString(\@cy, ' YA YB ')
                     && $msg =~ s@^(RF(\d\d\.\d)/(\d{3}\.\d)) @@)
            {
                push @{$metar{remark}}, { rainfall => {
                    s             => $1,
                    rainfall10min => $2 + 0,
                    rainfall0900  => $3 + 0
                }};
            } elsif (   _cyInString(\@cy, ' KHMS ')
                     && $msg =~ s@^(RSNK (M)?(\d\d(?:\.\d)?)/(${re_wind_dir}0$re_wind_speed(?:G$re_wind_speed)?)($re_wind_speed_unit)?) @@o)
            {
                my ($temp, $unit);

                $temp = $3 + 0;
                $temp *= -1 if defined $2;
                $unit = defined $5 ? $5 : 'KT';
                push @{$metar{remark}}, { RSNK => {
                    s    => $1,
                    air  => { temp => { v => $temp, u => 'C' } },
                    wind => _parseWind "$4$unit"
                }};
            } elsif (   _cyInString(\@cy, ' KNTD ')
                     && $msg =~ s@^(LAG ?PK (M)?(\d\d)/(M)?(\d\d)/(${re_wind_dir}0$re_wind_speed(?:G$re_wind_speed)?)) @@o)
            {
                my ($temp, $dewpoint);

                $temp = $3 + 0;
                $temp *= -1 if defined $2;
                $dewpoint = $5 + 0;
                $dewpoint *= -1 if defined $4;
                push @{$metar{remark}}, { LAG_PK => {
                    s        => $1,
                    air      => { temp => { v => $temp,     u => 'C' } },
                    dewpoint => { temp => { v => $dewpoint, u => 'C' } },
                    wind     => _parseWind "$6KT"
                }};
            } elsif (   _cyInString(\@cy, ' KAEG KAUN KBVS KUKT KW22 ')
                     && $msg =~ s@^(P(\d{3})) @@)
            {
                push @{$metar{remark}}, { precipHourly => {
                    s      => $1,
                    $2 == 0 ? (traces       => undef)
                            : (precipAmount => { v => $2 / 100, u => 'IN' }),
                    period => '1h'
                }};

# nearly the same pattern twice except for the position of $re_phen_desc
# (do not match too much but don't miss anything, either)

            } elsif ($msg =~ s@^(((?:$re_phen_desc[/ ]?)+)$re_phenomenon4( BBLO)?(?:$re_loc_quadr3$re_wx_mov_d2?|$re_loc_quadr3?$re_wx_mov_d2)( BBLO)?(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) @@o)
            {
                $r->{s} = $1;
                _parsePhenomDescr $r, 'phenomDescrPre', $2;
                if (defined $3) {
                    _parsePhenom \$r, $3;
                } elsif (defined $4) {
                    @{$r->{cloudType}} = split m@[/-]@, $4;
                } elsif (defined $5) {
                    # phenomenon _can_ have intensity, but it is an EXTENSION
                    $r->{weather} = _parseWeather $5, 'NI';
                } else {
                    $r->{cloudCover} = $6;
                }
                _parsePhenomDescr $r, 'phenomDescrPost', $7 if defined $7;
                $r->{locationAnd} = _parseLocations $8 if defined $8;
                $r->{locationAnd} = _parseQuadrants $10, $9 if defined $10;
                $r->{$11}{locationAnd} = _parseLocations $12 if defined $12;
                $r->{locationAnd} = _parseLocations $13 if defined $13;
                $r->{locationAnd} = _parseQuadrants $15, $14 if defined $15;
                $r->{$16}{locationAnd} = _parseLocations $17 if defined $17;
                _parsePhenomDescr $r, 'phenomDescrPost', $18 if defined $18;
                $r->{cloudTypeAsoctd} = $19 if defined $20;
                $r->{cloudTypeEmbd}   = $19 if defined $21;
                push @{$metar{remark}}, { phenomenonAtLoc => $r };
            } elsif ($msg =~ s@^($re_phenomenon4(?: IS)?((?:[/ ](?:$re_phen_desc|BBLO))*)(?:$re_loc_quadr3$re_wx_mov_d2?|$re_loc_quadr3?$re_wx_mov_d2)( BBLO)?(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) @@o)
            {
                $r->{s} = $1;
                if (defined $2) {
                    _parsePhenom \$r, $2;
                } elsif (defined $3) {
                    @{$r->{cloudType}} = split m@[/-]@, $3;
                } elsif (defined $4) {
                    # phenomenon _can_ have intensity, but it is an EXTENSION
                    $r->{weather} = _parseWeather $4, 'NI';
                } else {
                    $r->{cloudCover} = $5;
                }
                _parsePhenomDescr $r, 'phenomDescrPost', $6 if defined $6;
                $r->{locationAnd} = _parseLocations $7 if defined $7;
                $r->{locationAnd} = _parseQuadrants $9, $8 if defined $9;
                $r->{$10}{locationAnd} = _parseLocations $11 if defined $11;
                $r->{locationAnd} = _parseLocations $12 if defined $12;
                $r->{locationAnd} = _parseQuadrants $14, $13 if defined $14;
                $r->{$15}{locationAnd} = _parseLocations $16 if defined $16;
                _parsePhenomDescr $r, 'phenomDescrPost', $17 if defined $17;
                $r->{cloudTypeAsoctd} = $18 if defined $19;
                $r->{cloudTypeEmbd}   = $18 if defined $20;
                push @{$metar{remark}}, { phenomenonAtLoc => $r };
            } elsif ($msg =~ s@^((?:PR(?:ESENT )?WX )?((?:(?:AND )?$re_phen_desc[/ ]?)+)$re_phenomenon4( BBLO)?(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) @@o)
            {
                $r->{s} = $1;
                _parsePhenomDescr $r, 'phenomDescrPre', $2 if defined $2;
                if (defined $3) {
                    _parsePhenom \$r, $3;
                } elsif (defined $4) {
                    @{$r->{cloudType}} = split m@[/-]@, $4;
                } elsif (defined $5) {
                    # phenomenon _can_ have intensity, but it is an EXTENSION
                    $r->{weather} = _parseWeather $5, 'NI';
                } else {
                    $r->{cloudCover} = $6;
                }
                _parsePhenomDescr $r, 'phenomDescrPost', $7 if defined $7;
                $r->{cloudTypeAsoctd} = $8 if defined $9;
                $r->{cloudTypeEmbd}   = $8 if defined $10;
                push @{$metar{remark}}, { phenomenonOnly => $r };

            # first check a restricted set of phenomenon and descriptions!

            } elsif ($msg =~ s@^(GR ($re_gs_size)) @@o) {
                push @{$metar{remark}}, { hailStones => {
                    s             => $1,
                    hailStoneSize => _parseFraction $2, 'IN'
                }};
            } elsif ($msg =~ s@^((PRES[FR]R)( PAST HR)?) @@) {
                $r->{s} = $1;
                $r->{otherPhenom} = $2;
                _parsePhenomDescr $r, 'phenomDescrPost', $3 if defined $3;
                push @{$metar{remark}}, { phenomenonOnly => $r };
            } elsif ($msg =~ s@^(($re_phenomenon_other)(?: IS)?((?:[/ ](?:$re_phen_desc_other|BBLO))*)(?: ($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) @@o)
            {
                $r->{s} = $1;
                _parsePhenom \$r, $2;
                _parsePhenomDescr $r, 'phenomDescrPost', $3 if defined $3;
                $r->{cloudTypeAsoctd} = $4 if defined $5;
                $r->{cloudTypeEmbd}   = $4 if defined $6;
                push @{$metar{remark}}, { phenomenonOnly => $r };
            } elsif ($msg =~ s@^((?:PR(?:ESENT )?WX )?$re_phenomenon4((?: IS)?(?:[/ ](?:(?:AND )?$re_phen_desc|BBLO))*)(?: ?($re_cloud_type) (?:(ASOCTD?)|(EMBD)))?) @@o)
            {
                $r->{s} = $1;
                if (defined $2) {
                    _parsePhenom \$r, $2;
                } elsif (defined $3) {
                    @{$r->{cloudType}} = split m@[/-]@, $3;
                } elsif (defined $4) {
                    # phenomenon _can_ have intensity, but it is an EXTENSION
                    $r->{weather} = _parseWeather $4, 'NI';
                } else {
                    $r->{cloudCover} = $5;
                }
                _parsePhenomDescr $r, 'phenomDescrPost', $6 if defined $6;
                $r->{cloudTypeAsoctd} = $7 if defined $8;
                $r->{cloudTypeEmbd}   = $7 if defined $9;
                push @{$metar{remark}}, { phenomenonOnly => $r };
            } elsif ($msg =~ s@^(AO(?:1|2A?)) @@) {
                push @{$metar{remark}}, { obsStationType => {
                    s           => $1,
                    stationType => $1
                }};
            } elsif ($msg =~ s@^(CIG ($re_cloud_base)(?:(?:(?: (APCH))?(?: RWY ?| R?)?($re_rwy_des))(?: TO)?($re_phen_locdirs)?|(?: TO)?($re_phen_locdirs))) @@o)
            {
                $r->{s} = $1;
                $r->{cloudBase}  = { v => $2 * 100, u => 'FT' };
                $r->{isApproach} = undef if defined $3;
                $r->{rwyDesig} = $4 if defined $4;
                $r->{locationAnd} =_parseLocations $5 if defined $5;
                $r->{locationAnd} =_parseLocations $6 if defined $6;
                push @{$metar{remark}}, { ceilingAtLoc => $r };
            } elsif ($msg =~ s@^($re_vis ($re_vis_sm)(?: (APCH))?(?: RWY ?| R?)?($re_rwy_des)) @@o)
            {
                $r->{s} = $1;
                $r->{visibility} = _getVisibilitySM $2, $is_auto, \@cy;
                $r->{isApproach} = undef if defined $3;
                $r->{rwyDesig} = $4;
                push @{$metar{remark}}, { visibilityAtLoc => $r };
            } elsif ($msg =~ s@^(NOSIG|CAVU|$re_estmd PASS (?:OPEN|CLOSED|CLSD|MARGINAL|MRGL)|PASS $re_estmd CLSD|EPC|EPO|EPM|RTS) @@o) {
                $r->{s} = $1;
                $r->{v} = $1;
                if ($r->{v} =~ m@PASS OP@) {
                    $r->{v} = 'EPO';
                } elsif ($r->{v} =~ m@PASS .*CL@) {
                    $r->{v} = 'EPC';
                } elsif ($r->{v} =~ m@PASS M@) {
                    $r->{v} = 'EPM';
                }
                push @{$metar{remark}}, { keyword => $r };
            } elsif ($msg =~ s@^((?:CLIMAT ?)?($re_temp)/($re_temp)(?:/(TR|$re_precip)(?:/(NIL|$re_precip))?)?) @@o)
            {
                my ($precip1, $precip2);

                $r->{s} = $1;
                $r->{temp1}{temp} = { v => $2 + 0, u => 'C' };
                $r->{temp2}{temp} = { v => $3 + 0, u => 'C' };
                if (defined $4) {
                    if ($4 eq 'TR') {
                        $r->{traces} = undef;
                    } else {
                        $precip1 = $4;
                    }
                    if (defined $5) {
                        if ($5 eq 'NIL') {
                            $r->{precipAmount2MM} = 0;
                        } else {
                            $precip2 = $5;
                            if ($precip2 =~ s/ ?([MC]M)//) {
                                $r->{precipAmount2MM} = $precip2 + 0;
                                $r->{precipAmount2MM} *= 10 if $1 eq 'CM';
                            } else {
                                $r->{precipAmount2Inch} = $precip2 + 0;
                            }
                        }
                    }
                    if (defined $precip1) {
                        if ($precip1 =~ s/ ?([MC]M)//) {
                            $r->{precipAmount1MM} = $precip1 + 0;
                            $r->{precipAmount1MM} *= 10 if $1 eq 'CM';
                        } else {
                            $r->{precipAmount1Inch} = $precip1 + 0;
                        }
                    }
                }
                push @{$metar{remark}}, { climate => $r };
            } elsif ($msg =~ s@^((TORNADO|FUNNEL CLOUDS?|WATERSPOUT) ($re_be_weather_be+)(?: TO)?($re_phen_locdirs)$re_wx_mov_d2?) @@o) {
                $r->{s} = $1;
                $r->{tornadicActivityType} = lc $2;
                $r->{locationAnd} = _parseLocations $4;
                $r->{$5}{locationAnd} = _parseLocations $6 if defined $6;
                $r->{startEndArr} = ();
                for ($3 =~ /$re_be_weather_be/og) {
                    my %s_e;
                    /(.)(..)?(..)/;
                    $s_e{hour} = $2 if defined $2;
                    $s_e{minute} = $3;
                    if ($1 eq 'B') {
                        push @{$r->{startEndArr}}, { beganAt => \%s_e };
                    } else {
                        push @{$r->{startEndArr}}, { endedAt => \%s_e };
                    }
                }
                $r->{tornadicActivityType} =~ s/(CLOUD)S?/$1/;
                $r->{tornadicActivityType} =~ s/ /_/;
                push @{$metar{remark}}, { tornadicActivity => $r };
            } elsif ($msg =~ s@^((?:FIRST|FST)(?:( STAFFED| STFD)|( MANNED))?(?: OBS?)?)[ /]@@)
            {
                $r->{s} = $1;
                $r->{isStaffed} = undef if defined $2;
                $r->{isManned} = undef if defined $3;
                push @{$metar{remark}}, { firstObs => $r };
            } elsif ($msg =~ s@^(NEXT ($re_day)($re_hour)($re_min)(?: ?UTC| ?Z)?) @@o)
            {
                $r->{s} = $1;
                @{$r->{obsAt}}{'day', 'hour', 'minute'} = ($2, $3, $4)
                    if defined $2;
                push @{$metar{remark}}, { nextObs => $r };
            } elsif ($msg =~ s@^(LAST(?:( STAFFED)|( MANNED))?(?: OBS?)?(?: ($re_day)($re_hour)($re_min)(?: ?UTC| ?Z)?)?)[ /]@@o)
            {
                $r->{s} = $1;
                $r->{isStaffed} = undef if defined $2;
                $r->{isManned} = undef if defined $3;
                @{$r->{obsAt}}{'day', 'hour', 'minute'} = ($4, $5, $6)
                    if defined $4;
                push @{$metar{remark}}, { lastObs => $r };
            } elsif ($msg =~ s@^((QBB|QBJ)(\d\d0)) @@) {
                push @{$metar{remark}}, { $2 => {
                    s => $1,
                    altitude => { v => $3 + 0, u => 'M' }
                }};
            # from mdsplib: RADAT: elevation information
            } elsif ($msg =~ s@^(RADAT (?:(\d\d)(\d{3})|MISG)) @@) {
                $r->{s} = $1;
                if (defined $2) {
                    $r->{relHumid} = $2 + 0;
                    $r->{distance} = { v => $3 * 100, u => 'FT' };
                } else {
                    $r->{isMissing} = undef;
                }
                push @{$metar{remark}}, { RADAT => $r };
            } elsif ($cy[1] eq 'LF' && $msg =~ s@^([MB])(\d) @@) {
                push @{$metar{remark}}, { reportConcerns => {
                    s       => "$1$2",
                    change  => $1,
                    subject => $2
                }};
            } elsif ($msg =~ s@^(WI?ND (850 ?FT) ($re_wind)(?: ($re_wind_dir\d)V($re_wind_dir\d))?) @@o)
            {
                push @{$metar{remark}}, _parseWindAtLoc($1, $2, $3, $8, $9);
            } elsif ($msg =~ s@^((2000FT|CNTR RWY|HARBOR|ROOF|BAY TWR) WI?ND ($re_wind)(?: ($re_wind_dir\d)V($re_wind_dir\d))?) @@o)
            {
                push @{$metar{remark}}, _parseWindAtLoc($1, $2, $3, $8, $9);
            } elsif ($msg =~ s@^((BAY TWR|WNDY HILL|KAUKAU) ($re_wind)(?: ($re_wind_dir\d)V($re_wind_dir\d))?) @@o)
            {
                push @{$metar{remark}}, _parseWindAtLoc($1, $2, $3, $8, $9);
            } elsif (_cyInString \@cy, ' LI EQ ') {
                my $re_cond_moun =
                    '(?:LIB|CLD SCT|VERS INC|CNS POST|CLD CIME'
                    . '|CIME INC|GEN INC|INC|INVIS)';
                my $re_chg_moun =
                    '(?:NC|CUF'
                     . '|ELEV (?:SLW|RAPID|STF)'
                     . '|ABB (?:SLW|RAPID)'
                     . '|STF(?: ABB)?'
                     . '|VAR RAPID)';
                my $re_cond_vall =
                    '(?:NIL|FOSCHIA(?: SKC SUP)?|NEBBIA(?: SCT)?'
                    . '|CLD SCT(?: NEBBIA INF)?|MAR CLD|INVIS)';
                my $re_chg_vall =
                    '(?:NC|ELEV|DIM(?: ELEV| ABB)?|AUM(?: ELEV| ABB)?|ABB'
                    . '|NEBBIA INTER)';
                my $re_moun =
                            "$re_phen_locdirs? $re_cond_moun(?: $re_chg_moun)?";
                my $re_vall =
                            "$re_phen_locdirs? $re_cond_vall(?: $re_chg_vall)?";
                if ($msg =~ s@^(MON((?:$re_moun)+)) @@o) {
                    $r->{s} = $1;
                    $r->{condMountainLoc} = ();
                    for ($2 =~ m@$re_moun@og) {
                        my $m;
                        m@($re_phen_locdirs)? ($re_cond_moun)(?: ($re_chg_moun))?@o;

                        $m->{locationAnd}     =_parseLocations $1 if defined $1;
                        $m->{cloudMountain} = {
                                               'LIB'      => 0,
                                               'CLD SCT'  => 1,
                                               'VERS INC' => 2,
                                               'CNS POST' => 3,
                                               'CLD CIME' => 5,
                                               'CIME INC' => 6,
                                               'GEN INC'  => 7,
                                               'INC'      => 8,
                                               'INVIS'    => 9
                                               }->{$2};
                        $m->{cloudEvol} = {
                                           'NC'         => 0,
                                           'CUF'        => 1,
                                           'ELEV SLW'   => 2,
                                           'ELEV RAPID' => 3,
                                           'ELEV STF'   => 4,
                                           'ABB SLW'    => 5,
                                           'ABB RAPID'  => 6,
                                           'STF'        => 7,
                                           'STF ABB'    => 8,
                                           'VAR RAPID'  => 9
                                          }->{$3}
                            if defined $3;
                        push @{$r->{condMountainLoc}}, $m;
                    }
                    push @{$metar{remark}}, { conditionMountain => $r };
                } elsif ($msg =~ s@^(VAL((?:$re_vall)+)) @@o) {
                    $r->{s} = $1;
                    $r->{condValleyLoc} = ();
                    for ($2 =~ m@$re_vall@og) {
                        my $m;
                        m@($re_phen_locdirs)? ($re_cond_vall)(?: ($re_chg_vall))?@o;

                        $m->{locationAnd}     =_parseLocations $1 if defined $1;
                        $m->{cloudValley} = {
                                             'NIL'                => 0,
                                             'FOSCHIA SKC SUP'    => 1,
                                             'NEBBIA SCT'         => 2,
                                             'FOSCHIA'            => 3,
                                             'NEBBIA'             => 4,
                                             'CLD SCT'            => 5,
                                             'CLD SCT NEBBIA INF' => 6,
                                             'MAR CLD'            => 8,
                                             'INVIS'              => 9
                                            }->{$2};
                        $m->{cloudBelowEvol} = {
                                                'NC'           => 0,
                                                'DIM ELEV'     => 1,
                                                'DIM'          => 2,
                                                'ELEV'         => 3,
                                                'DIM ABB'      => 4,
                                                'AUM ELEV'     => 5,
                                                'ABB'          => 6,
                                                'AUM'          => 7,
                                                'AUM ABB'      => 8,
                                                'NEBBIA INTER' => 9
                                               }->{$3}
                            if defined $3;
                        push @{$r->{condValleyLoc}}, $m;
                    }
                    push @{$metar{remark}}, { conditionValley => $r };
                } elsif ($msg =~ s@^(QU(L|K) ?([\d/])(?: ?($re_compass_dir16))?) @@o)
                {
                    my $key = $2 eq 'K' ? 'sea' : 'swell';

                    $r->{s} = $1;
                    if ($3 eq '/') {
                        $r->{notAvailable} = undef;
                    } else {
                        $r->{"${key}CondVal"} = $3;
                    }
                    $r->{locationAnd} = _parseLocations $4 if defined $4;
                    push @{$metar{remark}}, { "${key}Condition" => $r };
                } elsif ($msg =~ s@^($re_vis (MAR) ([1-9]\d*) KM) @@) {
                    $r->{s} = $1;
                    $r->{locationAt} = $2;
                    $r->{visibility}{distance} = { v => $3, u => 'KM' };
                    push @{$metar{remark}}, { visibilityAtLoc => $r };
                } elsif ($msg =~ s@^((?:WIND THR ?|WT)($re_rwy_des) ($re_wind)) @@o)
                {
                    push @{$metar{remark}}, { thrWind => {
                        s        => $1,
                        rwyDesig => $2,
                        wind     => _parseWind $3
                    }};
                } elsif ($msg =~ s@^($re_vis ?MIN ($re_vis_m)($re_compass_dir)?) @@o) {
                    $r->{s} = $1;
                    $r->{distance} = _getVisibilityM $2;
                    $r->{compassDir} = $3 if defined $3;
                    push @{$metar{remark}}, { visMin => $r };
                } else {
                    $parsed = 0;
                }
            } elsif($cy[1] eq 'LK' && $msg =~ s@^(REG QNH ([01]\d{3})) @@) {
                push @{$metar{remark}}, { regQNH => {
                    s   => $1,
                    hPa => $2
                }};
            } elsif ($cy[0] eq 'ZMUB' && $msg =~ s@^(\d\d) @@) {
                push @{$metar{remark}}, { RH => {
                    s        => $1,
                    relHumid => $1
                }};
            } else {
                $parsed = 0;
            }
            if (!$parsed && _cyInString \@cy, ' cCA cJP cUS BG EG ET EQ LI NS ') {
                $parsed = 1;
                if ($msg =~ s@^(4([01]\d{3})([01]\d{3})?) @@) {
                    $r->{s} = $1;
                    $r->{tempMaxPeriod}{s}      = $2;
                    $r->{tempMaxPeriod}{period} = '24h';
                    $r->{tempMaxPeriod}{temp}   = _parseTemp $2;
                    if (defined $3) {
                        $r->{tempMinPeriod}{s}      = $3;
                        $r->{tempMinPeriod}{period} = '24h';
                        $r->{tempMinPeriod}{temp}   = _parseTemp $3;
                    }
                    push @{$metar{remark}}, { temp24h => $r };
                } elsif ($msg =~ s@^(4/(\d{3})) @@) {
                    push @{$metar{remark}}, { snowDepth => {
                        s            => $1,
                        precipAmount => { v => $2 + 0, u => 'IN' }
                    }};
                } elsif (   _cyInString(\@cy, ' cCA ')
                         && $msg =~ s@^(SOG (\d+)) @@)
                {
                    push @{$metar{remark}}, { snowDepth => {
                        s            => $1,
                        precipAmount => { v => $2 + 0, u => 'CM' }
                    }};
                } elsif (   !_cyInString(\@cy, ' KAEG KAUN KBVS KUKT KW22 ')
                         && $msg =~ s@^((P|6|7)(?:(\d{4})|////)) @@)
                {
                    # Air Force Manual 15-111, Dec. '03, 2.8.1
                    $r->{s} = $1;
                    if (defined $3) {
                        if ($3 == 0) {
                            $r->{traces} = undef;
                        } else {
                            $r->{precipAmount} = { v => $3 / 100, u => 'IN' };
                        }
                    } else {
                        $r->{notAvailable} = undef;
                    }
                    if ($2 eq 'P') {
                        $r->{period} = '1h';
                    } elsif ($2 eq '7') {
                        $r->{period} = '24h';
                    } else {
                        # EXTENSION: allow 20 minutes
                        if (exists $metar{obsTime}{hour}) {
                            if ((  $metar{obsTime}{hour}
                                 . $metar{obsTime}{minute}) =~
                                '(?:(?:23|05|11|17)[45]|(?:00|06|12|18)[01])\d')
                            {
                                $r->{period} = '6h';
                            } elsif (  ($metar{obsTime}{hour}
                                     . $metar{obsTime}{minute}) =~
                                '(?:(?:02|08|14|20)[45]|(?:03|09|15|21)[01])\d')
                            {
                                $r->{period} = '3h';
                            } else {
                                $r->{period3o6h} = undef;
                            }
                        } else {
                            $r->{period3o6h} = undef;
                        }
                    }
                    push @{$metar{remark}}, { precipHourly => $r };
                # EXTENSION: allow slash after 933 (PATA)
                } elsif ($msg =~ s@^(933/?(\d{3})) @@) {
                    push @{$metar{remark}}, { waterEquivOfSnow => {
                        s            => $1,
                        precipAmount => { v => $2 / 10, u => 'IN' }
                    }};
                } elsif ($msg =~ s@^(CIG (\d{3})V(\d{3})) @@) {
                    push @{$metar{remark}}, { variableCeiling => {
                        s => $1,
                        cloudBaseFrom => { v => $2 * 100, u => 'FT' },
                        cloudBaseTo   => { v => $3 * 100, u => 'FT' }
                    }};
                } elsif ($msg =~ s@^(CIG (\d{4})V(\d{4})) @@) {
                    push @{$metar{remark}}, { variableCeiling => {
                        s => $1,
                        cloudBaseFrom => { v => $2 + 0, u => 'FT' },
                        cloudBaseTo   => { v => $3 + 0, u => 'FT' }
                    }};
                } elsif ($msg =~ s@^(($re_be_weather+)($re_phen_locdirs)?$re_wx_mov_d2?) @@o)
                {
                    $r->{weatherHist}{s} = $1;
                    $r->{weatherHist}{locationAnd} = _parseLocations $3
                        if defined $3;
                    $r->{weatherHist}{$4}{locationAnd} = _parseLocations $5
                        if defined $5;
                    $r->{weatherHist}{weatherBeginEnd} = ();
                    for ($2 =~ /$re_be_weather/og) {
                        my (@startEndArr, $weatherBeginEnd);
                        my ($weather, $times) = /(.*?)($re_be_weather_be+)/g;
                        for ($times =~ /$re_be_weather_be/og) {
                            my %s_e;
                            /(.)(..)?(..)/;
                            $s_e{hour} = $2 if defined $2;
                            $s_e{minute} = $3;
                            if ($1 eq 'B') {
                                push @startEndArr, { beganAt => \%s_e };
                            } else {
                                push @startEndArr, { endedAt => \%s_e };
                            }
                        }
                        $weatherBeginEnd = _parseWeather $weather, 'NI';
                        delete $weatherBeginEnd->{s};
                        $weatherBeginEnd->{startEndArr} = \@startEndArr;
                        push @{$r->{weatherHist}{weatherBeginEnd}}, $weatherBeginEnd;
                    }
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ s@^WSHFT (($re_hour)?($re_min)( FROPA)?) @@o) {
                    $r->{windShift}{s} = "WSHFT $1";
                    $r->{windShift}{hour} = $2 if defined $2;
                    $r->{windShift}{minute} = $3;
                    $r->{windShift}{FROPA} = undef if defined $4;
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ s@^($re_vis ($re_vis_sm)(?: TO)? ($re_compass_dir16)) @@o){
                    push @{$metar{remark}}, { visListAtLoc => {
                        s => $1,
                        visLocData => [
                          { visibility  => _getVisibilitySM($2, $is_auto, \@cy),
                            locationAnd => _parseLocations $3
                    }]}};
                } elsif ($msg =~ s@^($re_vis($re_phen_locdirs ?$re_vis_sm(?: SM)?(?:(?: AND)?$re_phen_locdirs ?$re_vis_sm(?: SM)?)*)) @@o)
                {
                    my @visLocData;

                    $r->{visListAtLoc}{s} = $1;
                    $r->{visListAtLoc}{visLocData} = \@visLocData;
                    for ($2 =~ m@$re_phen_locdirs ?$re_vis_sm@og) {
                        m@($re_phen_locdirs) ?($re_vis_sm)@o;
                        push @visLocData, {
                            visibility => _getVisibilitySM($2, $is_auto, \@cy),
                            locationAnd => _parseLocations $1
                        };
                    }
                    push @{$metar{remark}}, $r;
                } else {
                    $parsed = 0;
                }
            }
            if (!$parsed && $msg =~ s@^(BA (?:GOOD|POOR)|THN SPTS IOVC|FUOCTY|ALTM MISG|ALL WNDS GRID|CONTRAILS?|(?:SKY|MT(?:N?S)?|RDGS) OBSC(?:URED|D)?|FOGGY|TWLGT|WEA) @@)
            {
                $parsed = 1;
                $r->{s} = $1;
                ($r->{v} = $1) =~ tr/\/ /_/;
                $r->{v} =~ s/CONTRAILS?/CONTRAILS/;
                $r->{v} =~ s/OBSC(?:URE)?D/OBSC/;
                $r->{v} =~ s/MT(?:N?S)?/MTNS/;
                push @{$metar{remark}}, { keyword => $r };
            }
            if (!$parsed) {
                $msg =~ s/(\S+) //;
                $notRecognised .= ' ' unless $notRecognised eq '';
                $notRecognised .= $1;
            }
            if ($parsed && $notRecognised ne '') {
                my $top = pop @{$metar{remark}};
                push @{$metar{remark}},
                                    { notRecognised => { s => $notRecognised }};
                push @{$metar{remark}}, $top;
                $notRecognised = '';
            }
        }
        push @{$metar{remark}}, { notRecognised => { s => $notRecognised }}
            if $notRecognised ne '';
    }

    $metar{ERROR} = _makeErrorMsg 'other', $corr_msg, $msg if $msg ne '';
    return %metar;
}

=head1 DEPENDENCIES

The function C<POSIX::floor> from the module C<POSIX> is required.

=head1 FUNCTIONS

=cut

########################################################################
# parseReport
########################################################################
sub parseReport {
    my ($msg, $is_taf, $is_synop) = @_;
    my ($is_speci, $msg_hdr);

=head2 parseReport($msg,$is_taf,$is_synop)

This is the main function of the module.

The following arguments are expected:

=over

=item C<msg>

string that contains the message

=item C<is_taf>

boolean value, indicating whether it is a TAF message or not

=item C<is_synop>

boolean value, indicating whether it is a SYNOP message or not

=back

Leading and trailing spaces are removed from C<msg>, multiple spaces are
replaced by a single one. Characters that are invalid in HTML or XML are also
removed.

=cut

    $msg =~ s/ +/ /g;
    $msg =~ s/ $//;
    $msg =~ s/^ //;
    $msg =~ s/[^ -~]/?/gos; # avoid invalid XML and HTML

=pod

If the message starts with C<METAR>, C<SPECI>, C<TAF>, or C<SYNOP>, the
arguments C<is_taf> and C<is_synop> are ignored.

=cut

    $msg_hdr = '';
    if ($msg =~ s/^(METAR )//) {
        $is_taf = 0;
        $is_synop = 0;
        $msg_hdr = $1;
    } elsif ($msg =~ s/^(SPECI )//) {
        $is_taf = 0;
        $is_synop = 0;
        $msg_hdr = $1;
        $is_speci = 1;
    } elsif ($msg =~ s/^(TAF )//) {
        $is_taf = 1;
        $is_synop = 0;
        $msg_hdr = $1;
    } elsif ($msg =~ s/^(SYNOP )//) {
        $is_synop = 1;
        $msg_hdr = $1;
    }

=pod

Then the correct function to parse the message is called.

=cut

    return _parseSynop $msg, $msg_hdr if $is_synop;
    return _parseMetarTaf $msg, $msg_hdr, $is_speci, $is_taf;
}

=pod

The return value is a hash with all the parsed components of the message. It
may also have values for the following keys:

=over

=item C<ERROR>

If a message could not be parsed C<ERROR> is a hash with the the keys
C<errorType> (the type of error) and C<s> - the original message with the
position of the error marked as C<< <@> >>.

=item C<warning>

An array of hashes describing problems which where encountered during
parsing but which didn't prevent complete parsing.

=back

=head1 SEE ALSO

=begin html

<a href=XML.pm.html>metaf2xml::XML(3pm)</a>,
<a href=metaf2xml.pl.html>metaf2xml(1)</a>,
<a href=metaf.pl.html>metaf(1),</a> <!--

=end html

B<metaf2xml::XML>(3pm),
B<metaf2xml>(1),
B<metaf>(1),

=for html -->

L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT, LICENSE, DISCLAIMER

copyright (c) 2006-2009 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

=cut

1;
